/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */


#include "jwp.h"

#include "idm.h"

#include <commdlg.h>


#define DICTINDENT   5
#define NRCACHE      20
#define STRIKEOUT    127

#define LEADINGONLY  FixedDictKeywords[0][0]

typedef struct {
    INT32 offset;
    BYTE far *string;
} DICTCACHE;

typedef struct {
    INT32 start;
    INT32 end;
} DICTBOUNDS;

static DICTCACHE cache[NRCACHE];

static DICTBOUNDS *DictBounds = NULL;
static int totalmatches;
static BOOL SearchCancelled;

#ifdef USEWIN32
  static HANDLE DictSearchThread;
  static DWORD DictThreadId;
#endif USEWIN32

static char *CurrentDict, *CurrentIndex, *CurrentDesc;

BOOL FAR PASCAL DictionaryOptionsProc (HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK EditDictionaryProc (HWND, UINT, WPARAM, LPARAM);

extern int bytelen (BYTE far *);



static BOOL CalcDictionaryBounds(void)
{
    int i;
    HFILE fd;
    INT32 offset, len;
    OFSTRUCT of;
    BOOL ok = FALSE;

    if (DictBounds != NULL) FreeMem(DictBounds);

    DictBounds = MemAlloc((global.nr_jdict + 1) * sizeof(DICTBOUNDS));
    if (DictBounds == NULL) return (FALSE);

    offset = 0L;

    for (i = 0; i < global.nr_jdict; i++) {
        DictBounds[i].start = DictBounds[i].end = -1;

        if (!global.jdictactive[i]) continue;
        fd = OpenFile(global.jdict[i], &of, OF_READ);
        if (fd == HFILE_ERROR) continue;

        len = _llseek(fd, 0L, 2);
        if (len > 0L) {
            DictBounds[i].start = offset;
            offset += len;
            DictBounds[i].end = offset - 1L;
            ok = TRUE;
        }
        _lclose(fd);
    }

    return (ok);
}



static void FlushCache (void)
{
    int i;

    for (i = 0; i < NRCACHE; i++) {
        if (cache[i].string != NULL) {
            FreeBlock(cache[i].string);
            cache[i].string = NULL;
        }
        cache[i].offset = 0L;
    }
}



static int GetDictionary(INT32 offset)
{
    int i;

    if (DictBounds == NULL) return (-1);

    for (i = 0; i < global.nr_jdict; i++) {
        if (DictBounds[i].start < 0L) continue;
        if (DictBounds[i].start <= offset && offset <= DictBounds[i].end) break;
    }

    if (i >= global.nr_jdict) return (-1);
    return (i);
}



static BOOL CensorEntry(BYTE *s)
{
    int i, j, k, x;
    BOOL censor;
    BYTE *p, *q;
    BYTE tmp[MAXLINELEN];

    p = strtok(s, "/");

    for (;;) {
        if (p == NULL || !p[0]) break;

        censor = FALSE;

        for (i = 0; ; i++) {
            for (; p[i] && p[i] != '('; i++);
            if (!p[i]) break;

            /* Found a paren - copy what's inside */

            for (i++, j = 0; p[i] && p[i] != ')'; i++, j++) tmp[j] = p[i];
            tmp[j] = '\0';

            /* Now we have what's inside */

            k = 0;

            for (;tmp[k];) {
                x = k;
                for (; tmp[k] && tmp[k] != ','; k++);

                if (tmp[k]) {
                    tmp[k++] = '\0';
                    tmp[k] = '\0';
                }

                for (j = 0; j < global.nr_dictkeywords; j++) {
                    if (global.dictkeywords[j][0] == '(') continue;

                    if (!bytecmp(global.dictkeywords[j], tmp + x) &&
                        !global.dictkeywordsshow[j]) {
                            censor = TRUE;
                            break;
                    }
                }

                if (censor) break;
            }

            /* Censor it? */

            if (censor) {
                for (j = 0; p[j]; j++) p[j] = STRIKEOUT;
                break;
            }

            if (!p[i]) break;
        }

        /* Put back the '/' */

        q = strtok(NULL, "/");

        k = bytelen(p);
        p[k] = (censor ? STRIKEOUT : '/');

        if (q == NULL) {
            p[k+1] = '\0';
            break;
        }
        p = q;
    }

    /* Censor the entire entry? */

    for (i = 0; s[i] && s[i] != '/'; i++);
    if(!s[i]) return (FALSE);

    for (i++; s[i]; i++) {
        if (s[i] != STRIKEOUT && s[i] != '\r' && s[i] != '\n' && s[i] != '/')
            return (TRUE);
    }
    return (FALSE);
}



static KANJI far *ConvertOffset (int id, LPARAM lParam, KANJI *buf)
{
    FILE *fp;
    int i, j, k, dictno;
    HFILE fd;
    INT32 offset;
    int length;
    OFSTRUCT dictof;
    BYTE tmp[MAXLINELEN];

    length = (lParam >> 24) & 0x00ff;
    offset = lParam & 0x00ffffffL;

    /* In the cache? */

    for (i = 0; i < NRCACHE; i++) {
        if (cache[i].string == NULL) continue;
        if (cache[i].offset == offset) break;
    }

    if (i < NRCACHE) goto ProcessLine;


    /* Not in the cache */

    dictno = GetDictionary(offset);
    if (dictno < 0) return ((KANJI far *) buf);

#ifdef USEWIN32
    fp = fopen(global.jdict[dictno], "rb");
    if (fp == NULL) {
        return ((KANJI far *) buf);
    }
#else
    fd = OpenFile(global.jdict[dictno], &dictof, OF_READ);
    fp = fdopen(fd, "rb");
    if (fp == NULL) {
        _lclose(fd);
        return ((KANJI far *) buf);
    }
#endif

    /* Prepare the cache */

    for (i = 0; i < NRCACHE && cache[i].string != NULL; i++);
    if (i >= NRCACHE) {
        INT32 diff, maxdiff;

        for (i = maxdiff = 0, j = -1; i < NRCACHE; i++) {
            diff = cache[i].offset - offset;
            if (diff < 0) diff = -diff;
            if (diff > maxdiff) {
                j = i;
                maxdiff = diff;
            }
        }
        if (j < 0) j = rand() % NRCACHE;
        i = j;
    }

    if (cache[i].string != NULL) FreeBlock(cache[i].string);
    cache[i].string = (BYTE far *) BlockAlloc((length + 5) * sizeof(BYTE));
    cache[i].offset = offset;

    /* Get the line */

    fseek(fp, offset - DictBounds[dictno].start, 0);

    for (j = 0; j < length; j++) {
        cache[i].string[j] = fgetc(fp);
    }
    cache[i].string[j] = '\0';
    fclose(fp);


ProcessLine:

    _fstrcpy(tmp, cache[i].string);
    CensorEntry(tmp);

    k = bytelen(tmp);
    if (tmp[k-1] == '\r' || tmp[k-1] == '\n' || tmp[k-1] == '/')
        tmp[k-1] = '\0';

    for (k = bytelen(tmp) - 1; k >= 0 && tmp[k] == STRIKEOUT; k--);
    if (k < 0) {
        buf[0] = 0;
        return ((KANJI far *) buf);
    }
    if (tmp[k] == '/') tmp[k] = '\0';

    for (j = k = 0; tmp[j]; j++) {
        if (tmp[j] == STRIKEOUT) continue;
        if (tmp[j] & 0x80) {
            buf[k] = (tmp[j++] & 0x007f) << 8;
            buf[k++] |= tmp[j] & 0x007f;
        } else if (tmp[j] == '/') {
            buf[k++] = ',';
            buf[k++] = ' ';
        } else {
            buf[k++] = tmp[j];
        }
    }

    buf[k] = 0;

    if (k <= 0) {
        return ((KANJI far *) buf);
    } else if (!ISKANJI(buf[0])) {
        for (i = k; i >= 0; i--) buf[i+DICTINDENT] = buf[i];
        for (i = 0; i < DICTINDENT; i++) buf[i] = ' ';
    } else {
        for (i = 0; buf[i] != 0; i++) {
            if (buf[i] == '[') buf[i] = 0x215a;
            else if (buf[i] == ']') buf[i] = 0x215b;
            else if (buf[i] == ' ') buf[i] = 0x2121;
        }
    }

    return ((KANJI far *) buf);
}



static BOOL CheckDictionary (int num)
{
    HFILE fd;
    OFSTRUCT dictof, idxof;
    INT32 indexsize;
    INT32 dictsize;


    for (;;) {
        fd = OpenFile(global.jdict[num], &dictof, OF_READ);
        if (fd != HFILE_ERROR) break;
        if (!RetryMessage ("Cannot open dictionary '%s'!", global.jdict[num]))
            return (FALSE);
    }
    dictsize = _llseek(fd, 0L, 2);
    _lclose(fd);

    for (;;) {
        fd = OpenFile(global.jdictidx[num], &idxof, OF_READ);
        if (fd != HFILE_ERROR) break;
        if (!RetryMessage ("Cannot open dictionary index '%s'!", global.jdictidx[num]))
            return (FALSE);
    }

    _llseek(fd, 0L, 0);
    _lread(fd, &indexsize, sizeof(INT32));
    _lclose(fd);

    if (indexsize != dictsize + 15) {
        ErrorMessage(global.hwnd, "This index file does not belong to this dictionary!");
        return (FALSE);
    }

    return (TRUE);
}



static int DictComp (BYTE *p1, BYTE *p2, int n)
{
    int i;
    BYTE c1, c2;
    BOOL f1 = FALSE, f2 = FALSE;

    for (i = 0; ; i++, p1++, p2++) {
        if (n > 0 && i >= n) return (0);

        c1 = *p1;
        c2 = *p2;

        if ('A' <= c1 && c1 <= 'Z') c1 += 32;
        if ('A' <= c2 && c2 <= 'Z') c2 += 32;

        if (c1 & 0x80) {
            f1 = !f1;
            if (f1 && c1 == 0xa5) c1 = 0xa4;        /* Katakana? */
        } else {
            f1 = FALSE;
        }
        if (c2 & 0x80) {
            f2 = !f2;
            if (f2 && c2 == 0xa5) c2 = 0xa4;        /* Katakana? */
        } else {
            f2 = FALSE;
        }

        if (c1 != c2) return (c1 - c2);
        if (c1 == 0) return (0);
    }

    return (c1 - c2);
}



static int SearchDictionary (int num, HWND DlgHwnd, HWND KeyHwnd, HWND ListHwnd)
{
    int i, j, k, last;
    int nr_matches, nr_rejected;
    HFILE fd;
    OFSTRUCT dictof, idxof;
    BOOL PassedKanji;
    FILE *fp;
    INT32 nr_index;
    INT32 top, bottom, middle;
    INT32 offset, cut;
    int diff, length;
    MSG msg;
    BYTE key[MAXLINELEN], target[MAXLINELEN];
    KANJI ch;
    KANJI far *up;
    RECT rect;


    while (PeekMessage(&msg, DlgHwnd, 0, 0, PM_REMOVE)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    /* Opens the dictionary and index files */

    if (!CheckDictionary(num)) return (0);

#ifdef USEWIN32
    fp = fopen(global.jdict[num], "rb");
    if (fp == NULL) return (0);
#else
    fd = OpenFile(global.jdict[num], &dictof, OF_READ);
    if (fd == HFILE_ERROR) return (0);

    fp = fdopen(fd, "rb");
    if (fp == NULL) {
        _lclose(fd);
        return (0);
    }
#endif USEWIN32

    fd = OpenFile(global.jdictidx[num], &idxof, OF_READ);
    if (fd == HFILE_ERROR) {
        fclose(fp);
        return (0);
    }

    nr_index = _llseek(fd, 0L, 2);       /* Move to EOF */

    nr_index = (nr_index / sizeof(INT32)) - 1;
    nr_matches = nr_rejected = 0;


    /* Create the key string */

    up = (KANJI far *) SendMessage(KeyHwnd, EM_GETLINE, 0, 0L);

    for (i = j = 0; ; j++) {
        ch = up[j];
        if (ch == 0) break;

        if (ISKANJI(ch)) {
            key[i++] = HIBYTE(ch) | 0x0080;
            key[i++] = LOBYTE(ch) | 0x0080;
        } else {
            key[i++] = ch;
        }
    }

    key[i] = '\0';
    length = i;


    if (totalmatches <= 0) {
        SetDlgItemText(DlgHwnd, 4202, "Searching...");
        InvalidateRect(GetDlgItem(DlgHwnd, 4202), NULL, FALSE);
        UpdateWindow(GetDlgItem(DlgHwnd, 4202));
    }


    /* Binary Search */

    top = 0;
    bottom = nr_index - 1;

    for (;;) {
        middle = (top + bottom) / 2;

        /* Get the offset */

        _llseek(fd, (middle + 1) * sizeof(INT32), 0);
        _lread(fd, &offset, sizeof(INT32));


        /* Get the target string */

        fseek(fp, offset - 1, 0);
        fgets(target, MAXLINELEN, fp);

        diff = DictComp(key, target, -1);

        if (diff == 0) {
            cut = middle;
            break;
        }

        if (top >= bottom - 1) {
            cut = bottom;
            break;
        }

        if (diff > 0) top = middle;
        else bottom = middle;
    }

    SendMessage(ListHwnd, WM_SETREDRAW, FALSE, 0L);

    GetClientRect(ListHwnd, &rect);
    rect.right -= 2 * GetSystemMetrics(SM_CXVSCROLL);


    for (;; cut++) {
        /* Now, did the user abort? */

#ifdef USEWIN32
        while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }

        if (DictSearchThread == NULL) {
#endif USEWIN32
            if (PeekMessage(&msg, DlgHwnd, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE)) {
                if (msg.message == WM_CHAR && msg.wParam == '\x1b') {
					if (YesNo(DlgHwnd, "Dictionary search interrupted!\n\nContinue?") != IDYES) {
						SearchCancelled = TRUE;
						break;
					}
				} else {
					TranslateMessage(&msg);
				}
				//DispatchMessage(&msg);
            }
#ifdef USEWIN32
        } else {
			if (SearchCancelled) {
				if (YesNo(DlgHwnd, "Dictionary search interrupted!\n\nContinue?") != IDYES) break;
			}
			SearchCancelled = FALSE;
        }

        if (totalmatches == 10) {
            SendMessage(ListHwnd, WM_SETREDRAW, TRUE, 0L);
            InvalidateRect(ListHwnd, NULL, TRUE);
        }
#endif USEWIN32

        _llseek(fd, (cut + 1) * sizeof(INT32), 0);
        _lread(fd, &offset, sizeof(INT32));

        /* Get the text string */

        offset--;
        fseek(fp, offset, 0);
        fgets(target, MAXLINELEN, fp);


        /* Still matches? */

        if (DictComp(key, target, length)) break;


        /* OK, now backseek to the beginning of the line, shall we? */

        fseek(fp, offset, 0);

        for (;;) {
            if (offset < MAXLINELEN) {
                fseek(fp, 0L, 0);
				j = offset;
                offset = 0L;
            } else {
                fseek(fp, -MAXLINELEN, 1);      /* Get the previous block */
                j = MAXLINELEN;
                offset -= MAXLINELEN;
            }
            fread(target, sizeof(BYTE), MAXLINELEN, fp);

            for(i = j - 1; i >= 0; i--) {
                if (target[i] == '\n' || target[i] == '\r') break;
            }

            if (i > 0) {
                i++;
                offset += i;
                break;
            }

            if (offset <= 0L) break;
        }


        fseek(fp, offset, 0);
		fgets(target, MAXLINELEN, fp);


        /* Discard non-leading entries? */

        for (i = 0; i < global.nr_dictkeywords; i++) {
            if (!bytecmp(global.dictkeywords[i], LEADINGONLY)) {
                break;
            }
        }

        if (i < global.nr_dictkeywords && !global.dictkeywordsshow[i]) {
            if (DictComp(key, target, length)) {
                /* Find the kana string */
                for (j = 0; target[j] && target[j] != '['; j++);
                if (target[j] != '[') continue;

                if (DictComp(key, target + j + 1, length)) continue;
            }
        }


        /* Censor the entry */

        if (!CensorEntry(target)) {
            nr_rejected++;
            continue;
        }


        /* Update the indicator */

		for (i = strlen(target) - 1; i >= 0; i--) {
            if (target[i] != '\n' && target[i] != '\r' && target[i] != '/') break;
        }
        target[i+1] = '\0';

        nr_matches++;
        totalmatches++;

        if (totalmatches > 0) {
            char tempbuf[50];

            sprintf(tempbuf, "%d match%s", totalmatches, (totalmatches > 1) ? "es" : "");
            SetDlgItemText(DlgHwnd, 4202, tempbuf);
            InvalidateRect(GetDlgItem(DlgHwnd, 4202), NULL, FALSE);
            UpdateWindow(GetDlgItem(DlgHwnd, 4202));
        }


        /* Now break it down into individual lines */

        last = -1;
        PassedKanji = FALSE;


        for (i = j = k = 0; ; i++) {
            if (target[i] <= ' ' || target[i] == '/') last = j;

            if (target[i]) {
                if (target[i] == STRIKEOUT) {
                    /* Nothing */
                } else if (target[i] == '/') {
                    if (!PassedKanji) {
                        SendMessage(ListHwnd, LB_ADDSTRING, 0,
                                    ((offset + DictBounds[num].start) & 0x00ffffffL) | (((LONG) j) << 24));
                        offset += (j + 1);
                        j = 0;
                        k = DICTINDENT * FontCharWidth(' ', 0);     /* System font */
                        PassedKanji = TRUE;
                        continue;
                    }
                    k += FontCharWidth(',', 0);     /* System font */
                    k += FontCharWidth(' ', 0);     /* System font */
                } else {
                    k += FontCharWidth(target[i], 0);   /* System font */
                }
            }

            if (!target[i]) {
                SendMessage(ListHwnd, LB_ADDSTRING, 0,
                        ((offset + DictBounds[num].start) & 0x00ffffffL) | (((LONG) j) << 24));
                break;
            } else if (PassedKanji && k >= rect.right) {
                SendMessage(ListHwnd, LB_ADDSTRING, 0,
                        ((offset + DictBounds[num].start) & 0x00ffffffL) | (((LONG) last + 1) << 24));

                offset += (last + 1);
                i -= (j - last);
                j = 0;
                k = DICTINDENT * FontCharWidth(' ', 0);     /* System font */
            } else {
                j++;
            }
        }
    }

    SendMessage(ListHwnd, WM_SETREDRAW, TRUE, 0L);
#ifndef USEWIN32
    InvalidateRect(ListHwnd, NULL, TRUE);
#endif USEWIN32

    fclose(fp);
    _lclose(fd);
    return (nr_matches);
}



LONG FAR PASCAL InputModeIconProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case EM_SETHANDLE:
#ifdef USEWIN32
            SetProp(hwnd, PROP_HWND, (HANDLE) lParam);
#else
            SetProp(hwnd, PROP_HWND, LOWORD(lParam));
#endif USEWIN32
            return (0);

        case WM_LBUTTONDOWN: {
            FILEOPTIONS *f;
            HWND hwnd1;

            hwnd1 = (HWND) GetProp(hwnd, PROP_HWND);
            f = (FILEOPTIONS *) SendMessage(hwnd1, EM_GETHANDLE, 0, 0L);
            if (f->type & FN_EITHER && f->nr_bytes > 0) {
                MessageBeep(MB_ICONHAND);
                return (0);
            }
            SendMessage(hwnd1, WM_COMMAND, IDM_TOGGLEMODE, 0L);
            InvalidateRect(hwnd, NULL, FALSE);
            return (0);
        }

        case WM_PAINT: {
            HDC hdc, hdcmem;
            HBITMAP hbitmap;
            PAINTSTRUCT ps;

            SetWindowPos(hwnd, NULL, 0, 0, 24, 24, SWP_NOMOVE | SWP_NOZORDER);

            /* Put in the bitmap */
            hdc = BeginPaint(hwnd, &ps);
            hdcmem = CreateCompatibleDC(hdc);

            switch (global.mode) {
                case M_ASCII: hbitmap = LoadBitmap(hInstance, "ASCIIBmp"); break;

                default:
                case M_KANA:  hbitmap = LoadBitmap(hInstance, "KanaBmp"); break;
            }

            SelectObject(hdcmem, hbitmap);
            BitBlt(hdc, 0, 0, 24, 24, hdcmem, 0, 0, SRCCOPY);
            DeleteDC(hdcmem);
            DeleteObject(hbitmap);
            EndPaint(hwnd, &ps);
            return (0);
        }

        case WM_DESTROY:
            ClearAllProperties(hwnd);
            break;
    }

    return (DefWindowProc(hwnd, message, wParam, lParam));
}



#ifdef USEWIN32

static DWORD DictSearchThreadProc (HWND hwnd)
{
    int dictno, n;

	EnableWindow(GetDlgItem(hwnd, 4211), FALSE);    /* Search button */
	SearchCancelled = FALSE;

    AnimateIcon(GetDlgItem(hwnd, 4241), "X_BOOKOPEN", 6, 2.0);

    n = totalmatches = 0;

    for (dictno = 0; dictno < global.nr_jdict; dictno++) {
        if (!global.jdictactive[dictno]) continue;
        if (!CheckDictionary(dictno)) continue;
        n += SearchDictionary(dictno, hwnd, GetDlgItem(hwnd, 4201), GetDlgItem(hwnd, 4221));
        if (SearchCancelled) break;
    }

    AnimateIcon(NULL, NULL, 0, 0.0);

    if (n <= 0) {
        MessageBeep(MB_ICONHAND);
		SetDlgItemText(hwnd, 4202, "No matches");
		InvalidateRect(GetDlgItem(hwnd, 4202), NULL, FALSE);
		UpdateWindow(GetDlgItem(hwnd, 4202));
	} else {
		SetFocus(GetDlgItem(hwnd, 4221));
	}

	DictSearchThread = NULL;
	EnableWindow(GetDlgItem(hwnd, 4211), TRUE);
	return (n);
}

#endif USEWIN32


BOOL FAR PASCAL DictionaryProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {

        case WM_INITDIALOG: {
            int i, len;
            BOOL IsKanji1, IsKanji2;
            KANJI buf[MAXLINELEN];
            FILEOPTIONS *f, *f1;

            /* Set the type and mode-change icon of the Jedit control */
            SendDlgItemMessage(hwnd, 4201, EM_SETMODIFY, 0, FN_CONTROL | FN_EITHER);
            SendDlgItemMessage(hwnd, 4201, EM_SETRECT, 0, (LPARAM) GetDlgItem(hwnd, 4231));

            f = (FILEOPTIONS *) SendDlgItemMessage(hwnd, 4201, EM_GETHANDLE, 0, 0L);

            SendDlgItemMessage(hwnd, 4231, EM_SETHANDLE, 0, (LPARAM) f->hwnd);  /* icon-change */

            if (global.active != NULL) f1 = global.active;

            if (global.active == NULL || SELPARA1(f1) == NULL) {
                if (global.mode != M_ASCII) ToggleInputMode();
            } else {
                /* Copy the selected text */

                if (SELPARA1(f1) != SELPARA2(f1)) {
                    len = unitlen(SELPARA1(f1)->text) - SELPOS1(f1);
                } else {
                    len = SELPOS2(f1) - SELPOS1(f1) + 1;
                }

                IsKanji1 = ISKANJI(KANJIOF(SELCHAR1(f1)));

                for (i = 0; i < len; i++) {
                    buf[i] = KANJIOF(TEXTOF(SELPARA1(f1), SELPOS1(f1) + i));
                    IsKanji2 = ISKANJI(buf[i]);
                    if (IsKanji1 && !IsKanji2) break;
                    if (!IsKanji1 && IsKanji2) break;
                }
                buf[i] = 0;

                SendDlgItemMessage(hwnd, 4201, EM_REPLACESEL, 0, (LPARAM) buf);

                if (len > 0) {
                    SendDlgItemMessage(hwnd, 4201, EM_SETSEL, len, MAKELONG(0, len - 1));
                }
            }

            EnableWindow(GetDlgItem(hwnd, 4212), FALSE);
            EnableWindow(GetDlgItem(hwnd, 4211), (f->nr_bytes > 0));

            for (i = 0; i < NRCACHE; i++) cache[i].string = NULL;

            SetFocus(GetDlgItem(hwnd, 4201));

            TrapHelpInDialogBoxes(hwnd, "DictionaryDlg", 1, TRUE);
            CenterDialogBox(hwnd);

            return (TRUE);
        }

        case WM_PAINT: {
            HDC hdc;
            PAINTSTRUCT ps;

            hdc = BeginPaint(hwnd, &ps);

            DrawBoundingBox(hwnd, hdc, 4201);

            EndPaint(hwnd, &ps);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND: {
            int i, n, dictno;
            KANJI far *up;
            HCURSOR hCursor;
            HWND ChildHwnd;


            switch (WINCMDID) {
                case 4201:          /* Japanese Edit Control */
                    if (NOTIFYCODE == EN_CHANGE) {
                        if (SendMessage(NOTIFYHWND, EM_LINELENGTH, 0, 0L) > 0) {
                            EnableWindow(GetDlgItem(hwnd, 4211), TRUE);
                        } else {
                            EnableWindow(GetDlgItem(hwnd, 4211), FALSE);
                        }
                    }
                    return (TRUE);

                case 4211:          /* Search Button */
                case IDOK:
                    if (global.nr_jdict <= 0) {
                        ErrorMessage(hwnd, "You do not have any dictionaries installed!  "
                                           "Press the Options button to add dictionaries.");
                        return (TRUE);
                    }
#ifdef USEWIN32
                    if (DictSearchThread != NULL) {
                        MessageBeep(MB_ICONHAND);
                        return (TRUE);
                    }
#endif USEWIN32
                    FlushCache();
                    if (!CalcDictionaryBounds()) {
                        MessageBeep(MB_ICONHAND);
                        return (TRUE);
                    }

                    up = (KANJI far *) SendDlgItemMessage(hwnd, 4201, EM_GETLINE, 0, 0L);
                    i = kanjilen(up);

                    if (!ISKANJI(up[0]) && i <= 2) {
                        /* Cannot search for less than 3 letters */
                        MessageBeep(MB_ICONHAND);
                    } else if (ISKANJI(up[0]) && i <= 1 &&
                               (HIBYTE(up[0]) == 0x24 || HIBYTE(up[0]) == 0x25)) {

                        /* Cannot search for less than 2 kana */
                        MessageBeep(MB_ICONHAND);
                    } else if (i > 0) {
                        hCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
                        ShowCursor(TRUE);
                        EnableWindow(GetDlgItem(hwnd, 4212), FALSE);

                        SetDlgItemText(hwnd, 4202, "");
                        n = totalmatches = 0;
						SearchCancelled = FALSE;

						/* Deletes everything from the list box */

						if (SendMessage(GetDlgItem(hwnd, 4221), LB_GETCOUNT, 0, 0L) > 0)
							SendMessage(GetDlgItem(hwnd, 4221), LB_RESETCONTENT, 0, 0L);

#ifdef USEWIN32
						DictSearchThread = CreateThread(NULL, BLOCKSIZE, (LPTHREAD_START_ROUTINE) DictSearchThreadProc, hwnd, 0, &DictThreadId);

						if (DictSearchThread == NULL) {
                            AnimateIcon(GetDlgItem(hwnd, 4241), "X_BOOKOPEN", 6, 2.0);
                            for (dictno = 0; dictno < global.nr_jdict; dictno++) {
                                if (!global.jdictactive[dictno]) continue;
                                if (!CheckDictionary(dictno)) continue;
                                n += SearchDictionary(dictno, hwnd, GetDlgItem(hwnd, 4201), GetDlgItem(hwnd, 4221));
                                if (SearchCancelled) break;
                            }
                            AnimateIcon(NULL, NULL, 0, 0.0);

                            if (n <= 0) {
                                MessageBeep(MB_ICONHAND);
                                SetDlgItemText(hwnd, 4202, "No matches");
                            } else {
                                SetFocus(GetDlgItem(hwnd, 4221));
                            }
                        }
#else
                        for (dictno = 0; dictno < global.nr_jdict; dictno++) {
                            if (!global.jdictactive[dictno]) continue;
                            if (!CheckDictionary(dictno)) continue;
                            n += SearchDictionary(dictno, hwnd, GetDlgItem(hwnd, 4201), GetDlgItem(hwnd, 4221));
                            if (SearchCancelled) break;
                        }

                        if (n <= 0) {
                            MessageBeep(MB_ICONHAND);
                            SetDlgItemText(hwnd, 4202, "No matches");
                        } else {
                            SetFocus(GetDlgItem(hwnd, 4221));
                        }
#endif USEWIN32

                        ShowCursor(FALSE);
                        SetCursor(hCursor);
                    } else {
                        MessageBeep(MB_ICONHAND);
                    }
                    return (TRUE);

                case 4212: {        /* Insert Button */
                    KANJI buf[MAXLINELEN], far *kp;
                    LONG offset;
                    FILEOPTIONS *f;

                    ChildHwnd = GetDlgItem(hwnd, 4221);

                    if (NOTIFYCODE == BN_CLICKED && global.active != NULL) {
                        n = SendMessage(ChildHwnd, LB_GETCURSEL, 0, 0L);
                        if (n == LB_ERR) {
                            EnableWindow(NOTIFYHWND, FALSE);
                            return (TRUE);
                        }

                        f = global.active;
                        offset = SendMessage(ChildHwnd, LB_GETITEMDATA, n, 0L);
                        kp = ConvertOffset(0, offset, buf);

                        /* Trim surrounding spaces */
                        for (i = kanjilen(kp); i > 0; i--) {
                            if (ISKANJI(kp[i])) {
                                if (kp[i] > 0x2121) break;
                            } else {
                                if (kp[i] > ' ') break;
                            }
                        }

                        kp[i+1] = 0;

                        for (i = 0; kp[i]; i++) {
                            if (ISKANJI(kp[i])) {
                                if (kp[i] > 0x2121) break;
                            } else {
                                if (kp[i] > ' ') break;
                            }
                        }

                        if (kp[i] && CheckInputTextType(f, PARAOF(f->current)->text, kp + i)) {
                            TurnOffSelection(f);
                            UndoAddTyping(f, f->current, kanjilen(kp + i), FALSE);
                            InsertString(f, f->current, kp + i,
                                OP_REFORMAT | OP_UPDATE | OP_MOVESEL | OP_MOVETOEND);
                        } else {
                            MessageBeep(MB_ICONHAND);
                            return (TRUE);
                        }
                    }

                    //SendMessage(hwnd, WM_COMMAND, IDCANCEL, 0L);
                    return (TRUE);
                }

                case 4213: {        /* Options Button */
#ifdef USEWIN32
                    if (DictSearchThread != NULL) {
                        MessageBeep(MB_ICONHAND);
                        return (TRUE);
                    }
#endif USEWIN32
                    DialogBox(hInstance, "DictionaryOptions", hwnd, DictionaryOptionsProc);
                    return (TRUE);
                }

                case 4221:          /* Japanese List Box */
                    ChildHwnd = GetDlgItem(hwnd, 4212);

                    if (NOTIFYCODE == LBN_SELCHANGE || NOTIFYCODE == LBN_DBLCLK) {
                        n = SendMessage(NOTIFYHWND, LB_GETCURSEL, 0, 0L);
                        if (n == LB_ERR) {
                            EnableWindow(ChildHwnd, FALSE);
                            return (TRUE);
                        }

                        if (global.active != NULL) EnableWindow(ChildHwnd, TRUE);

                        if (NOTIFYCODE == LBN_DBLCLK) {
                            SENDNOTIFY(hwnd, WM_COMMAND, 4212, BN_CLICKED, ChildHwnd);
                        }
                    }
                    return (TRUE);

                case IDCANCEL:      /* Done Button */
#ifdef USEWIN32
                    if (DictSearchThread != NULL) {
                        SearchCancelled = TRUE;
                        return (TRUE);
                    }
#endif USEWIN32
                    for (i = 0; i < NRCACHE; i++) {
                        if (cache[i].string != NULL) FreeBlock(cache[i].string);
                        cache[i].string = NULL;
                    }
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);

                default:
                    break;
            }
            break;
        }

        case WM_COMPAREITEM:
        case WM_DELETEITEM:
        case WM_DRAWITEM:
        case WM_MEASUREITEM:
            return (JlistProc(hwnd, message, wParam, lParam, TRUE, ConvertOffset));
    }

    return (FALSE);
}



static KANJI far *ConvertDictOffsets (int id, LPARAM lParam, KANJI *buf)
{
    int i, j;
    BYTE *p;

    switch (id) {
		case 4221: p = global.dictkeywordsdesc[lParam]; break;
        case 4222: p = global.jdictdesc[lParam]; break;
        default:   buf[0] = 0; return ((KANJI far *) buf);
    }

    for (i = j = 0; p[i]; i++, j++) {
        if (p[i] & 0x80) {
            buf[j] = (p[i++] & 0x7f) << 8;
            buf[j] |= (p[i] & 0x7f);
        } else {
            buf[j] = p[i];
        }
        if (!p[i]) break;
    }

    buf[j] = 0;

    return ((KANJI far *) buf);
}



static BOOL ExtendJdict(int new)
{
    char * *p;

    p = (char * *) MemAlloc((global.nr_jdict + new + 1) * sizeof(char *) + 5);
    if (p == NULL) return (FALSE);

    memcpy(p, global.jdict, global.nr_jdict * sizeof(char *));
    if (global.jdict != NULL) FreeMem(global.jdict);
    global.jdict = p;

    p = (char * *) MemAlloc((global.nr_jdict + new + 1) * sizeof(char *) + 5);
    if (p == NULL) return (FALSE);

    memcpy(p, global.jdictidx, global.nr_jdict * sizeof(char *));
    if (global.jdictidx != NULL) FreeMem(global.jdictidx);
    global.jdictidx = p;

    p = (char * *) MemAlloc((global.nr_jdict + new + 1) * sizeof(char *) + 5);
    if (p == NULL) return (FALSE);

    memcpy(p, global.jdictdesc, global.nr_jdict * sizeof(char *));
    if (global.jdictdesc != NULL) FreeMem(global.jdictdesc);
    global.jdictdesc = p;

    p = (BOOL *) MemAlloc((global.nr_jdict + new + 1) * sizeof(BOOL) + 5);
    if (p == NULL) return (FALSE);

    memcpy(p, global.jdictactive, global.nr_jdict * sizeof(BOOL));
    global.jdictactive = p;

    return (TRUE);
}



BOOL FAR PASCAL DictionaryOptionsProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {

        case WM_INITDIALOG: {
            int i;

            /* Fill in the dictionaries */

            for (i = 0; i < global.nr_jdict; i++) {
                SendDlgItemMessage(hwnd, 4222, LB_ADDSTRING, 0, (LPARAM) i);
            }
            for (i = global.nr_jdict - 1; i >= 0; i--) {
                SendDlgItemMessage(hwnd, 4222, LB_SETSEL, global.jdictactive[i], (LPARAM) i);
            }

            /* Fill in the keywords */

            for (i = 0; i < global.nr_dictkeywords; i++) {
                SendDlgItemMessage(hwnd, 4221, LB_ADDSTRING, 0, (LPARAM) i);
            }

            for (i = global.nr_dictkeywords - 1; i >= 0; i--) {
                SendDlgItemMessage(hwnd, 4221, LB_SETSEL, !global.dictkeywordsshow[i], (LPARAM) i);
            }
            TrapHelpInDialogBoxes(hwnd, "DictionaryOptionsDlg", 1, TRUE);
            CenterDialogBox(hwnd);

            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND: {
            switch (WINCMDID) {
                case 4211: {    /* Add */
                    int i, j, len;

                    CurrentDict = CurrentIndex = CurrentDesc = NULL;
                    if (!DialogBox(hInstance, "EditDictionary", hwnd, EditDictionaryProc)) {
                        CurrentDict = CurrentIndex = CurrentDesc = NULL;
                        SetFocus(GetDlgItem(hwnd, 4222));
                        return (TRUE);
                    }

                    /* Put it in the list */

                    if (!ExtendJdict(1)) {
                        ErrorMessage(hwnd, "Cannot allocate memory to hold dictionary!");
                        if (CurrentDict != NULL) FreeMem(CurrentDict);
                        if (CurrentIndex != NULL) FreeMem(CurrentIndex);
                        if (CurrentDesc != NULL) FreeMem(CurrentDesc);
                        CurrentIndex = CurrentDict = CurrentDesc = NULL;
                        return (TRUE);
                    }

                    i = global.nr_jdict++;

                    global.jdict[i] = CurrentDict;
                    global.jdictidx[i] = CurrentIndex;
                    global.jdictdesc[i] = CurrentDesc;
                    global.jdictactive[i] = TRUE;

                    SendDlgItemMessage(hwnd, 4222, LB_ADDSTRING, 0, (LPARAM) i);

                    /* Which item is the one just added? */

                    len = SendDlgItemMessage(hwnd, 4222, LB_GETCOUNT, 0, 0L);

                    for (j = 0; j < len; j++) {
                        if (SendDlgItemMessage(hwnd, 4222, LB_GETITEMDATA, j, 0L) == i) break;
                    }

                    if (j < len) {
                        SendDlgItemMessage(hwnd, 4222, LB_SETCURSEL, j, 0L);
                        SendDlgItemMessage(hwnd, 4222, LB_SETSEL, TRUE, j);
                    }

                    SENDNOTIFY(hwnd, WM_COMMAND, 4222, LBN_SELCHANGE, hwnd);
                    OptionsChanged = TRUE;

                    CurrentIndex = CurrentDict = CurrentDesc = NULL;

                    SetFocus(GetDlgItem(hwnd, 4222));
                    return (TRUE);
                }

                case 4212: {    /* Edit */
                    int i;

                    i = SendDlgItemMessage(hwnd, 4222, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) {
                        MessageBeep(MB_ICONEXCLAMATION);
                        SetFocus(GetDlgItem(hwnd, 4222));
                        return (TRUE);
                    }

                    i = SendDlgItemMessage(hwnd, 4222, LB_GETITEMDATA, i, 0L);

                    CurrentDict = global.jdict[i];
                    CurrentIndex = global.jdictidx[i];
                    CurrentDesc = global.jdictdesc[i];

                    if (!DialogBox(hInstance, "EditDictionary", hwnd, EditDictionaryProc)) {
                        CurrentDict = CurrentIndex = CurrentDesc = NULL;
                        return (TRUE);
                    }

                    if (global.jdict[i] != NULL) FreeMem(global.jdict[i]);
                    if (global.jdictidx[i] != NULL) FreeMem(global.jdictidx[i]);
                    if (global.jdictdesc[i] != NULL) FreeMem(global.jdictdesc[i]);

                    global.jdict[i] = CurrentDict;
                    global.jdictidx[i] = CurrentIndex;
                    global.jdictdesc[i] = CurrentDesc;

                    CurrentDict = CurrentIndex = CurrentDesc = NULL;

                    OptionsChanged = TRUE;

                    SetFocus(GetDlgItem(hwnd, 4222));
                    return (TRUE);
				}

                case 4213: {    /* Delete */
                    int i, j;

                    if (global.nr_jdict <= 0) return (TRUE);
                    i = SendDlgItemMessage(hwnd, 4222, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) return (TRUE);

                    if (YesNo(hwnd, "Do you REALLY want to delete this dictionary from the list?") != IDYES)
                        return (TRUE);

                    j = SendDlgItemMessage(hwnd, 4222, LB_GETITEMDATA, i, 0L);
                    SendDlgItemMessage(hwnd, 4222, LB_DELETESTRING, i, 0L);

                    if (global.jdict[j] != NULL) FreeMem(global.jdict[j]);
                    if (global.jdictidx[j] != NULL) FreeMem(global.jdictidx[j]);
                    if (global.jdictdesc[j] != NULL) FreeMem(global.jdictdesc[j]);

                    global.jdict[j] = global.jdictidx[j] = global.jdictdesc[j] = NULL;
                    global.jdictactive[j] = FALSE;

                    if (SendDlgItemMessage(hwnd, 4222, LB_GETCOUNT, 0, 0L) > 0) {
                        SendDlgItemMessage(hwnd, 4222, LB_SETCURSEL, (i < 1) ? 0 : i-1, 0L);
                    }
                    SENDNOTIFY(hwnd, WM_COMMAND, 4222, LBN_SELCHANGE, hwnd);
                    OptionsChanged = TRUE;

                    return (TRUE);
                }

                case IDOK:      /* Done Button */
                case IDCANCEL: {
                    int i, j;
                    LONG r, n;

                    n = SendDlgItemMessage(hwnd, 4222, LB_GETCOUNT, 0, 0);

                    for (i = 0; i < n; i++) {
                        r = SendDlgItemMessage(hwnd, 4222, LB_GETSEL, i, 0L);
                        if (r != LB_ERR) {
                            if (global.jdictactive[i]) {
                                if (r != 0L) OptionsChanged = TRUE;
                            } else {
                                if (r == 0L) OptionsChanged = TRUE;
                            }
                            global.jdictactive[i] = (r != 0L);
                        }
                    }

                    /* Compact the entries */

                    for (i = 0; i < global.nr_jdict; i++) {
                        if (global.jdict[i] == NULL) {
                            for (j = i + 1; j < global.nr_jdict; j++) {
                                global.jdict[j-1] = global.jdict[j];
                                global.jdictidx[j-1] = global.jdictidx[j];
                                global.jdictdesc[j-1] = global.jdictdesc[j];
                                global.jdictactive[j-1] = global.jdictactive[j];
                            }

                            global.nr_jdict--;
                            i--;
                        }
                    }

                    for (i = 0; i < global.nr_dictkeywords; i++) {
                        r = SendDlgItemMessage(hwnd, 4221, LB_GETSEL, i, 0L);
                        if (r != LB_ERR) {
                            if (global.dictkeywordsshow[i]) {
                                if (r != 0L) OptionsChanged = TRUE;
                            } else {
                                if (r == 0L) OptionsChanged = TRUE;
                            }
                            global.dictkeywordsshow[i] = (r == 0L);
                        }
                    }
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
                }

                default:
                    break;
            }
            break;
        }

        case WM_KEYDOWN:
        case WM_CHAR:
            switch (wParam) {
                case VK_ESCAPE: SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                                return (TRUE);
            }
            break;

        case WM_COMPAREITEM:
        case WM_DELETEITEM:
        case WM_DRAWITEM:
        case WM_MEASUREITEM:
            return (JlistProc(hwnd, message, wParam, lParam, TRUE, ConvertDictOffsets));
    }

    return (FALSE);
}



BOOL CALLBACK EditDictionaryProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i, j;
            FILEOPTIONS *f;
            KANJI buffer[MAXLINELEN];

            /* Set the type and mode-change icon */
            SendDlgItemMessage(hwnd, 4202, EM_SETMODIFY, 0, FN_CONTROL);
            SendDlgItemMessage(hwnd, 4202, EM_SETRECT, 0, (LPARAM) GetDlgItem(hwnd, 4211));

            f = (FILEOPTIONS *) SendDlgItemMessage(hwnd, 4202, EM_GETHANDLE, 0, 0L);
            SendDlgItemMessage(hwnd, 4211, EM_SETHANDLE, 0, (LPARAM) f->hwnd);  /* mode-change icon */

            if (CurrentDict != NULL) {
                SendDlgItemMessage(hwnd, 4201, WM_SETTEXT, 0, (LPARAM) CurrentDict);
            }
            if (CurrentIndex != NULL) {
                SendDlgItemMessage(hwnd, 4203, WM_SETTEXT, 0, (LPARAM) CurrentIndex);
                SetFocus(GetDlgItem(hwnd, 4202));
            }

            if (CurrentDesc != NULL) {
                /* Edit */
                for (i = j = 0; CurrentDesc[i]; i++) {
                    if (CurrentDesc[i] & 0x80) {
                        buffer[j] = (CurrentDesc[i++] & 0x7f) << 8;
                        buffer[j++] |= (CurrentDesc[i] & 0x7f);
                    } else {
                        buffer[j++] = CurrentDesc[i];
                    }
                }
                buffer[j] = 0;
                SendDlgItemMessage(hwnd, 4202, EM_REPLACESEL, 0, (LPARAM) buffer);
                i = kanjilen(buffer);
                SendDlgItemMessage(hwnd, 4202, EM_SETSEL, i, MAKELONG(0, i-1));
                SetWindowText(GetDlgItem(hwnd, 1), "&Done!");
                SetWindowText(hwnd, "Edit an Existing Dictionary Item");
                TrapHelpInDialogBoxes(hwnd, "DictionaryEditDlg", 1, TRUE);
            } else {
                /* Add */
                SetWindowText(GetDlgItem(hwnd, 1), "&Add!");
                SetWindowText(hwnd, "Add a Dictionary Item");
                TrapHelpInDialogBoxes(hwnd, "DictionaryAddDlg", 1, TRUE);
            }

            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case 4221:
                case 4222: {    /* Find */
                    OPENFILENAME ofn;
                    char found[MAXFILENAMELEN];

                    switch (WINCMDID) {
                        case 4221: SendDlgItemMessage(hwnd, 4201, WM_GETTEXT, MAXFILENAMELEN, (LPARAM) found); break;
                        case 4222: SendDlgItemMessage(hwnd, 4203, WM_GETTEXT, MAXFILENAMELEN, (LPARAM) found); break;
                        default: return (TRUE);
                    }

                    memset(&ofn, 0, sizeof(OPENFILENAME));

                    ofn.lStructSize = sizeof(OPENFILENAME);
                    ofn.hwndOwner = hwnd;
                    ofn.hInstance = hInstance;
                    ofn.lpstrFilter = "All files\0*.*\0\0\0";
                    ofn.lpstrCustomFilter = NULL;
                    ofn.nMaxCustFilter = 0L;
                    ofn.nFilterIndex = 1;

                    ofn.lpstrFile= found;
                    ofn.nMaxFile = MAXFILENAMELEN;
                    ofn.lpstrFileTitle = NULL;
                    ofn.nMaxFileTitle = 0;
                    ofn.lpstrInitialDir = NULL;
                    ofn.lpstrTitle = "Find File...";
                    ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
                    ofn.nFileOffset = 0;
                    ofn.nFileExtension = 0;
                    ofn.lpstrDefExt = NULL;
                    ofn.lpfnHook = NULL;

                    ofn.lpTemplateName = "FileOpen";

                    if (GetOpenFileName(&ofn)) {
                        switch (WINCMDID) {
                            case 4221: SendDlgItemMessage(hwnd, 4201, WM_SETTEXT, 0, (LPARAM) found); break;
                            case 4222: SendDlgItemMessage(hwnd, 4203, WM_SETTEXT, 0, (LPARAM) found); break;
                            default: return (TRUE);
                        }
                    }
                    return (TRUE);
                }

                case IDCANCEL:
                    CurrentIndex = CurrentDict = CurrentDesc = NULL;

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);

                case IDOK: {
                    int i, j;
                    KANJI far *up;
                    OFSTRUCT of;
                    char buffer[MAXLINELEN];

                    SendDlgItemMessage(hwnd, 4201, WM_GETTEXT, MAXLINELEN, (LPARAM) buffer);
                    if (OpenFile(buffer, &of, OF_READ | OF_EXIST) == -1) {
                        ErrorMessage(hwnd, "The Dictionary File Name is not valid.");
						SetFocus(GetDlgItem(hwnd, 4201));
						return (TRUE);
					}
					CurrentDict = DupString(buffer);

					SendDlgItemMessage(hwnd, 4203, WM_GETTEXT, MAXLINELEN, (LPARAM) buffer);
                    if (OpenFile(buffer, &of, OF_READ | OF_EXIST) == -1) {
						if (CurrentDict != NULL) FreeMem(CurrentDict);
                        ErrorMessage(hwnd, "The Dictionary Index is not valid.");
						SetFocus(GetDlgItem(hwnd, 4203));
						return (TRUE);
                    }
                    CurrentIndex = DupString(buffer);

                    up = (KANJI far *) SendDlgItemMessage(hwnd, 4202, EM_GETLINE, 0, 0L);
                    if (kanjilen(up) <= 0) {
                        CurrentDesc = DupString(CurrentDict);
                    } else {
                        for (i = j = 0; up[i]; i++) {
                            if (ISKANJI(up[i])) {
                                buffer[j++] = HIBYTE(up[i]) | 0x80;
                                buffer[j++] = LOBYTE(up[i]) | 0x80;
                            } else {
                                buffer[j++] = up[i];
                            }
                        }
                        buffer[j] = '\0';
                        CurrentDesc = DupString(buffer);
                    }

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, TRUE);
                    return (TRUE);
                }
            }
            return (TRUE);

        case WM_PAINT: {
            HDC hdc;
            PAINTSTRUCT ps;

            hdc = BeginPaint(hwnd, &ps);

            DrawBoundingBox(hwnd, hdc, 4202);

            EndPaint(hwnd, &ps);
            return (TRUE);
        }
    }
    return (FALSE);
}



