/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */

#include "jwp.h"

#include "idm.h"


typedef struct {
    KANJI far *shortcut;
    KANJI far *phrase;
    BOOL dynamic;
    BOOL hidden;
} GLOSSARY;


#define GBLOCKSIZE  10

static GLOSSARY *Glossary = NULL;
static int MaxGlossary = 0;
static int NrGlossary = 0;
static BOOL GlossaryChanged = FALSE;
static BOOL sizing = FALSE;

static KANJI far *CurrentShortCut;
static KANJI far *CurrentPhrase;



static KANJI far *ConvertGlossary (int id, LONG lParam, KANJI *buf)
{
    return (Glossary[lParam].shortcut);
}



static KANJI far *ConvertGlossary1 (int id, LONG lParam, KANJI *buf)
{
    return (Glossary[lParam].phrase);
}



static int GlossaryComp (void const *v1, void const *v2)
{
    return (kanjicmp(((GLOSSARY *) v1)->shortcut, ((GLOSSARY *) v2)->shortcut));
}



void ReadGlossary (char *filename)
{
    int i, j, len;
    HFILE fd;
    OFSTRUCT of;

    /* Clear the old glossary */

    if (Glossary != NULL) {
        for (i = 0; i < NrGlossary; i++) {
            if (Glossary[i].shortcut != NULL) FreeBlock(Glossary[i].shortcut);
            if (Glossary[i].phrase != NULL) FreeBlock(Glossary[i].phrase);
        }
        FreeMem(Glossary);
        Glossary = NULL;
    }

    fd = OpenFile(filename, &of, OF_READ);
    if (fd == HFILE_ERROR) return;

    /* Read the glossary file */

    _llseek(fd, 0L, 0);
    ReadShortInt(fd, &NrGlossary);

    if (NrGlossary <= 0) {
        _lclose(fd);
        return;
    }

    Glossary = (GLOSSARY *) MemAlloc((NrGlossary + 1) * sizeof(GLOSSARY));
    if (Glossary == NULL) {
        ErrorMessage(global.hwnd, "Cannot allocate memory to hold glossary!");
        _lclose(fd);
        return;
    }

    MaxGlossary = NrGlossary;

    for (i = 0; i < NrGlossary; i++) {
        /* Read the options */
        ReadShortInt(fd, &j);

        Glossary[i].dynamic = j & 0x0001;
        Glossary[i].hidden = j & 0x0002;

        /* Read the short-cut */
        ReadShortInt(fd, &len);
        len *= sizeof(KANJI);
        Glossary[i].shortcut = (KANJI far *) BlockAlloc(len);
        if (Glossary[i].shortcut == NULL) break;
        _lread(fd, (char far *) Glossary[i].shortcut, len);

        /* Read the phrase */
        ReadShortInt(fd, &len);
        len *= sizeof(KANJI);
        Glossary[i].phrase = (KANJI far *) BlockAlloc(len);
        if (Glossary[i].shortcut == NULL) break;
        _lread(fd, (char far *) Glossary[i].phrase, len);
    }

    _lclose(fd);

    if (NrGlossary > 0) {
        qsort(Glossary, NrGlossary, sizeof(GLOSSARY), GlossaryComp);
    }
}



void WriteGlossary (char *filename)
{
    int i, j, len;
    HFILE fd;
    OFSTRUCT of;


    if (!GlossaryChanged) return;

    fd = OpenFile(filename, &of, OF_WRITE | OF_CREATE);
    if (fd == HFILE_ERROR) {
        ErrorMessage(global.hwnd, "Cannot create glossary file %s!", filename);
        return;
    }

    for (i = j = 0; i < NrGlossary; i++) {
        if (Glossary[i].shortcut != NULL) j++;
    }

    _llseek(fd, 0L, 0);
    WriteShortInt(fd, j);

    for (i = 0; i < NrGlossary; i++) {
        if (Glossary[i].shortcut == NULL) continue;

        /* Write the options flag */
        j = 0;
        if (Glossary[i].dynamic) j |= 0x0001;
        if (Glossary[i].hidden) j |= 0x0002;
        WriteShortInt(fd, j);

        /* Write the short-cut */
        len = kanjilen(Glossary[i].shortcut) + 1;
        WriteShortInt(fd, len);

        len *= sizeof(KANJI);
        _lwrite(fd, (char far *) Glossary[i].shortcut, len);

        /* Write the phrase */
        len = kanjilen(Glossary[i].phrase) + 1;
        WriteShortInt(fd, len);

        len *= sizeof(KANJI);
        _lwrite(fd, (char far *) Glossary[i].phrase, len);
    }

    _lclose(fd);
}



int SearchGlossary (KANJI far *key)
{
    int len;
    int top, bottom, middle;
    int diff;


    if (NrGlossary <= 0) return (-2);


    /* Now binary search */

    top = 0;
    bottom = NrGlossary - 1;

    len = kanjilen(key);

    if (!kanjicmp(key, Glossary[top].shortcut)) {
        if (Glossary[top].dynamic) return (top);
    } else if (!kanjincmp(key, Glossary[top].shortcut, len)) {
        if (Glossary[top].dynamic) return (-1);
    }

    if (!kanjicmp(key, Glossary[bottom].shortcut)) {
        return (Glossary[bottom].dynamic ? bottom : -2);
    } else if (!kanjincmp(key, Glossary[bottom].shortcut, len)) {
        return (Glossary[bottom].dynamic ? -1 : -2);
    }


    for (;;) {
        middle = (top + bottom) / 2;

        diff = kanjicmp(key, Glossary[middle].shortcut);

        if (diff == 0) {
            return (Glossary[middle].dynamic ? middle : -2);
        }

        if (top >= bottom - 1) {
            for (; bottom < NrGlossary; bottom++) {
                diff = kanjincmp(key, Glossary[bottom].shortcut, len);
                if (!Glossary[bottom].dynamic || diff != 0) continue;
                return (-1);
            }
            return (-2);
        }

        if (diff > 0) top = middle;
        else bottom = middle;
    }
}



KANJI far *GetGlossary (int index)
{
    return (Glossary[index].phrase);
}




BOOL CALLBACK EditGlossaryProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i;
            FILEOPTIONS *f;

            /* Set the type and mode-change icon */
            SendDlgItemMessage(hwnd, 4201, EM_SETMODIFY, 0, FN_CONTROL | FN_NOKANJI);
            SendDlgItemMessage(hwnd, 4201, EM_SETRECT, 0, (LONG) GetDlgItem(hwnd, 4211));

            SendDlgItemMessage(hwnd, 4202, EM_SETMODIFY, 0, FN_CONTROL);
            SendDlgItemMessage(hwnd, 4202, EM_SETRECT, 0, (LONG) GetDlgItem(hwnd, 4211));

            f = (FILEOPTIONS *) SendDlgItemMessage(hwnd, 4201, EM_GETHANDLE, 0, 0L);
            SendDlgItemMessage(hwnd, 4211, EM_SETHANDLE, 0, (LPARAM) f->hwnd);  /* mode-change icon */

            if (CurrentPhrase != NULL) {
                /* Edit */
                SendDlgItemMessage(hwnd, 4201, EM_REPLACESEL, 0, (LPARAM) CurrentShortCut);
                SendDlgItemMessage(hwnd, 4202, EM_REPLACESEL, 0, (LPARAM) CurrentPhrase);
                i = kanjilen(CurrentPhrase);
                SendDlgItemMessage(hwnd, 4202, EM_SETSEL, i, MAKELONG(0, i-1));
                SetWindowText(GetDlgItem(hwnd, 1), "&Done!");
                SetWindowText(hwnd, "Edit an Existing Glossary Item");
                TrapHelpInDialogBoxes(hwnd, "GlossaryEditDlg", 1, TRUE);

                EnableWindow(GetDlgItem(hwnd, 4201), FALSE);
                SetFocus(GetDlgItem(hwnd, 4202));
            } else {
                /* Add */
                SetWindowText(GetDlgItem(hwnd, 1), "&Add!");
                SetWindowText(hwnd, "Add a Glossary Item");
                TrapHelpInDialogBoxes(hwnd, "GlossaryAddDlg", 1, TRUE);
            }

            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDCANCEL:
                    CurrentShortCut = CurrentPhrase = NULL;

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);

                case IDOK: {
                    KANJI far *up, far *up1;

                    up = (KANJI far *) SendDlgItemMessage(hwnd, 4201, EM_GETLINE, 0, 0L);
                    if (kanjilen(up) <= 0) {
                        ErrorMessage(hwnd, "Sorry, you must enter a short-cut for the phrase.");
                        SetFocus(GetDlgItem(hwnd, 4201));
                        return (TRUE);
                    }

                    up1 = (KANJI far *) SendDlgItemMessage(hwnd, 4202, EM_GETLINE, 0, 0L);
                    if (kanjilen(up1) <= 0) {
                        ErrorMessage(hwnd, "Sorry, you must enter a phrase for the short-cut.");
                        SetFocus(GetDlgItem(hwnd, 4202));
                        return (TRUE);
                    }

                    if (CurrentPhrase == NULL) {
                        CurrentShortCut = (KANJI far *) BlockAlloc((kanjilen(up) + 5) * sizeof(KANJI));
                        kanjicpy(CurrentShortCut, up);
                    } else {
                        CurrentShortCut = NULL;
                    }

                    CurrentPhrase = (KANJI far *) BlockAlloc((kanjilen(up1) + 5) * sizeof(KANJI));
                    kanjicpy(CurrentPhrase, up1);

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, TRUE);
                    return (TRUE);
                }
            }
            return (TRUE);

        case WM_PAINT: {
            HDC hdc;
            PAINTSTRUCT ps;

            hdc = BeginPaint(hwnd, &ps);

            DrawBoundingBox(hwnd, hdc, 4201);
            DrawBoundingBox(hwnd, hdc, 4202);

            EndPaint(hwnd, &ps);
            return (TRUE);
        }
    }
    return (FALSE);
}



BOOL CALLBACK GlossaryProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i;

            for (i = 0; i < NrGlossary; i++) {
                SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LPARAM) i);
            }

            if (NrGlossary > 0) {
                SendDlgItemMessage(hwnd, 4211, LB_SETCURSEL, 0, 0L);
                i = SendDlgItemMessage(hwnd, 4211, LB_GETITEMDATA, 0, 0L);
                SendDlgItemMessage(hwnd, 4201, EM_REPLACESEL, 0, (LPARAM) Glossary[i].shortcut);
                SendDlgItemMessage(hwnd, 4202, EM_REPLACESEL, 0, (LPARAM) Glossary[i].phrase);
                CheckDlgButton (hwnd, 4231, Glossary[0].dynamic);
                CheckDlgButton (hwnd, 4232, Glossary[0].hidden);
            } else {
                EnableWindow(GetDlgItem(hwnd, 4231), FALSE);
                EnableWindow(GetDlgItem(hwnd, 4232), FALSE);
            }

            SetFocus(GetDlgItem(hwnd, 4211));

            TrapHelpInDialogBoxes(hwnd, "GlossaryDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

        case WM_KEYDOWN:
            switch (wParam) {
                case VK_ESCAPE: SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                                return (TRUE);
            }
            break;

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND: {
            int i, j;

            switch (WINCMDID) {
                case 4211:      /* List Box */
                    switch (NOTIFYCODE) {
                        case LBN_SELCHANGE:
                            i = SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);

                            if (NrGlossary <= 0 || i == LB_ERR) {
NoGlossary:
                                SendDlgItemMessage(hwnd, 4201, EM_REPLACESEL, 0, (LPARAM) NULL);
                                SendDlgItemMessage(hwnd, 4202, EM_REPLACESEL, 0, (LPARAM) NULL);
                                InvalidateRect(GetDlgItem(hwnd, 4201), NULL, TRUE);
                                InvalidateRect(GetDlgItem(hwnd, 4202), NULL, TRUE);
                                CheckDlgButton (hwnd, 4231, FALSE);
                                CheckDlgButton (hwnd, 4232, FALSE);
                                EnableWindow(GetDlgItem(hwnd, 4231), FALSE);
                                EnableWindow(GetDlgItem(hwnd, 4232), FALSE);
                                return (TRUE);
                            }

                            i = SendDlgItemMessage(hwnd, 4211, LB_GETITEMDATA, i, 0L);
                            if (i >= NrGlossary || Glossary[i].shortcut == NULL) goto NoGlossary;

                            SendDlgItemMessage(hwnd, 4201, EM_REPLACESEL, 0, (LPARAM) Glossary[i].shortcut);
                            SendDlgItemMessage(hwnd, 4202, EM_REPLACESEL, 0, (LPARAM) Glossary[i].phrase);
                            InvalidateRect(GetDlgItem(hwnd, 4201), NULL, TRUE);
                            InvalidateRect(GetDlgItem(hwnd, 4202), NULL, TRUE);
                            CheckDlgButton (hwnd, 4231, Glossary[i].dynamic);
                            CheckDlgButton (hwnd, 4232, Glossary[i].hidden);
                            EnableWindow(GetDlgItem(hwnd, 4231), TRUE);
                            EnableWindow(GetDlgItem(hwnd, 4232), TRUE);
                            return (TRUE);

                        case LBN_DBLCLK:
                            SendMessage(hwnd, WM_COMMAND, 4222, 0L);    /* Edit */
                            return (TRUE);
                    }
                    return (TRUE);

                case 4221: {    /* Add */
                    int len;

                    CurrentShortCut = CurrentPhrase = NULL;
                    if (!DialogBox(hInstance, "EditGlossary", hwnd, EditGlossaryProc)) {
                        CurrentShortCut = CurrentPhrase = NULL;
                        return (TRUE);
                    }

                    if (CurrentShortCut == NULL || CurrentPhrase == NULL) {
                        CurrentShortCut = CurrentPhrase = NULL;
                        return (TRUE);
                    }


                    /* Trim trailing blanks */

                    len = kanjilen(CurrentShortCut);

                    for (i = len - 1; i >= 0; i--) {
                        CurrentShortCut[i] &= 0x7f7f;
                        if (CurrentShortCut[i] != 0x2121 && CurrentShortCut[i] > ' ') break;
                    }

                    CurrentShortCut[i+1] = 0;

                    /* Duplicate? */

                    for (i = 0; i < NrGlossary; i++) {
                        if (Glossary[i].shortcut == NULL) continue;
                        if (!kanjicmp(CurrentShortCut, Glossary[i].shortcut)) break;
                    }

                    if (i < NrGlossary) {
                        if (YesNo(hwnd, "The short-cut you want to add already exists!  "
                                        "Do you REALLY want to overwrite it?") != IDYES) {

                            if (CurrentShortCut != NULL) FreeBlock(CurrentShortCut);
                            if (CurrentPhrase != NULL) FreeBlock(CurrentPhrase);
                            CurrentShortCut = CurrentPhrase = NULL;
                            return (TRUE);
                        }

                        FreeBlock((void far *) CurrentShortCut);
                        /* Replace the phrase */
                        FreeBlock((void far *) Glossary[i].phrase);
                        Glossary[i].phrase = CurrentPhrase;
                        InvalidateRect(GetDlgItem(hwnd, 4221), NULL, TRUE);
                    } else {
                        /* Put it in the glossary */

                        if (Glossary == NULL) {
                            Glossary = (GLOSSARY *) MemAlloc((GBLOCKSIZE + 1) * sizeof(GLOSSARY));
                            if (Glossary == NULL) {
                                ErrorMessage(global.hwnd, "Cannot allocate memory to hold glossary!");
                                if (CurrentShortCut != NULL) FreeBlock(CurrentShortCut);
                                if (CurrentPhrase != NULL) FreeBlock(CurrentPhrase);
                                CurrentPhrase = CurrentShortCut = NULL;
                                SetFocus(GetDlgItem(hwnd, 4211));
                                return (TRUE);
                            }
                            MaxGlossary = GBLOCKSIZE;
                        } else if (NrGlossary >= MaxGlossary) {
                            GLOSSARY *tmp;

                            tmp = (GLOSSARY *) MemAlloc((MaxGlossary + GBLOCKSIZE) * sizeof(GLOSSARY));
                            if (tmp == NULL) {
                                ErrorMessage(global.hwnd, "Cannot allocate memory to hold glossary!");
                                if (CurrentShortCut != NULL) FreeBlock(CurrentShortCut);
                                if (CurrentPhrase != NULL) FreeBlock(CurrentPhrase);
                                CurrentPhrase = CurrentShortCut = NULL;
                                SetFocus(GetDlgItem(hwnd, 4211));
                                return (TRUE);
                            }
                            memcpy(tmp, Glossary, NrGlossary * sizeof(GLOSSARY));
                            FreeMem(Glossary);
                            Glossary = tmp;
                            MaxGlossary += GBLOCKSIZE;
                        }

                        Glossary[NrGlossary].shortcut = CurrentShortCut;
                        Glossary[NrGlossary].phrase = CurrentPhrase;

                        Glossary[NrGlossary].dynamic = TRUE;
                        Glossary[NrGlossary].hidden = FALSE;

                        SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LPARAM) NrGlossary);

                        i = NrGlossary++;
                    }

                    /* Which item is the one just added? */

                    len = SendDlgItemMessage(hwnd, 4211, LB_GETCOUNT, 0, 0L);

                    for (j = 0; j < len; j++) {
                        if (SendDlgItemMessage(hwnd, 4211, LB_GETITEMDATA, j, 0L) == i) break;
                    }

                    if (j < len) SendDlgItemMessage(hwnd, 4211, LB_SETCURSEL, j, 0L);

                    SENDNOTIFY(hwnd, WM_COMMAND, 4211, LBN_SELCHANGE, hwnd);

                    GlossaryChanged = TRUE;

                    CurrentPhrase = CurrentShortCut = NULL;

                    SetFocus(GetDlgItem(hwnd, 4211));
                    return (TRUE);
                }

                case 4222:     /* Edit */
                    i = SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) {
                        MessageBeep(MB_ICONEXCLAMATION);
                        return (TRUE);
                    }

                    i = SendDlgItemMessage(hwnd, 4211, LB_GETITEMDATA, i, 0L);

                    CurrentShortCut = Glossary[i].shortcut;
                    CurrentPhrase = Glossary[i].phrase;

                    if (!DialogBox(hInstance, "EditGlossary", hwnd, EditGlossaryProc)) {
                        CurrentShortCut = CurrentPhrase = NULL;
                        return (TRUE);
                    }

                    if (CurrentPhrase != NULL) {
                        FreeBlock(Glossary[i].phrase);
                        Glossary[i].phrase = CurrentPhrase;
                    }

                    CurrentShortCut = CurrentPhrase = NULL;

                    SENDNOTIFY(hwnd, WM_COMMAND, 4211, LBN_SELCHANGE, hwnd);

                    GlossaryChanged = TRUE;

                    return (TRUE);

                case 4223:      /* Delete */
                    if (NrGlossary <= 0) return (TRUE);
                    i = SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) return (TRUE);

                    if (YesNo(hwnd, "Do you REALLY want to delete this item?") != IDYES)
                        return (TRUE);

                    j = SendDlgItemMessage(hwnd, 4211, LB_GETITEMDATA, i, 0L);
                    SendDlgItemMessage(hwnd, 4211, LB_DELETESTRING, i, 0L);

                    if (Glossary[j].shortcut != NULL) FreeBlock(Glossary[j].shortcut);
                    if (Glossary[j].phrase != NULL) FreeBlock(Glossary[j].phrase);

                    Glossary[j].shortcut = Glossary[j].phrase = NULL;

                    if (SendDlgItemMessage(hwnd, 4211, LB_GETCOUNT, 0, 0L) > 0) {
                        SendDlgItemMessage(hwnd, 4211, LB_SETCURSEL, (i < 1) ? 0 : i-1, 0L);
                    }
                    SENDNOTIFY(hwnd, WM_COMMAND, 4211, LBN_SELCHANGE, hwnd);
                    GlossaryChanged = TRUE;

                    SetFocus(GetDlgItem(hwnd, 4211));

                    return (TRUE);

                case 4231:      /* Dynamic button */
                    i = SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) return (TRUE);

                    j = SendDlgItemMessage(hwnd, 4211, LB_GETITEMDATA, i, 0L);
                    Glossary[j].dynamic = IsDlgButtonChecked(hwnd, 4231);
                    GlossaryChanged = TRUE;
                    return (TRUE);

                case 4232:      /* Hidden button */
                    i = SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) return (TRUE);

                    j = SendDlgItemMessage(hwnd, 4211, LB_GETITEMDATA, i, 0L);
                    Glossary[j].hidden = IsDlgButtonChecked(hwnd, 4232);
                    GlossaryChanged = TRUE;
                    return (TRUE);

                case IDOK:
                    /* Compact the entries */

                    for (i = 0; i < NrGlossary; i++) {
                        if (Glossary[i].shortcut == NULL) {
                            for (j = i + 1; j < NrGlossary; j++)
                                Glossary[j-1] = Glossary[j];

                            NrGlossary--;
                            i--;
                        }
                    }

                    if (NrGlossary > 0) {
                        qsort(Glossary, NrGlossary, sizeof(GLOSSARY), GlossaryComp);
                    }

                    if (global.glisthwnd != NULL)
                        SendMessage(global.glisthwnd, EM_SETRECT, 0, 0L);

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }

            return (FALSE);
        }

        case WM_COMPAREITEM:
        case WM_DELETEITEM:
        case WM_DRAWITEM:
        case WM_MEASUREITEM:
            return (JlistProc(hwnd, message, wParam, lParam, TRUE, ConvertGlossary));
    }

    return (FALSE);
}



BOOL CALLBACK GlossaryListProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int i, j;

    switch (message) {
        case WM_INITDIALOG: {
            RECT rect;

            sizing = FALSE;

            GetWindowRect(global.hwnd, &rect);

            SetWindowPos(hwnd, NULL,
                         rect.left + global.glistsize.left,
                         rect.top + global.glistsize.top,
                         0, 0, SWP_NOSIZE | SWP_NOZORDER);

            SetWindowPos(hwnd, NULL, 0, 0, global.glistsize.right, global.glistsize.bottom, SWP_NOMOVE | SWP_NOZORDER);

            SendMessage(hwnd, WM_USER, 0, 0L);

            return (TRUE);
        }

        case WM_USER:
            SendDlgItemMessage(hwnd, 4201, LB_RESETCONTENT, 0, 0L);
            for (i = 0; i < NrGlossary; i++) {
                if (Glossary[i].hidden) continue;
                SendDlgItemMessage(hwnd, 4201, LB_ADDSTRING, 0, (LPARAM) i);
            }
            return (TRUE);

        case WM_MOVE: {
            RECT rect1, rect2;

            GetWindowRect(global.hwnd, &rect1);
            GetWindowRect(hwnd, &rect2);

            global.glistsize.left = rect2.left - rect1.left;
            global.glistsize.top = rect2.top - rect1.top;
            return (TRUE);
        }

        case WM_SIZE: {
            RECT rect1, rect2, rect3;

            if (sizing) return (TRUE);

            GetWindowRect(hwnd, &rect1);
            GetClientRect(hwnd, &rect2);

            MoveWindow(GetDlgItem(hwnd, 4201), 0, 0, rect2.right, rect2.bottom, TRUE);
            GetWindowRect(GetDlgItem(hwnd, 4201), &rect3);

            rect3.right += (rect1.right - rect1.left - rect2.right);
            rect3.bottom += (rect1.bottom - rect1.top - rect2.bottom);

            sizing = TRUE;

            global.glistsize.right = i = rect3.right - rect3.left;
            global.glistsize.bottom = j = rect3.bottom - rect3.top;

            SetWindowPos(hwnd, NULL, 0, 0, i, j, SWP_NOMOVE | SWP_NOZORDER);
            sizing = FALSE;

            return (TRUE);
        }

        case WM_SYSCOMMAND:
            if ((wParam & 0xfff0) == SC_CLOSE) {
                SetActiveWindow(global.hwnd);

                EndDialog(hwnd, FALSE);
                SendMessage(global.tbhwnd, BM_SETSTATE, 0, ID_GLOSSARY);
                CheckMenuItem(hmenu, IDM_VIEWGLOSSARY, MF_UNCHECKED | MF_BYCOMMAND);
                global.glisthwnd = NULL;
                return (TRUE);
            }
            break;

        case WM_COMMAND:
            switch (WINCMDID) {
                case 4201:
                    switch (NOTIFYCODE) {
                        case LBN_SELCHANGE: {
                            FILEOPTIONS *f;

                            i = SendDlgItemMessage(hwnd, 4201, LB_GETCURSEL, 0, 0L);
                            if (i == LB_ERR) return (TRUE);
                            if (global.active == NULL) {
                                MessageBeep(MB_ICONHAND);
                                SendDlgItemMessage(hwnd, 4201, LB_SETCURSEL, LB_ERR, 0L);
                                return (TRUE);
                            }

                            i = SendDlgItemMessage(hwnd, 4201, LB_GETITEMDATA, i, 0L);

                            f = global.active;

                            if (!CheckInputTextType(f, PARAOF(f->current)->text, Glossary[i].phrase)) {
                                MessageBeep(MB_ICONHAND);
                                SendDlgItemMessage(hwnd, 4201, LB_SETCURSEL, LB_ERR, 0L);
                                return (TRUE);
                            }

                            TurnOffSelection(f);
                            UndoAddTyping(f, f->current, kanjilen(Glossary[i].phrase), FALSE);
                            InsertString(f, f->current, Glossary[i].phrase,
                                OP_REFORMAT | OP_UPDATE | OP_MOVESEL | OP_MOVETOEND);

                            SendDlgItemMessage(hwnd, 4201, LB_SETCURSEL, LB_ERR, 0L);
                            SetActiveWindow(global.hwnd);

                            switch (global.glistbehaviour) {
                                case 0: break;
                                case 1: EndDialog(hwnd, FALSE);
                                        CheckMenuItem(hmenu, IDM_VIEWGLOSSARY, MF_UNCHECKED | MF_BYCOMMAND);
                                        global.glisthwnd = NULL;
                                        SendMessage(global.tbhwnd, BM_SETSTATE, 0, ID_GLOSSARY);
                                        break;
                            }
                            return (TRUE);
                        }
                    }
                    return (TRUE);
            }
            return (FALSE);

        case WM_KEYDOWN:
        case WM_CHAR:
            switch (wParam) {
                case VK_ESCAPE: SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                                return (TRUE);
            }
            break;

        case WM_COMPAREITEM:
        case WM_DELETEITEM:
        case WM_DRAWITEM:
        case WM_MEASUREITEM:
            return (JlistProc(hwnd, message, wParam, lParam, TRUE, ConvertGlossary1));
    }

    return (FALSE);
}
