/* Copyright (C) Stephen Chung, 1995. All rights reserved. */

#include "jwp.h"

#include <stdarg.h>
#include <shellapi.h>
#include <commdlg.h>
#include <sys/stat.h>

static int FileFormatIcons[] = { -1, 2, 3, 4, 5, 6, 7, 8 };

char *FixedDictKeywords[][2] = {
    { "(m)",    "not starting with the search text" },

    { "pn",     "person names (family or given)" },
    { "pl",     "place names" },
    { "giv",    "given names" },
    { "fem",    "female given names" },
    { "male",   "male given names" },

    { "vulg",   "vulgar expressions or words" },
    { "X",      "rude or X-rated terms" },

    { "col",    "colloquialism" },
    { "m-sl",   "manga slangs" },
    { "MA",     "martial arts terms" },
    { "id",     "idiomatic expressions" },
    { "arch",   "archaism" },
    { "fam",    "familiar language" },
    { "pol",    "polite (ǫ) language" },
    { "hum",    "humble () language" },
    { "hon",    "honorific or respectful (ºɸ) language" },

    { "an",     "adjectival nouns or quasi-adjectives (ư)" },
    { "a-no",   "nouns which may take the genitive case particle " },
    { "vs",     "nouns or participles which take the auxillary verb " },
    { "vt",     "transitive verbs" },
    { "vi",     "intransitive verbs" },
    { "I",      "Type I () verbs (only when type is not implicit)" },
    { "IV",     "Type IV (irregular) verbs, such as " },

    { "pref",   "prefix" },
    { "suf",    "suffix" },

    { "uk",     "words usually written using kana alone" },
    { "uK",     "words usually written using kanji alone" },
    { "oK",     "words containing out-dated kanji" },
    { "ik",     "words containing irregular kana usage" },
    { "iK",     "words containing irregular kanji usage" },
    { "io",     "words containing irregular okurigana usage" },

    { NULL,     NULL }
};



static void RegisterExtensions (void)
{
    FILEFORMAT i;
    LONG size;
    char cmdline[MAXLINELEN];
    char doctype[MAXLINELEN];
    char buffer[MAXLINELEN];
    char value[MAXLINELEN];

    if (GetModuleFileName(hInstance, cmdline, MAXLINELEN) <= 0) return;

    for (i = FF_NORMAL; FileFormatNames[i] != NULL; i++) {
        sprintf(doctype, "%s%s.file", PROGNAME, FileExtensions[i]);

        if (RegQueryValue(HKEY_CLASSES_ROOT, FileExtensions[i], value, &size) == ERROR_SUCCESS) {
            if (stricmp(doctype, value)) {
                if (YesNo(global.hwnd, "The file extension \'%s\' is already linked to another program. "
                                       "Do you want " PROGNAME " " JWPVERSION " to change \'%s\' to a "
                                       "\'%s\' file instead?\n\n"
                                       "If you say yes, then " PROGNAME " " JWPVERSION " will be used to "
                                       "open files with this extension in the future.",
                                       FileExtensions[i], FileExtensions[i], FileFormatNames[i]) != IDYES)
                    continue;
            }
        }

        RegSetValue(HKEY_CLASSES_ROOT, FileExtensions[i], REG_SZ, doctype, strlen(doctype));

        switch (i) {
            case FF_NORMAL:     strcpy(value, "Japanese Text File (" PROGNAME " Normal Format)"); break;
            case FF_TEMPLATE:   strcpy(value, "Japanese Text File (" PROGNAME " Template Format)"); break;
            default:            sprintf(value, "Japanese Text File (%s Format)", FileFormatNames[i]); break;
        }

        RegSetValue(HKEY_CLASSES_ROOT, doctype, REG_SZ, value, strlen(value));

        sprintf(buffer, "%s\\DefaultIcon", doctype);
        sprintf(value, "%s,%d", cmdline, FileFormatIcons[i]);
        RegSetValue(HKEY_CLASSES_ROOT, buffer, REG_SZ, value, strlen(value));

        sprintf(buffer, "%s\\Shell\\open\\command", doctype);

#ifdef USEWIN32
        sprintf(value, "%s \"%%1\"", cmdline);
#else USEWIN32
        strcpy(value, cmdline);
#endif USEWIN32

        RegSetValue(HKEY_CLASSES_ROOT, buffer, REG_SZ, value, strlen(value));
    }
}



void SaveKeyAndINI(FILE *fp, char *section, char *option, char *fmt, ...)
{
    va_list argptr;
    char value[MAXLINELEN];
#ifdef USEWIN32
    DWORD size;
    HKEY subkey;
    char buffer[MAXLINELEN];
    char oldvalue[MAXLINELEN];
#endif USEWIN32

    va_start(argptr, fmt);
    vsprintf(value, fmt, argptr);
    va_end(argptr);

    if (fp != NULL) fprintf(fp, "%s=%s\n", option, value);

#ifdef USEWIN32
    sprintf(buffer, "Software\\" PROGNAME "\\Options\\%s", section);
    if (RegCreateKey(HKEY_CURRENT_USER, buffer, &subkey) != ERROR_SUCCESS) return;

    /* Won't need to duplicate value writes */

    size = MAXLINELEN;
    if (RegQueryValueEx(subkey, option, NULL, NULL, oldvalue, &size) == ERROR_SUCCESS) {
        if (!strcmp(oldvalue, value)) {
            RegCloseKey(subkey);
            return;
        }
    }

    RegSetValueEx(subkey, option, 0, REG_SZ, value, strlen(value) + 1);
    RegCloseKey(subkey);
#endif USEWIN32
}



static BOOL LocateFile(FILE *fp, char *section, char *option, char *filename, char *desc, char *MainDirectory)
{
    int i;
    OFSTRUCT of;
    OPENFILENAME ofn;
    BOOL update;
    char buffer[MAXLINELEN];
    char found[MAXFILENAMELEN];


    update = (fp != NULL || section != NULL);


    /* Is the file there in the JWP directory? */

    if (desc == NULL) desc = option;

    sprintf(found, "%s\\%s", MainDirectory, filename);
    if (OpenFile(found, &of, OF_READ | OF_EXIST) != -1) {
        if (update)
            SaveKeyAndINI(fp, section, option, "%s", found);
        else
            strcpy(filename, of.szPathName);
        return (TRUE);
    }

    ErrorMessage (global.hwnd, "Cannot find the %s [%s]!", desc, filename);


    /* Fill in the common dialog box structure */

    memset(&ofn, 0, sizeof(OPENFILENAME));
    strcpy(found, filename);

    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = global.hwnd;
    ofn.hInstance = hInstance;
    ofn.lpstrFilter = "All files\0*.*\0\0\0";
    ofn.lpstrCustomFilter = NULL;
    ofn.nMaxCustFilter = 0L;
    ofn.nFilterIndex = 1;

    ofn.lpstrFile= found;
    ofn.nMaxFile = MAXFILENAMELEN;
    ofn.lpstrFileTitle = NULL;
    ofn.nMaxFileTitle = 0;
    ofn.lpstrInitialDir = MainDirectory;

    sprintf(buffer, "Looking for the %s [%s]...", desc, filename);

    ofn.lpstrTitle = buffer;
    ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
    ofn.nFileOffset = 0;
    ofn.nFileExtension = 0;
    ofn.lpstrDefExt = NULL;
    ofn.lpfnHook = NULL;

    ofn.lpTemplateName = "FileOpen";


    /* Open the file using a common dialog box */

    if (!GetOpenFileName(&ofn)) {
        if (IDYES != YesNo (global.hwnd, "Cannot find the %s [%s].\n\n"
                                         PROGNAME " " JWPVERSION " may not run properly.  "
                                         "Do you want to continue anyway?",
                                         desc, filename)) {
            if (fp != NULL) fclose(fp);
            return (FALSE);
        } else {
            if (update) SaveKeyAndINI(fp, section, option, "%s\\%s", MainDirectory, filename);
            return (TRUE);
        }
    }

    i = strlen(found);
    if (found[i-1] == '\\') found[i-1] = '\0';

    if (update)
        SaveKeyAndINI(fp, section, option, "%s", found);
    else
        strcpy(filename, found);

    return (TRUE);
}



BOOL CreateOptionsFile(char *filename)
{
    int i, j;
    DWORD size;
    OFSTRUCT of;
    HKEY hsubkey;
    char MainDirectory[MAXFILENAMELEN];
    char WindowsDirectory[MAXFILENAMELEN];
    char buffer[MAXLINELEN];
    char line[MAXLINELEN];
    time_t t;
    FILE *fp;
    struct stat sb;


#ifdef USEWIN32

    fp = NULL;

    if (filename != NULL) filename[0] == '\0';

    if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\" PROGNAME "\\Program", 0, KEY_QUERY_VALUE, &hsubkey) == ERROR_SUCCESS) {
        size = MAXLINELEN;

        if (RegQueryValueEx(hsubkey, "Version", NULL, NULL, buffer, &size) == ERROR_SUCCESS) {
            RegCloseKey(hsubkey);

            if (!buffer[0] || stricmp(buffer, JWPVERSION)) {
                if (IDYES != YesNo (global.hwnd, "The existing options in the registry is for %s version of " PROGNAME
                                                 " - version %s.\n\n"
                                                 "Do you want " PROGNAME " " JWPVERSION " to overwrite it with its own?\n\n"
                                                 "(If you choose Yes, all settings will revert to " PROGNAME " " JWPVERSION "'s default.)",
                                                 (stricmp(buffer, JWPVERSION) < 0) ? "an older" : "a newer", buffer)) {

                    return (TRUE);
                }
            } else {
                /* Good set of options */
                return (TRUE);
            }
        } else {
            RegCloseKey(hsubkey);
        }

        RegisterExtensions();
        goto Create_Options_Set;
    }

    RegisterExtensions();
#endif USEWIN32


    /* Get Windows Directory */

    if (GetWindowsDirectory(WindowsDirectory, MAXLINELEN) <= 0) return (FALSE);
    i = strlen(WindowsDirectory);
    if (WindowsDirectory[i-1] == '\\') WindowsDirectory[i-1] = '\0';


    /* Check for existence of the file */

    sprintf(buffer, "%s\\" PROGNAME ".INI", WindowsDirectory);

    if (OpenFile(buffer, &of, OF_READ | OF_EXIST) == -1) {

#ifdef USEWIN32
        goto Create_Options_Set;
#endif USEWIN32

        if (IDYES != YesNo (global.hwnd, PROGNAME " " JWPVERSION " cannot find the initialization file %s\n\n"
                                         "Do you want " PROGNAME " " JWPVERSION " to create it?", buffer))
            return (FALSE);
    } else {

        line[0] = '\0';

#ifdef USEWIN32
        GetPrivateProfileString("Program", "Version", " <ERROR> ", line, MAXLINELEN, buffer);

        if (!line[0]) strcpy(line, "Unknown");
        if (IDYES != YesNo (global.hwnd, PROGNAME " " JWPVERSION " found an initialization file %s "
                                         "from a previous version (%s).  Do you want to use the settings "
                                         "in it?", buffer, line))
            goto Create_Options_Set;
#endif USEWIN32

        /* Checks whether it is complete */

        GetPrivateProfileString("Hot Keys", "Help", " <ERROR> ", line, MAXLINELEN, buffer);

        if (!line[0] || !stricmp(line, " <ERROR> ")) {
            if (IDYES != YesNo (global.hwnd, "The existing initialization file %s is corrupted!\n\n"
                                             "Do you want " PROGNAME " " JWPVERSION " to re-create it?", buffer))
                return (FALSE);
        } else {
            GetPrivateProfileString("Program", "Version", " <ERROR> ", line, MAXLINELEN, buffer);

            if (!line[0] || stricmp(line, JWPVERSION)) {

                if (IDYES != YesNo (global.hwnd, "The existing initialization file %s is for %s version of " PROGNAME ".\n\n"
                                                 "Do you want " PROGNAME " " JWPVERSION " to overwrite it with its own?\n\n"
                                                 "(If you choose Yes, all settings will revert to " PROGNAME " " JWPVERSION "'s default.)",
                                                 buffer, (stricmp(line, JWPVERSION) < 0) ? "an older" : "a newer")) {

                    if (filename != NULL) strcpy(filename, of.szPathName);
                    return (TRUE);
                }
            } else {
                /* Good file */

                if (filename != NULL) strcpy(filename, of.szPathName);
                return (TRUE);
            }
        }
    }


    /* Create the file */

    fp = fopen(buffer, "w");
    if (fp == NULL) {
        ErrorMessage(global.hwnd, "Cannot create initialization file %s!\n", buffer);
        return (FALSE);
    }

    if (filename != NULL) strcpy(filename, buffer);


Create_Options_Set:

    time(&t);

    if (fp != NULL) {
        fprintf(fp, "; JWP.INI - Initialization file for JWP\n");
        fprintf(fp, ";           Version " JWPVERSION "\n");
        fprintf(fp, "; Last Modified %s", ctime(&t));
    }

    if (fp != NULL) fprintf(fp, "\n[Program]\n");
    if (fp != NULL) fprintf(fp, "%s=%s\n", "Version", JWPVERSION);

    /* Get the JWP directory */

    if (GetModuleFileName(hInstance, MainDirectory, MAXFILENAMELEN) <= 0) {
        ErrorMessage(global.hwnd, "Cannot get the JWP directory!\n");
        return (FALSE);
    }

    i = strlen(MainDirectory) - 1;
    if (MainDirectory[i-1] == '\\') i--;

    for (; i >= 0; i--) {
        if (MainDirectory[i] == '\\') break;
    }
    if (i >= 0)
        MainDirectory[i] = '\0';
    else {
        ErrorMessage(global.hwnd, "Cannot get the JWP directory!\n");
        return (FALSE);
    }

    if (fp != NULL) fprintf(fp, "\n[File Names]\n");
    SaveKeyAndINI(fp, "File Names", PROGNAME " Directory", "%s\\", MainDirectory);
    SaveKeyAndINI(fp, "File Names", "Templates", "%s\\%s\\", MainDirectory, "TEMPLATE");


    /* Process support files */

    if (!LocateFile(fp, "File Names", "Help File", PROGNAME ".HLP", NULL, MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "File Names", "Conversion Dictionary", "WNN.DCT", NULL, MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "File Names", "Conversion Dictionary Index", "WNN.IDX", NULL, MainDirectory)) return (FALSE);
    SaveKeyAndINI(fp, "File Names", "User Conversion Dictionary", "%s\\%s", MainDirectory, "USERDICT.DCT");
    SaveKeyAndINI(fp, "File Names", "Conversion Cache", "%s\\%s", MainDirectory, "CNVCACHE.SAV");
    if (!LocateFile(fp, "File Names", "Kanji Info", "KINFO.DAT", "Kanji Info Database", MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "File Names", "Kanji Info Index (Bushu)", "KINFOBSH.IDX", NULL, MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "File Names", "Kanji Info Index (Stroke)", "KINFOSTR.IDX", NULL, MainDirectory)) return (FALSE);

    if (!LocateFile(fp, "File Names", "Radical Lookup Database (Radical)", "RADKANJI.DAT", NULL, MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "File Names", "Radical Lookup Database Index (Radical)", "RADKANJI.IDX", NULL, MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "File Names", "Radical Lookup Database (Stroke)", "STROKNJI.DAT", NULL, MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "File Names", "Radical Lookup Database Index (Stroke)", "STROKNJI.IDX", NULL, MainDirectory)) return (FALSE);

    if (!LocateFile(fp, "File Names", "Global Glossary", "GLOSSARY.DAT", NULL, MainDirectory)) return (FALSE);


    /* Other sections */

    if (fp != NULL) fprintf(fp, "\n[Page Setup]\n");
    SaveKeyAndINI(fp, "Page Setup", "Measurement Unit",         "%s",   "inches");
    SaveKeyAndINI(fp, "Page Setup", "Margins",                  "%s",   "1.00 1.00 1.00 1.00");
    SaveKeyAndINI(fp, "Page Setup", "Long Text Justification",  "%s",   "Justify");
    SaveKeyAndINI(fp, "Page Setup", "Short Text Justification", "%s",   "Center");

    if (fp != NULL) fprintf(fp, "\n[Base ASCII Font]\n");
    SaveKeyAndINI(fp, "Base ASCII Font", "Face Name",           "%s",   "Times New Roman");
    SaveKeyAndINI(fp, "Base ASCII Font", "Size",                "%s",   "12.0");

    if (fp != NULL) fprintf(fp, "\n[Kanji Fonts]\n");
    SaveKeyAndINI(fp, "Kanji Fonts", "Number of Kanji Fonts",   "%d",   3);
    SaveKeyAndINI(fp, "Kanji Fonts", "System Font",             "%d",   1);
    SaveKeyAndINI(fp, "Kanji Fonts", "Default Base Font",       "%d",   1);
    SaveKeyAndINI(fp, "Kanji Fonts", "Default Print Font",      "%d",   3);
    SaveKeyAndINI(fp, "Kanji Fonts", "Default Print Size",      "%s",   "12.0");
    if (!LocateFile(fp, "Kanji Fonts", "Font #1", "K16X16.F00", "16x16 Kanji Font", MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "Kanji Fonts", "Font #2", "K24X24.F00", "24x24 Kanji Font", MainDirectory)) return (FALSE);
    if (!LocateFile(fp, "Kanji Fonts", "Font #3", "K48X48.F00", "48x48 Kanji Font", MainDirectory)) return (FALSE);

    if (fp != NULL) fprintf(fp, "\n[Glossary]\n");
    SaveKeyAndINI(fp, "Glossary", "Dynamic Glossary",           "%s",   "Yes");
    SaveKeyAndINI(fp, "Glossary", "Glossary List Disappears",   "%s",   "No");
    SaveKeyAndINI(fp, "Glossary", "Glossary List Visible",      "%s",   "No");
    SaveKeyAndINI(fp, "Glossary", "Glossary List Position",     "%d %d", 480, 80);
    SaveKeyAndINI(fp, "Glossary", "Glossary List Size",         "%d %d", 120, 140);

    if (fp != NULL) fprintf(fp, "\n[Undo]\n");
    SaveKeyAndINI(fp, "Undo", "Number of Undo Levels",          "%d",    50);
    SaveKeyAndINI(fp, "Undo", "Undo Levels Saved",              "%s",    "All");

    if (fp != NULL) fprintf(fp, "\n[Others]\n");
    SaveKeyAndINI(fp, "Others", "Program Position",             "%s",    "Maximized");
    SaveKeyAndINI(fp, "Others", "Show Input Line Editor",       "%s",    "No");
    SaveKeyAndINI(fp, "Others", "Relax Margin for Punctuation", "%s",    "Yes");
    SaveKeyAndINI(fp, "Others", "Dynamic Conversion",           "%s",    "No");
    SaveKeyAndINI(fp, "Others", "Conversion Bar Position",      "%s",    "Bottom");
    SaveKeyAndINI(fp, "Others", "Conversion Bar Size",          "%d %d", 300, 70);
    SaveKeyAndINI(fp, "Others", "Clipboard Format",             "%s",    "EUC");
    SaveKeyAndINI(fp, "Others", "Default Open Format",          "%s",    "Normal");
    SaveKeyAndINI(fp, "Others", "Draft View",                   "%s",    "No");
    SaveKeyAndINI(fp, "Others", "Show Special Symbols",         "%s",    "No");
    SaveKeyAndINI(fp, "Others", "Show Iconbar",                 "%s",    "Yes");
    SaveKeyAndINI(fp, "Others", "Show Status Bar",              "%s",    "Yes");
    SaveKeyAndINI(fp, "Others", "Show File Ruler",              "%s",    "Yes");
    SaveKeyAndINI(fp, "Others", "3D Dialog Boxes",              "%s",    "Yes");
    SaveKeyAndINI(fp, "Others", "Automatic Radical Lookup",     "%s",    "No");
    SaveKeyAndINI(fp, "Others", "Radical Lookup Sort Method",   "%d  ; 0 = No sort, 1 = strokes, 2 = level/strokes\n", 1);
    SaveKeyAndINI(fp, "Others", "Automatic Repagenation",       "%s",    "No");

    if (fp != NULL) fprintf(fp, "\n[Quick Files]\n");
    SaveKeyAndINI(fp, "Quick Files", "Number of Quick Files",   "%d",   0);

    if (fp != NULL) fprintf(fp, "\n[Dictionary]\n");
    SaveKeyAndINI(fp, "Dictionary", "Number of Dictionaries",   "%d",   1);

    strcpy(buffer, "EDICT");
    if (!LocateFile(fp, NULL, NULL, buffer, "Jim Breen's Japanese-English Dictionary", MainDirectory)) return (FALSE);
    strcpy(line, buffer);

    strcpy(buffer, "EDICT.JDX");
    if (!LocateFile(fp, NULL, NULL, buffer, "Index to Jim Breen's Japanese-English Dictionary", MainDirectory)) return (FALSE);
    strcat(line, "|");
    strcat(line, buffer);
    strcat(line, "|Yes|");
    strcat(line, "EDICT: Jim Breen's Japanese-English Dictionary");
    SaveKeyAndINI(fp, "Dictionary", "Dictionary #1",            "%s",   line);

    for (j = 0; FixedDictKeywords[j][0] != NULL; j++);
    SaveKeyAndINI(fp, "Dictionary", "Number of Keywords",       "%d",   j);
    for (i = 0; i < j; i++) {
        sprintf(buffer, "Key #%d", i+1);
        sprintf(line, "%s|Yes|%s", FixedDictKeywords[i][0], FixedDictKeywords[i][1]);
        SaveKeyAndINI(fp, "Dictionary", buffer,        "%s",   line);
    }

    if (fp != NULL) fprintf(fp, "\n[Hot Keys]\n");
    SaveKeyAndINI(fp, "Hot Keys", "Cut Text",                    "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Copy Text",                   "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Paste Text",                  "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Select All",                  "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Goto",                        "%s", "^G");
    SaveKeyAndINI(fp, "Hot Keys", "Insert Page Break",           "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Insert Soft Return",          "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Undo Last Change",            "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Redo Last Change",            "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Create New File",             "%s", "^N");
    SaveKeyAndINI(fp, "Hot Keys", "Open File",                   "%s", "^O");
    SaveKeyAndINI(fp, "Hot Keys", "Close File",                  "%s", "^F4");
    SaveKeyAndINI(fp, "Hot Keys", "Save File",                   "%s", "F11");
    SaveKeyAndINI(fp, "Hot Keys", "Save as Something Else",      "%s", "F12");
    SaveKeyAndINI(fp, "Hot Keys", "Find File",                   "%s", "^F");
    SaveKeyAndINI(fp, "Hot Keys", "Print File",                  "%s", "+^F12");
    SaveKeyAndINI(fp, "Hot Keys", "Page Setup",                  "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Repagenate File",             "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Printer Setup",               "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Edit File Summary",           "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Edit Header/Footer",          "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Kana/Text mode",       "%s", "F4");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle ASCII / JASCII mode",  "%s", "F3");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Input Line Editor",    "%s", "F5");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Active Input Window",  "%s", "^V");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Glossary List",        "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Draft View",           "%s", "^F2");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Iconbar",              "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Status Bar",           "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle File Ruler",           "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Special Symbols",      "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Toggle Dynamic Glossary",     "%s", "+^G");
    SaveKeyAndINI(fp, "Hot Keys", "Search for Text",             "%s", "^S");
    SaveKeyAndINI(fp, "Hot Keys", "Continue Search/Replace",     "%s", "F7");
    SaveKeyAndINI(fp, "Hot Keys", "Replace Text",                "%s", "^R");
    SaveKeyAndINI(fp, "Hot Keys", "Convert Kana to Kanji (L->R)","%s", "F2");
    SaveKeyAndINI(fp, "Hot Keys", "Convert Kana to Kanji (R->L)","%s", "+F2");
    SaveKeyAndINI(fp, "Hot Keys", "Kanji Lookup",                "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Radical Lookup",              "%s", "^L");
    SaveKeyAndINI(fp, "Hot Keys", "Kanji Info",                  "%s", "^K");
    SaveKeyAndINI(fp, "Hot Keys", "JIS Input",                   "%s", "^J");
    SaveKeyAndINI(fp, "Hot Keys", "JIS Table",                   "%s", "^T");
    SaveKeyAndINI(fp, "Hot Keys", "Set Options",                 "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Change Hot-Keys",             "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Maintain Kanji Fonts",        "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Show Statistics",             "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Dictionary",                  "%s", "F6");
    SaveKeyAndINI(fp, "Hot Keys", "Edit Glossary",               "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Edit User Conversions",       "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Cascade File Windows",        "%s", "+F4");
    SaveKeyAndINI(fp, "Hot Keys", "Tile File Windows",           "%s", "+F5");
    SaveKeyAndINI(fp, "Hot Keys", "Close all File Windows",      "%s", "None");
    SaveKeyAndINI(fp, "Hot Keys", "Help",                        "%s", "F1");

    if (fp != NULL) fclose(fp);

    RegisterExtensions();

#ifdef USEWIN32
    strcpy(buffer, ctime(&t));
    buffer[strlen(buffer) - 1] = '\0';

    if (RegCreateKey(HKEY_CURRENT_USER, "Software\\" PROGNAME "\\Program", &hsubkey) == ERROR_SUCCESS) {
        RegSetValueEx(hsubkey, "Options Saved", 0, REG_SZ, buffer, strlen(buffer) + 1);
        RegSetValueEx(hsubkey, "Version", 0, REG_SZ, JWPVERSION, strlen(JWPVERSION) + 1);
        RegCloseKey(hsubkey);
    }
#endif USEWIN32

    return (TRUE);
}
