/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */

#include "jwp.h"


static unsigned INT32 TotalAllocated = 0L;


#ifndef USEWIN32

#define MAGIC           0x42022667L

typedef struct MemoryStruct {
    INT32 magic;
    void far *page;
    unsigned int size;
    BOOL allocated;
	struct MemoryStruct far *next, far *prev;
} MEMHEADER;

typedef struct PageHeaderStruct {
    INT32 magic;
    HANDLE handle;
    unsigned int size;
    unsigned int used;
    unsigned int overhead;
    MEMHEADER far *data, far *empty;
    struct PageHeaderStruct far *next, far *prev;
} MEMPAGEHEADER;

typedef struct {
    MEMPAGEHEADER far *pages;
    int nr_pages;
} MAINMEMHEADER;

#define PAGESIZE        (4 * 1024)
#define USEABLESIZE     (PAGESIZE - sizeof(MEMPAGEHEADER) - sizeof(MEMHEADER))


static MAINMEMHEADER header = { NULL, 0 };



/* Make room by trimming off undo's */

static BOOL MakeRoom (void)
{
    FILEOPTIONS *f, *f1;
    UNDOBUF far *up;
    int largest;


    largest = -1;

    for (f = fileoptions; f != NULL; f = f->next) {
        if (f->undo != NULL && f->undolevels > largest) {
            largest = f->undolevels;
            f1 = f;
        }
    }

    if (largest < 0) return (FALSE);


    /* Start trimming */

    up = f1->undotail->prev;
    FreeUndo(f1->undotail);

    f1->undolevels--;
    if (up != NULL) up->next = NULL;

    if (up == NULL) {
        f1->undo = f1->undotail = NULL;
        f1->undolevels = 0;
    }

    return (TRUE);
}



static MEMPAGEHEADER far *AddPage(unsigned int n)
{
    void far *cp;
    MEMHEADER far *mp;
    MEMPAGEHEADER far *p;
    HANDLE handle = NULL;

    handle = GlobalAlloc(GHND, n);
    if (handle == NULL) return (NULL);

    if (header.pages == NULL || header.nr_pages <= 0) {
        p = header.pages = (MEMPAGEHEADER far *) GlobalLock(handle);
        p->prev = NULL;
    } else {
        for (p = header.pages; p->next != NULL; p = p->next);
        p->next = (MEMPAGEHEADER far *) GlobalLock(handle);
        p->next->prev = p;
        p = p->next;
    }

    p->magic = MAGIC;
    p->handle = handle;
    p->next = NULL;
    p->size = n;
    p->used = 0;
    p->overhead = sizeof(MEMPAGEHEADER) + sizeof(MEMHEADER);

    cp = ((char far *) p) + sizeof(MEMPAGEHEADER);
    mp = (MEMHEADER far *) cp;

    p->data = p->empty = mp;

    mp->magic = 0L;
    mp->allocated = FALSE;
    mp->page = p;
    mp->size = p->size - p->overhead;
    mp->next = mp->prev = NULL;

    header.nr_pages++;

	return (p);
}



static void DeletePage (MEMPAGEHEADER far *p)
{
    if (p->next == NULL && p->prev == NULL) {
        header.pages = NULL;
        header.nr_pages = 0;
        if (GlobalUnlock(p->handle) || GlobalFree(p->handle) != NULL) {
            ErrorMessage(global.hwnd, "Cannot free memory page 0x%x!", p->handle);
        }
	} else {
        if (p == header.pages) header.pages = p->next;
        header.nr_pages--;

        if (p->prev != NULL) p->prev->next = p->next;
        if (p->next != NULL) p->next->prev = p->prev;

        if (GlobalUnlock(p->handle) || GlobalFree(p->handle) != NULL) {
            ErrorMessage(global.hwnd, "Cannot free memory page 0x%x!", p->handle);
        }
    }
}

#endif USEWIN32


void far *SegHeapAlloc (unsigned INT32 n)
{
#ifdef USEWIN32

    void *p;

    for (;;) {
        p = (void *) LocalAlloc(LPTR, n);
        if (p != NULL) break;

        if (!RetryMessage("Out of Memory:\nAllocating %lu bytes", n))
            return (NULL);
    }

    TotalAllocated += LocalSize(p);

    return (p);

#else

    MEMPAGEHEADER far *p = NULL;
    MEMHEADER far *mp;
    char far *cp;

    if (n >= C64K) {
        ErrorMessage(global.hwnd, "SegHeapAlloc: size (%lu) > 64K!\n\n"
                                  "The program is not stable.  You should save your work NOW!",
                                  n);
        n = C64K;
    }

    /* Larger than page size? */

    if (n > USEABLESIZE) {
        for (;;) {
            p = AddPage(n + sizeof(MEMPAGEHEADER) + sizeof(MEMHEADER));

            if (p != NULL) break;

            if (MakeRoom()) continue;
            if (!RetryMessage("Out of Global Memory:\nAllocating %lu bytes", n))
                return (NULL);
        }

        mp = p->data;
        mp->magic = MAGIC;
        mp->allocated = TRUE;

        mp->next = mp->prev = NULL;

        p->used = n;
        p->empty = NULL;

        cp = ((char far *) mp) + sizeof(MEMHEADER);
        _fmemset(cp, 0, n);

        return ((void far *) cp);
    }


LookForHole:

    /* Search for the hole */

    for (p = header.pages; p != NULL; p = p->next) {
        /* Scan the chains */
        if (p->size - p->used - p->overhead <= 0) continue;
        if (p->empty == NULL) continue;

        for (mp = p->empty; mp != NULL; mp = mp->next) {
            if (!mp->allocated && mp->size >= n) break;
        }

        if (mp != NULL) break;
    }

    /* New page needed? */

    if (p == NULL) {
        for (;;) {
            p = AddPage(PAGESIZE);

            if (p != NULL) break;

            if (MakeRoom()) goto LookForHole;
            if (!RetryMessage("Out of Global Memory:\nAllocating %lu bytes", n))
                return (NULL);
        }

        mp = p->data;
    }

    /* Do we need to break it up? */

    if (mp->size - n > sizeof(MEMHEADER)) {
        MEMHEADER far *mp2;

        cp = ((char far *) mp) + n + sizeof(MEMHEADER);
        mp2 = (MEMHEADER far *) cp;

        mp2->magic = 0L;
        mp2->allocated = FALSE;
        mp2->page = p;
        mp2->size = mp->size - n - sizeof(MEMHEADER);

        mp2->next = mp->next;
        mp2->prev = mp;
        if (mp->next != NULL) mp->next->prev = mp2;
        mp->next = mp2;


        p->overhead += sizeof(MEMHEADER);

        mp->size = n;
    }

    mp->magic = MAGIC;
    mp->allocated = TRUE;

    p->used += n;
    cp = ((char far *) mp) + sizeof(MEMHEADER);

    /* Search for the next empty hole */

    for (; mp != NULL; mp = mp->next) {
        if (!mp->allocated && mp->size > 0) break;
    }

    p->empty = mp;

    _fmemset(cp, 0, n);

    return ((void far *) cp);

#endif USEWIN32
}


void SegHeapFree (void far *vp)
{
#ifdef USEWIN32

    TotalAllocated -= LocalSize(vp);
    LocalFree(vp);

#else

    MEMPAGEHEADER far *p;
    MEMHEADER far *mp, far *mp2;
    char far *cp;

    cp = ((char far *) vp) - sizeof(MEMHEADER);
    mp = (MEMHEADER far *) cp;

    if (mp->magic != MAGIC || !mp->allocated) {
        ErrorMessage(global.hwnd, "Trying to deallocate invalid memory block [%Fp, %u byte%s]!\n\n"
                                  "The program is not stable.  You should save your work NOW!",
                        vp, mp->size, (mp->size > 1) ? "s" : "");
        return;
    }

    p = (MEMPAGEHEADER far *) mp->page;
    p->used -= mp->size;

    mp->magic = 0L;
    mp->allocated = FALSE;

    /* Merge? */

    mp2 = mp->prev;

    if (mp2 != NULL && !mp2->allocated) {
        mp2->next = mp->next;
        if (mp->next != NULL) mp->next->prev = mp2;
        mp2->size += mp->size + sizeof(MEMHEADER);

        p->overhead -= sizeof(MEMHEADER);

        mp = mp2;
    }

    mp2 = mp->next;

    if (mp2 != NULL && !mp2->allocated) {
        mp->next = mp2->next;
        if (mp2->next != NULL) mp2->next->prev = mp;

        mp->size += mp2->size + sizeof(MEMHEADER);

        p->overhead -= sizeof(MEMHEADER);
    }

    if (mp->prev == NULL && mp->next == NULL) {
        DeletePage(p);
    } else {
        if (p->empty == NULL || mp < p->empty) p->empty = mp;
    }

#endif USEWIN32
}



void far *SegHeapRealloc (void far *p, unsigned INT32 n)
{
#ifdef USEWIN32

    void *p1;

    TotalAllocated -= LocalSize(p);

    for (;;) {
        p1 = (void *) LocalReAlloc(p, n, LMEM_MOVEABLE | LMEM_ZEROINIT);
        if (p1 != NULL) break;

        if (!RetryMessage("Out of Memory:\nReallocating %lu bytes", n))
            return (NULL);
    }

    TotalAllocated += LocalSize(p);
    return (p1);

#else

    MEMHEADER far *mp;
    char far *cp;

    /* Block already large enough? */

    cp = ((char far *) p) - sizeof(MEMHEADER);
    mp = (MEMHEADER far *) cp;

    if (mp->magic != MAGIC) {
        ErrorMessage(global.hwnd, "Trying to reallocate invalid memory block [%Fp, %u byte%s]!\n\n"
                                  "The program is not stable.  You should save your work NOW!",
                        p, mp->size, (mp->size > 1) ? "s" : "");

        /* Minimize the damage */

        cp = SegHeapAlloc(n);
        if (cp == NULL) return (NULL);
        _fmemcpy(cp, p, (mp->size >= n) ? n : mp->size);
        return (cp);
    }

    if (mp->size >= n) return (p);      /* No need to do anything */

    /* Else swap to another block */

    cp = SegHeapAlloc(n);
    if (cp == NULL) return (NULL);

    _fmemcpy(cp, p, (mp->size >= n) ? n : mp->size);
    SegHeapFree(p);

    return ((void far *) cp);

#endif USEWIN32
}



unsigned INT32 SegHeapGetSize (void far *p)
{
#ifdef USEWIN32

    return (LocalSize(p));

#else

    MEMHEADER far *mp;
    char far *cp;

    cp = ((char far *) p) - sizeof(MEMHEADER);

    mp = (MEMHEADER far *) cp;

    if (mp->magic != MAGIC) {
        ErrorMessage(global.hwnd, "Trying to get size of invalid memory block [%Fp, %u byte%s]!\n\n"
                                  "The program is not stable.  You should save your work NOW!",
                        p, mp->size, (mp->size > 1) ? "s" : "");
        return (0);
    }

    return (mp->size);

#endif USEWIN32
}



void far *SegHeapDuplicate (void far *p)
{
    unsigned INT32 len;
    void far *p1;

    len = SegHeapGetSize(p);
    p1 = SegHeapAlloc(len);
    _fmemcpy(p1, p, len);

    return (p1);
}



void *MyLocalAlloc (unsigned INT32 size)
{
    void *p = NULL;

    while (p == NULL) {
        p = (void *) LocalAlloc(LPTR, size);
        if (p == NULL) {
            if (!RetryMessage("Out of Local Memory:\nAllocating %lu bytes", size))
                return (NULL);
        }
    }

    //TotalAllocated += LocalSize(p);

    return (p);
}



void CountMemoryUsage (INT32 *allocated, INT32 *used, INT32 *overhead)
{
#ifdef USEWIN32

    int i;
    HANDLE ProcessHeap;
    PROCESS_HEAP_ENTRY heapinfo;

    ProcessHeap = GetProcessHeap();

    *allocated = *used = *overhead = 0L;

    HeapLock(ProcessHeap);

    heapinfo.lpData = NULL;

    if (!HeapWalk(ProcessHeap, &heapinfo)) {
        HeapUnlock(ProcessHeap);
        *allocated = *used = TotalAllocated;
        *overhead = 0L;
        return;
    }

    do {
        if (heapinfo.wFlags & PROCESS_HEAP_REGION) {
            *allocated += heapinfo.cbData;
            *overhead += heapinfo.cbOverhead;
        } else if (heapinfo.wFlags & PROCESS_HEAP_UNCOMMITTED_RANGE) {
            *overhead += heapinfo.cbOverhead;
        } else if (heapinfo.wFlags & PROCESS_HEAP_ENTRY_BUSY) {
            *used += heapinfo.cbData;
            *overhead += heapinfo.cbOverhead;
        } else {
            *overhead += heapinfo.cbOverhead;
        }
    } while (HeapWalk(ProcessHeap, &heapinfo));

    HeapUnlock(ProcessHeap);

    return;

#else

    MEMPAGEHEADER far *p;

    *allocated = *used = *overhead = 0L;

    for (p = header.pages; p != NULL; p = p->next) {
        *allocated += p->size;
        *used += p->used;
        *overhead += p->overhead;
    }

#endif USEWIN32
}


void FreeAllMemory (void)
{
#ifdef USEWIN32

#else

    MEMPAGEHEADER far *p, far *p1;

    for (p = header.pages; p != NULL; ) {
        p1 = p->next;
        GlobalFree(GlobalUnlock(p->handle));
        p = p1;
    }

    header.pages = NULL;
    header.nr_pages = 0;

#endif USEWIN32
}
