/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */

#include "jwp.h"
#include <commdlg.h>
#include <cderr.h>
#include <math.h>


MEASUREINFO Measurements[] = {
    { "inches",         "\"",       2,  1.00 },
    { "centimeters",    " cm",      1,  2.536 },
    { "millimeters",    " mm",      0,  25.36 },
    { "points",         " pts",     1,  72.00 },
    { "pixels",         " dots",    0,  0.00 },
    { NULL,             NULL,       0,  0.00 }
};

static int PrintFrom, PrintTo;
static HWND PrintDialogBoxHwnd;
static BOOL UserAbort, PrintError;

#define ESCAPECHAR      '&'

static struct {
    char escape;
    int summary;
} HeaderStrings[] = {
    { 'L', 0 },
    { 'S', 1 },
    { 'A', 2 },
    { 'K', 3 },
    { 'C', 4 },
    { 'T', -1 },
    { 'D', -2 },
    { 'P', -3 },
    { 'F', -4 },
    { '\0', 0 }
};

static KANJI NoRotate[] = {
    0x213a, 0x213b, 0x213c,
    0x2141, 0x2142, 0x2143, 0x2144, 0x2145,
    0x214a, 0x214b, 0x214c, 0x214d, 0x214e, 0x214f,
    0x2150, 0x2151, 0x2152, 0x2153, 0x2154, 0x2155, 0x2156,
    0x2157, 0x2158, 0x2159, 0x215a, 0x215b,
    0x2161, 0x2162, 0x2163, 0x2164, 0x2165, 0x2166, 0x2167,
    0x222a, 0x222b, 0x222e,
    0x0000
};

static KANJI NeedOffset[] = {
    0x2122, 0x2123, 0x2124, 0x2125,
    0x2421, 0x2423, 0x2425, 0x2427, 0x2429, 0x2443, 0x2463, 0x2465, 0x2467, 0x246e,
    0x2521, 0x2523, 0x2525, 0x2527, 0x2529, 0x2543, 0x2563, 0x2565, 0x2567, 0x256e,
    0x2575, 0x2576,
    0x0000
};


static struct {
    char *desc;
    LONG errcode;
} PrinterErrors[] = {
    { "Program error - dialog box failure",       CDERR_DIALOGFAILURE    },

    { "General Codes",                            CDERR_GENERALCODES     },
    { "Program error - wrong struct size",        CDERR_STRUCTSIZE       },
    { "Program error - cannot initialize",        CDERR_INITIALIZATION   },
    { "Program error - cannot find template",     CDERR_NOTEMPLATE       },
    { "Program error - cannot find instance",     CDERR_NOHINSTANCE      },
    { "Program error - cannot load strings",      CDERR_LOADSTRFAILURE   },
    { "Program error - cannot find resource",     CDERR_FINDRESFAILURE   },
    { "Program error - cannot load resource",     CDERR_LOADRESFAILURE   },
    { "Program error - cannot lock resource",     CDERR_LOCKRESFAILURE   },
    { "Program error - cannot allocate memory",   CDERR_MEMALLOCFAILURE  },
    { "Program error - cannot lock memory",       CDERR_MEMLOCKFAILURE   },
    { "Program error - cannot find hook",         CDERR_NOHOOK           },
    { "Program error - cannot register messages", CDERR_REGISTERMSGFAIL  },

    { "Printer Codes",                            PDERR_PRINTERCODES     },
    { "Cannot setup printer",                     PDERR_SETUPFAILURE     },
    { "Printer parse failed",                     PDERR_PARSEFAILURE     },
    { "Cannot get default printer",               PDERR_RETDEFFAILURE    },
    { "Cannot load printer driver",               PDERR_LOADDRVFAILURE   },
    { "Cannot get printer modes",                 PDERR_GETDEVMODEFAIL   },
    { "Cannot initialize printer",                PDERR_INITFAILURE      },
    { "No printer device found",                  PDERR_NODEVICES        },
    { "No default printer found",                 PDERR_NODEFAULTPRN     },
    { "Printer modes mismatch",                   PDERR_DNDMMISMATCH     },
    { "Cannot create Printer IC",                 PDERR_CREATEICFAILURE  },
    { "Cannot find printer",                      PDERR_PRINTERNOTFOUND  },
    { "Default printer is changed",               PDERR_DEFAULTDIFFERENT },

    { "Cannot select printer fonts",              CFERR_CHOOSEFONTCODES  },
    { "Cannot find printer fonts",                CFERR_NOFONTS          },
    { "Maximum less than minimum",                CFERR_MAXLESSTHANMIN   },

    { "Program error - bad file name codes",      FNERR_FILENAMECODES    },
    { "Program error - cannot subclass",          FNERR_SUBCLASSFAILURE  },
    { "Program error - invalid filename",         FNERR_INVALIDFILENAME  },
    { "Program error - buffer too small",         FNERR_BUFFERTOOSMALL   },

    { "Program error - find / replace codes",     FRERR_FINDREPLACECODES },
    { "Program error - buffer length is zero",    FRERR_BUFFERLENGTHZERO },

    { "Program error - choose color codes",       CCERR_CHOOSECOLORCODES },
    { NULL,                                       0x0000                 }
};



static PrinterErrorText (LONG errcode)
{
    int i;

    for (i = 0; PrinterErrors[i].desc != NULL && PrinterErrors[i].errcode != errcode; i++);

    if (PrinterErrors[i].desc == NULL) return ("Cannot initialize printer");
    return (PrinterErrors[i].desc);
}


static HGLOBAL GlobalDuplicate (HGLOBAL handle)
{
    HGLOBAL dup;
    void far *p, far *q;
    long int size;

    size = GlobalSize(handle);
    if (size <= 0) return (NULL);

    dup = GlobalAlloc(GHND, size);
    if (dup == NULL) return (NULL);

    p = GlobalLock(handle);
    if (p == NULL) {
        GlobalFree(dup);
        return (NULL);
    }

    q = GlobalLock(dup);
    if (q == NULL) {
        GlobalFree(dup);
        GlobalUnlock(handle);
        return (NULL);
    }

    _fmemcpy(q, p, size);

    GlobalUnlock(handle);
    GlobalUnlock(dup);
    return (dup);
}


static BOOL WithinSet (KANJI ch, KANJI set[])
{
    int i;

    ch &= 0x7f7f;

    for (i = 0; set[i] != 0; i++) {
        if (set[i] == ch) return (TRUE);
    }
    return (FALSE);
}



static void CalcRotateOffset (BYTE far *bitsbuffer, int dim, int bmsize, int *x, int *y)
{
    int i, j;
    int linesize = bmsize / dim;
    RECT rect;
    static int masks[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

    for (i = 0; i < dim; i++) {
        for (j = 0; j < dim; j++) {
            if (!(bitsbuffer[linesize * i + (j / 8)] & masks[j % 8])) break;
        }
        if (j < dim) break;
    }

    rect.top = i;

    for (i = dim - 1; i >= 0; i--) {
        for (j = 0; j < dim; j++) {
            if (!(bitsbuffer[linesize * i + (j / 8)] & masks[j % 8])) break;
        }
        if (j < dim) break;
    }

    rect.bottom = dim - 1 - i;

    for (j = 0; j < dim; j++) {
        for (i = 0; i < dim; i++) {
            if (!(bitsbuffer[linesize * i + (j / 8)] & masks[j % 8])) break;
        }
        if (i < dim) break;
    }

    rect.left = j;

    for (j = dim - 1; j >= 0; j--) {
        for (i = 0; i < dim; i++) {
            if (!(bitsbuffer[linesize * i + (j / 8)] & masks[j % 8])) break;
        }
        if (i < dim) break;
    }

    rect.right = dim - 1 - j;

    //*x = rect.left - rect.bottom;
    *x = rect.left;
    *y = rect.top - rect.right;
}



MEASUREMENT FindMeasurement (char *name, MEASUREMENT measure)
{
    int i;

    if (measure >= 0) {
        strcpy(name, Measurements[measure].name);
        return (measure);
    }

    for (i = 0; Measurements[i].name != NULL; i++) {
        if (!stricmp(name, Measurements[i].name)) return ((MEASUREMENT) i);
    }
    return (-1);
}



static BOOL PageOrientation (BOOL landscape, BOOL set)
{
    DEVMODE far *devmode;

    if (global.hdevmode == NULL) return (landscape);

    devmode = GlobalLock(global.hdevmode);
    if (devmode == NULL) return (landscape);

    if (set) {
        if (landscape) {
            devmode->dmOrientation = DMORIENT_LANDSCAPE;
        } else {
            devmode->dmOrientation = DMORIENT_PORTRAIT;
        }
    } else {
        switch (devmode->dmOrientation) {
            case DMORIENT_LANDSCAPE:    landscape = TRUE; break;
            case DMORIENT_PORTRAIT:     landscape = FALSE; break;
        }
    }

    GlobalUnlock(global.hdevmode);
    return (landscape);
}



HDC GetPrinterDC (BOOL IcOnly, BOOL landscape, char *name)
{
    HDC hdc;
    char far *device, far *driver, far *output;
    char buffer[MAXLINELEN];
    DEVMODE far *devmode;
    DEVNAMES far *devnames;

    /* Opens the printer device */

    if (name != NULL) strcpy(name, "None");

    if (global.hdevnames != NULL) {
        devnames = GlobalLock(global.hdevnames);
        if (devnames == NULL) return (NULL);
    } else {
        return (NULL);
    }

    if (global.hdevmode != NULL) {
        (void) PageOrientation(landscape, TRUE);
        devmode = GlobalLock(global.hdevmode);
    } else {
        devmode = NULL;
    }

    driver = ((char far *) devnames) + devnames->wDriverOffset;
    device = ((char far *) devnames) + devnames->wDeviceOffset;
    output = ((char far *) devnames) + devnames->wOutputOffset;

    _fstrcpy(buffer, device);
    if (name != NULL) sprintf(name, "%s on ", buffer);
    _fstrcpy(buffer, output);
    if (name != NULL) strcat(name, buffer);

    if (IcOnly) {
        hdc = CreateIC(driver, device, output, devmode);
    } else {
        hdc = CreateDC(driver, device, output, devmode);
    }

    if (devmode != NULL) GlobalUnlock(global.hdevmode);
    GlobalUnlock(global.hdevnames);
    return (hdc);
}



BOOL SetupPrinter (BOOL ReloadDriver, BOOL landscape)
{
    int x;
    HDC hdc;
    POINT printdots;
    double temp, printscale;
    double nr_points[2], dpp[2];
    char buffer[MAXLINELEN];

    if (ReloadDriver) {
        hdc = GetPrinterDC(TRUE, landscape, buffer);

        if (global.printer != NULL) {
            FreeMem(global.printer);
            global.printer = NULL;
        }

        if (hdc == NULL) {
            if (global.hdevnames != NULL && global.hdevmode != NULL) {
                ErrorMessage(global.hwnd, "Cannot initialize the chosen printer!");
            }
            hdc = CreateDC("DISPLAY", NULL, NULL, NULL);
            global.printer = DupString("None");
        } else if (!(GetDeviceCaps(hdc, RASTERCAPS) & RC_BITBLT)) {
UseDisplay: DeleteDC(hdc);
            hdc = CreateDC("DISPLAY", NULL, NULL, NULL);
            global.printer = DupString("None");
        } else {
            global.printer = DupString(buffer);
        }

        global.resolution.x = GetDeviceCaps(hdc, LOGPIXELSX);
        global.resolution.y = GetDeviceCaps(hdc, LOGPIXELSY);

        global.printable.x = GetDeviceCaps(hdc, HORZRES);
        global.printable.y = GetDeviceCaps(hdc, VERTRES);

        Measurements[M_PIXELS].conversion = global.resolution.y;

        global.paper[0] = (double) GetDeviceCaps(hdc, HORZSIZE) / Measurements[M_MM].conversion;
        global.paper[1] = (double) GetDeviceCaps(hdc, VERTSIZE) / Measurements[M_MM].conversion;

        if (global.resolution.x <= 0 || global.resolution.y <= 0 ||
            global.printable.x <= 0 || global.printable.y <= 0 ||
            global.paper[0] <= 0 || global.paper[1] <= 0) {

                ErrorMessage(global.hwnd, "The chosen printer driver, \"%s\", "
                                          "has problems - some paper-size/resolution "
                                          "measurements are zero.",
                             buffer);
                goto UseDisplay;
        }
    }

    dpp[0] = (double) global.resolution.x / 72.0;
    dpp[1] = (double) global.resolution.y / 72.0;


    /* We compensate for the printer's aspect ratio.  If necessary, */
    /* stretching or compressing the horizontal axis.               */

    if (InstalledFonts != NULL) {
        x = (int) floor(global.printscale * (double) global.resolution.y / 72.0 + 0.5);      /* dots */
        printscale = (double) x / (double) PRINTFONT->height;

        printdots.y = PRINTFONT->height * printscale;
        temp = (double) PRINTFONT->width * printscale;
        temp = temp * (double) global.resolution.x / (double) global.resolution.y;
        printdots.x = temp;

        /* This is the number of points per printing font */

        nr_points[0] = (double) printdots.x / dpp[0];
        nr_points[1] = (double) printdots.y / dpp[1];

        global.dispscale = (double) BASEFONT->height / (double) PRINTFONT->height;
    }

    if (ReloadDriver && hdc != NULL) DeleteDC(hdc);
    return (TRUE);
}



BOOL InitPrinting (void)
{
    BOOL DefPrinterOK = TRUE;
    PRINTDLG pd;

    memset(&pd, 0, sizeof(PRINTDLG));

    pd.lStructSize = sizeof(PRINTDLG);
    pd.hwndOwner = global.hwnd;
    pd.Flags = PD_RETURNDEFAULT;

    if (PrintDlg(&pd) == 0 || pd.hDevNames == NULL || pd.hDevMode == NULL) {
        DefPrinterOK = FALSE;
        global.hdevnames = global.hdevmode = NULL;
    } else {
        global.hdevnames = pd.hDevNames;
        global.hdevmode = pd.hDevMode;
    }

    if (!SetupPrinter(TRUE, global.pagesetup.landscape))
        DefPrinterOK = FALSE;

    return (DefPrinterOK);
}


void ChangePrinterOrFonts (FILEOPTIONS *f, BOOL always, BOOL reformat)
{
    int x;
    double printscale;
    HCURSOR hcursor;
    HFONT hfont;
    HDC hdc;
    HWND FocusHwnd = NULL;
    DEVMODE far *devmode;
    static DEVMODE cdevmode;
    static BOOL NoPrinter = TRUE;


    hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
    ShowCursor(TRUE);


    /* Anything changed? */

    if (!always) {
        if (global.hdevmode != NULL) {
            (void) PageOrientation(f->pagesetup.landscape, TRUE);
            devmode = GlobalLock(global.hdevmode);
            if (devmode != NULL) {
                if (!_fmemcmp(&cdevmode, devmode, sizeof(DEVMODE))) {
                    GlobalUnlock(global.hdevmode);
                    goto Do_Reformat;
                }
                GlobalUnlock(global.hdevmode);
            }
        } else {
            if (NoPrinter) goto Do_Reformat;
        }
    } else {
        _fmemset(&cdevmode, 0, sizeof(DEVMODE));
        NoPrinter = (global.hdevmode == NULL);
    }


    FocusHwnd = SetFocus(NULL);

    /* Change the printer setting */

    if (f == NULL || !(f->type & FN_NORMAL)) {
        SetupPrinter(TRUE, global.pagesetup.landscape);
        hdc = GetPrinterDC(TRUE, global.pagesetup.landscape, NULL);
    } else {
        SetupPrinter(TRUE, f->pagesetup.landscape);
        hdc = GetPrinterDC(TRUE, f->pagesetup.landscape, NULL);
    }

    if (hdc == NULL) return;


    /* Store the information into the cache */

    if (global.hdevmode != NULL) {
        devmode = GlobalLock(global.hdevmode);
        if (devmode != NULL) {
            _fmemcpy(&cdevmode, devmode, sizeof(DEVMODE));
            NoPrinter = FALSE;
            GlobalUnlock(global.hdevmode);
        }
    } else {
        NoPrinter = TRUE;
    }


    /* Get font width tables */

    if (hdc == NULL) hdc = CreateIC("DISPLAY", NULL, NULL, NULL);

    hfont = SelectAsciiFont(hdc, DefAsciiFont.facename, DefAsciiFont.size, NULL);

    if (DefAsciiFont.hfont != NULL) DeleteObject(DefAsciiFont.hfont);

    x = (int) floor(global.printscale * (double) global.resolution.y / 72.0 + 0.5);      /* dots */
    printscale = (double) x / (double) PRINTFONT->height;

    DefAsciiFont.hfont =
        FindMatchingScreenFont(hfont, global.dispscale / printscale,
                               &(DefAsciiFont.textmetric));
    DeleteObject(hfont);
    DeleteDC(hdc);

    FontCharWidth (0, -1);      /* Reload width tables */

    global.leading = (int) ((double) PRINTFONT->leading * global.dispscale);
    if (global.leading % 2 != 0) global.leading++;

    global.spacing = (int) ((double) PRINTFONT->spacing * global.dispscale);
    if (global.spacing % 2 != 0) global.spacing++;

Do_Reformat:

    if (reformat && f != NULL && (f->type & FN_NORMAL)) {
        PARAGRAPH far *p;
        double spc;

        for (p = f->paragraph; p != NULL; p = p->next) {
            spc = ((double) global.spacing) / ((double) f->spacing);
            p->spacing = (int) (((double) p->spacing) * spc);
            if (p->spacing % 2 != 0) p->spacing++;
        }

        f->leading = global.leading;
        f->spacing = global.spacing;
        f->basefont = BASEFONT;
        f->linelen = CalcLineLength(f);

        f->repage = TRUE;

        ReformatFile(f);

        if (!FindCaret(f, TRUE)) MoveIntoWindow(f);
        InvalidateRect(f->hwnd, NULL, TRUE);
        if(RULER(f).hwnd != NULL) InvalidateRect(RULER(f).hwnd, NULL, TRUE);
    } else {
        if (global.autorepage) RepagenateFile(f);
    }

    ShowCursor(FALSE);
    SetCursor(hcursor);

    if (FocusHwnd != NULL) SetFocus(FocusHwnd);
}



static void FormatTimeDate (char *timep, char *datep)
{
    time_t t;
    struct tm *tmp;
    static int idate, itime;
    static char sdate[5], stime[5], sam[10], spm[10];
    static BOOL Filled = FALSE;

    if (!Filled || (timep == NULL && datep == NULL)) {
#ifdef USEWIN32
        int size;
        char buffer[MAXLINELEN];

        if (RegQueryValue(HKEY_CURRENT_USER, "Control Panel\\International\\iDate", buffer, &size) == ERROR_SUCCESS) {
            idate = atoi(buffer);
        } else {
            idate = GetProfileInt("intl", "iDate", 0);
        }
        if (RegQueryValue(HKEY_CURRENT_USER, "Control Panel\\International\\iTime", buffer, &size) == ERROR_SUCCESS) {
            itime = atoi(buffer);
        } else {
            itime = GetProfileInt("intl", "iTime", 0);
        }
        if (RegQueryValue(HKEY_CURRENT_USER, "Control Panel\\International\\sDate", buffer, &size) == ERROR_SUCCESS) {
            strncpy(sdate, buffer, 4);
        } else {
            GetProfileString("intl", "sDate", "/", sdate, 4);
        }
        if (RegQueryValue(HKEY_CURRENT_USER, "Control Panel\\International\\sTime", buffer, &size) == ERROR_SUCCESS) {
            strncpy(stime, buffer, 4);
        } else {
            GetProfileString("intl", "sTime", ":", stime, 4);
        }
        if (RegQueryValue(HKEY_CURRENT_USER, "Control Panel\\International\\s1159", buffer, &size) == ERROR_SUCCESS) {
            strncpy(sam, buffer, 9);
        } else {
            GetProfileString("intl", "s1159", "AM", sam, 9);
        }
        if (RegQueryValue(HKEY_CURRENT_USER, "Control Panel\\International\\s2359", buffer, &size) == ERROR_SUCCESS) {
            strncpy(spm, buffer, 9);
        } else {
            GetProfileString("intl", "s2359", "PM", spm, 9);
        }
#else
        idate = GetProfileInt("intl", "iDate", 0);
        itime = GetProfileInt("intl", "iTime", 0);

        GetProfileString("intl", "sDate", "/", sdate, 4);
        GetProfileString("intl", "sTime", ":", stime, 4);
        GetProfileString("intl", "s1159", "AM", sam, 9);
        GetProfileString("intl", "s2359", "PM", spm, 9);
#endif USEWIN32
        Filled = TRUE;
    }

    time(&t);
    tmp = localtime(&t);

    if (datep != NULL) {
        sprintf(datep, "%d%s%d%s%d",
                idate == 1 ? tmp->tm_mday : idate == 2 ? (tmp->tm_year >= 2000 ? tmp->tm_year : tmp->tm_year % 100) : (tmp->tm_mon + 1), sdate,
                idate == 1 ? (tmp->tm_mon + 1) : idate == 2 ? (tmp->tm_mon + 1) : tmp->tm_mday, sdate,
                idate == 1 ? (tmp->tm_year >= 2000 ? tmp->tm_year : tmp->tm_year % 100) : idate == 2 ? tmp->tm_mday : (tmp->tm_year >= 2000 ? tmp->tm_year : tmp->tm_year % 100));
    }
    if (timep != NULL) {
        if (itime == 1) {
            sprintf(timep, "%02d%s%02d", tmp->tm_hour, stime, tmp->tm_min);
        } else {
            sprintf(timep, "%d%s%02d %s",
                    (tmp->tm_hour % 12) ? (tmp->tm_hour % 12) : 12, stime,
                    tmp->tm_min, (tmp->tm_hour / 12) ? spm : sam);
        }
    }
}



static void BuildHeaderString (FILEOPTIONS *f, KANJI far *kp, int PageNum, KANJI far *out)
{
    int i, j, k;
    char cch;
    KANJI kch;
    KANJI far *kp1;

    j = 0;

    for (i = 0; kp[i]; i++) {
        kch = kp[i];
        if (ISKANJI(kch)) cch = LOBYTE(TranslateJAscii(kch, TRUE));
        else cch = LOBYTE(kch);
        if ('a' <= cch && cch <= 'z') cch -= 32;

        if (cch == ESCAPECHAR) {
            i++;
            if (ISKANJI(kp[i])) cch = LOBYTE(TranslateJAscii(kp[i], TRUE));
            else cch = LOBYTE(kp[i]);
            if ('a' <= cch && cch <= 'z') cch -= 32;

            for (k = 0; HeaderStrings[k].escape; k++) {
                if (HeaderStrings[k].escape == cch) break;
            }

            if (!HeaderStrings[k].escape) {
                out[j++] = kch; out[j++] = kp[i]; break;
            } else if (HeaderStrings[k].summary >= 0) {
                kp1 = f->summary[HeaderStrings[k].summary];
                if (kp1 != NULL) for (k = 0; kp1[k]; k++) out[j++] = kp1[k];
            } else {
                char buffer[50];

                switch (HeaderStrings[k].summary) {
                    case -1:    /* Time */
                        FormatTimeDate(buffer, NULL);
                        break;

                    case -2:    /* Date */
                        FormatTimeDate(NULL, buffer);
                        break;

                    case -3:    /* Page */
                        sprintf(buffer, "%d", PageNum);
                        break;

                    case -4:    /* File name */
                        for (k = strlen(f->filename) - 1; k >= 0; k--) {
                            if (strchr(":/\\", f->filename[k]) != NULL) break;
                        }
                        strcpy(buffer, f->filename + k + 1);
                        break;
                }

                for (k = 0; buffer[k]; k++) out[j++] = buffer[k];
            }
            continue;
        }

        out[j++] = kch;
    }

    out[j] = 0;
}



static BOOL MoveToNextLine (PARAGRAPH far **pp, ONELINE far **lp, int Direction)
{
    if (Direction > 0) {
        *lp = (*lp)->next;
        if (*lp == NULL) {
            *pp = (*pp)->next;
            if (*pp == NULL) return (FALSE);
            *lp = (*pp)->lines;
        }
        return (TRUE);
    } else if (Direction < 0) {
        *lp = (*lp)->prev;
        if (*lp == NULL) {
            *pp = (*pp)->prev;
            if (*pp == NULL) return (FALSE);
            *lp = (*pp)->lastline;
        }
        return (TRUE);
    } else {
        return (TRUE);
    }
}



int RepagenateFile(FILEOPTIONS *f)
{
    int i, x, y, size;
    int VerticalWhiteSpace, MarginSpace;
    int PageLength, LineHeight;
    double printscale;
    PARAGRAPH far *pp;
    ONELINE far *lp;
    BOOL changed = FALSE;
    //char buffer[MAXLINELEN];

    //StatusMessage("Repagenating File...");


    VerticalWhiteSpace = ((global.paper[1] * global.resolution.y) - global.printable.y) / 2;
    MarginSpace = (f->pagesetup.margins[2] * global.resolution.y) - VerticalWhiteSpace;
    MarginSpace += (f->pagesetup.margins[3] * global.resolution.y) - VerticalWhiteSpace;

    PageLength = global.printable.y - MarginSpace;

    x = (int) floor(global.printscale * (double) global.resolution.y / 72.0 + 0.5);      /* dots */
    printscale = (double) x / (double) PRINTFONT->height;


    i = y = 0;

    for (pp = f->paragraph, lp = pp->lines; ; ) {
        if (lp->prev == NULL) {
            /* Recalculate the line spacing for each paragraph */
            LineHeight = (int) floor((double) (PRINTFONT->height + PRINTFONT->spacing) * printscale * ((double) pp->spacemulti) / 100.0 + 0.5);
        }

        if ((pp == f->paragraph && lp == pp->lines) ||
             y + LineHeight >= PageLength || pp->pagebreak) {

            if (f->pages == NULL) {
                f->pages = (POSITION far *) BlockAlloc(BLOCKSIZE);
            } else {
                size = SegHeapGetSize(f->pages);
                if (i > (size / sizeof(POSITION))) {
                    f->pages = BlockRealloc(f->pages, size + BLOCKSIZE);
                    if (f->pages == NULL) return (FALSE);
                }
            }

            if (PARAOF(f->pages[i]) != pp || LINEOF(f->pages[i]) != lp)
                changed = TRUE;

            PARAOF(f->pages[i]) = pp;
            LINEOF(f->pages[i]) = lp;
            POSOF(f->pages[i]) = 0;
            i++;
            y = 0;
            //sprintf(buffer, "Repagenating File... Page %d", i);
            //StatusMessage(buffer);
        }
        y += LineHeight;
        if (!MoveToNextLine (&pp, &lp, +1)) break;
    }

    if (f->nr_pages != i) changed = TRUE;

    f->nr_pages = i;
    f->repage = FALSE;

    //sprintf(buffer, "Total %d page%s.", i, (i > 1) ? "s" : "");
    //StatusMessage(buffer);

    return (changed);
}



// PageNum <> 0: Find the page
// PageNum == 0: Find the page containing pos

int FindPage (FILEOPTIONS *f, int PageNum, POSITION *pos)
{
    int i;
    PARAGRAPH far *pp;
    ONELINE far *lp;


    if (f->pages == NULL) return (0);
    POSOF(*pos) = 0;

    /* Find absolute page */

    if (PageNum > 0) {
        if (PageNum > f->nr_pages) return (0);
        *pos = f->pages[PageNum-1];
        return (PageNum);
    }


    /* Find page number */

    for (i = 0, pp = f->paragraph, lp = pp->lines; ; ) {
        if (PARAOF(f->pages[i]) == pp && LINEOF(f->pages[i]) == lp) i++;
        if (pp == PARAOF(*pos) && lp == LINEOF(*pos)) {
            *pos = f->pages[i-1];
            return (i);
        }
        if (!MoveToNextLine (&pp, &lp, +1)) break;
    }
    return (0);
}



BOOL CALLBACK PrintAbortProc (HDC hdc, short nCode)
{
    MSG msg;

    while (!UserAbort && PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
        if (PrintDialogBoxHwnd == NULL || !IsDialogMessage(PrintDialogBoxHwnd, &msg)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }
    return (!UserAbort);
}



BOOL CALLBACK PrintDlgProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG:
            CenterDialogBox(hwnd);
            return (TRUE);

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDCANCEL:
                    UserAbort = TRUE;
                    return (TRUE);
            }
            break;
    }

    return (FALSE);
}



static void PrintErrorMessage (int error, char *action)
{
    switch (error) {
        case SP_ERROR:      ErrorMessage(global.hwnd, "Cannot %s!", action); break;
        case SP_APPABORT:
        case SP_USERABORT:  break;
        case SP_OUTOFDISK:  ErrorMessage(global.hwnd, "Out of Disk Space!  Cannot continue printing."); break;
        case SP_OUTOFMEMORY:ErrorMessage(global.hwnd, "Out of Memory!  Cannot continue printing."); break;
    }
}



void DoPrinterSetup ()
{
    int r;
    PRINTDLG pd;
    FILEOPTIONS *f;

    /* Initialize the common dialog box */

printer_setup:

    if (global.hdevnames != NULL && global.hdevmode != NULL) {
        if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
            (void) PageOrientation(global.pagesetup.landscape, TRUE);
        } else {
            (void) PageOrientation(global.active->pagesetup.landscape, TRUE);
        }
    }

    memset(&pd, 0, sizeof(PRINTDLG));

    pd.lStructSize = sizeof(PRINTDLG);
    pd.hwndOwner = global.hwnd;
    pd.hDevMode = GlobalDuplicate(global.hdevmode);
    pd.hDevNames = GlobalDuplicate(global.hdevnames);
    pd.Flags = PD_PRINTSETUP;

    r = PrintDlg(&pd);


    if (r == 0) {
        LONG errcode;

        errcode = CommDlgExtendedError();
        if (errcode == PDERR_DEFAULTDIFFERENT) {
            /* The default printer has changed */
            if (pd.hDevMode != NULL) GlobalFree(pd.hDevMode);
            if (pd.hDevNames != NULL) GlobalFree(pd.hDevNames);
            if (global.hdevmode != NULL) GlobalFree(global.hdevmode);
            if (global.hdevnames != NULL) GlobalFree(global.hdevnames);
            global.hdevmode = global.hdevnames = NULL;
            InitPrinting();
            goto printer_setup;
        } else if (errcode != CDERR_GENERALCODES &&
                   pd.hDevMode != NULL && pd.hDevNames != NULL) {

            ErrorMessage(global.hwnd, "%s!\n(Error code = %ld 0x%lx)",
                            PrinterErrorText(errcode), errcode, errcode);
            if (pd.hDevMode != NULL) GlobalFree(pd.hDevMode);
            if (pd.hDevNames != NULL) GlobalFree(pd.hDevNames);
        }
        return;
    }

    if (global.hdevmode != NULL) GlobalFree(global.hdevmode);
    if (global.hdevnames != NULL) GlobalFree(global.hdevnames);
    global.hdevmode = pd.hDevMode;
    global.hdevnames = pd.hDevNames;

    if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
        global.pagesetup.landscape =
            PageOrientation(global.pagesetup.landscape, FALSE);
    } else {
        if (global.active->pagesetup.landscape !=
            PageOrientation(global.active->pagesetup.landscape, FALSE))
                global.active->changed = TRUE;

        global.active->pagesetup.landscape =
            PageOrientation(global.active->pagesetup.landscape, FALSE);
    }

    for (f = fileoptions; f != NULL; f = f->next) {
        ChangePrinterOrFonts(f, f == fileoptions, TRUE);
    }
}



#define LEFTMG           (MarginDots[0])
#define RIGHTMG          (MarginDots[1])
#define TOPMG            (MarginDots[2])
#define BOTTOMMG         (MarginDots[3])

void PrintFile (FILEOPTIONS *f)
{
    PRINTDLG pd;
    int i, j, k, h, n, r, bmsize;
    int x, y, x1, y1;
    int num, temp, ScreenLen;
    int PrintWidth, PrintHeight;
    int PrintLeading, PrintSpacing;
    int ThisParaSpacing, ThisParaHeight;    
    int PrintCell, PageNum, NrPages;
    int errorcode;
    BOOL ChangeView = FALSE;
    JUSTIFICATION Jmode;
    double ratio, printscale;
    double PrintRatioX, PrintRatioY;
    KUNIT far *up;
    HDC hdc, hdcmem;
    FARPROC AbortPrinting;
    HFONT hfont;
    HCURSOR hcursor;
    RECT rect;
    PARAGRAPH far *pp;
    ONELINE far *lp;
    POSITION StartPos, NextPos;
    HBITMAP hbitmap;
    TEXTMETRIC tm;
    BYTE far *cbufp;
    char buffer[MAXLINELEN];
    int WhiteSpaces[2], MarginDots[4];
    int Widths[256];


    RepagenateFile(f);

    /* Initialize the common dialog box */

print_dlg:

    if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
        (void) PageOrientation(global.pagesetup.landscape, TRUE);
    } else {
        (void) PageOrientation(global.active->pagesetup.landscape, TRUE);
    }

    memset(&pd, 0, sizeof(PRINTDLG));

    pd.lStructSize = sizeof(PRINTDLG);
    pd.hwndOwner = global.hwnd;
    pd.hDevMode = GlobalDuplicate(global.hdevmode);
    pd.hDevNames = GlobalDuplicate(global.hdevnames);
    pd.hDC = NULL;
    pd.Flags = PD_RETURNDC | PD_USEDEVMODECOPIES;
    pd.nFromPage = 1;
    pd.nToPage = f->nr_pages;
    pd.nMinPage = 1;
    pd.nMaxPage = f->nr_pages;
    pd.nCopies = 1;

    r = PrintDlg(&pd);


    if (r == 0) {
        LONG errcode;

        errcode = CommDlgExtendedError();
        if (errcode == PDERR_DEFAULTDIFFERENT) {
            /* The default printer has changed */
            if (pd.hDevMode != NULL) GlobalFree(pd.hDevMode);
            if (pd.hDevNames != NULL) GlobalFree(pd.hDevNames);
            if (global.hdevmode != NULL) GlobalFree(global.hdevmode);
            if (global.hdevnames != NULL) GlobalFree(global.hdevnames);
            global.hdevmode = global.hdevnames = NULL;
            InitPrinting();
            goto print_dlg;
        } else if (errcode == PDERR_NODEVICES) {
            ErrorMessage(global.hwnd, "There are no printers installed!");
            if (pd.hDevMode != NULL) GlobalFree(pd.hDevMode);
            if (pd.hDevNames != NULL) GlobalFree(pd.hDevNames);
        } else if (errcode != CDERR_GENERALCODES) {
            ErrorMessage(global.hwnd, "%s!\n(Error code = %ld 0x%lx)",
                            PrinterErrorText(errcode), errcode, errcode);
            if (pd.hDevMode != NULL) GlobalFree(pd.hDevMode);
            if (pd.hDevNames != NULL) GlobalFree(pd.hDevNames);
        }
        return;
    }


    if (global.hdevmode != NULL) GlobalFree(global.hdevmode);
    if (global.hdevnames != NULL) GlobalFree(global.hdevnames);
    global.hdevmode = pd.hDevMode;
    global.hdevnames = pd.hDevNames;


    hdc = pd.hDC;

    if (hdc == NULL) {
        ErrorMessage(global.hwnd, "Cannot get printer driver!");
        return;
    }

    if (!(GetDeviceCaps(hdc, RASTERCAPS) & RC_BITBLT)) {
        DeleteDC(hdc);
        ErrorMessage(global.hwnd, "The chosen printer is not capable of printing graphics!");
        return;
    }

    if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
        global.pagesetup.landscape =
            PageOrientation(global.pagesetup.landscape, FALSE);
    } else {
        global.active->pagesetup.landscape =
            PageOrientation(global.active->pagesetup.landscape, FALSE);
    }

    ChangePrinterOrFonts(f, TRUE, TRUE);

    if (r == 0) return;

    if (pd.Flags & PD_SELECTION) {
        PrintFrom = PrintTo = -1;
    } else if (pd.Flags & PD_PAGENUMS) {
        PrintFrom = pd.nFromPage;
        PrintTo = pd.nToPage;
        if (PrintFrom < 0) PrintFrom = 0;
        if (PrintTo < PrintFrom) PrintTo = PrintFrom;
    } else {
        PrintFrom = PrintTo = 0;
    }


    hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
    ShowCursor(TRUE);


    /* Initialize stuff */

    x = (int) floor(global.printscale * (double) global.resolution.y / 72.0 + 0.5);      /* dots */
    printscale = (double) x / (double) PRINTFONT->height;

    SetBkMode(hdc, TRANSPARENT);

    ratio = (double) global.resolution.x / (double) global.resolution.y;
    PrintHeight = floor((double) PRINTFONT->height * printscale + 0.5);
    PrintWidth = floor((double) PrintHeight * ratio + 0.5);
    PrintLeading = floor((double) PRINTFONT->leading * printscale * ratio + 0.5);
    PrintSpacing = floor((double) PRINTFONT->spacing * printscale + 0.5);
    PrintCell = PrintWidth + PrintLeading;

    PrintRatioX = ((double) PrintWidth / (double) PRINTFONT->width);
    PrintRatioY = ((double) PrintHeight / (double) PRINTFONT->height);

    hbitmap = CreateBitmap(PRINTFONT->width, PRINTFONT->height, 1, 1, NULL);
    hdcmem = CreateCompatibleDC(hdc);
    SetMapMode(hdcmem, GetMapMode(hdc));

    /* Load the ASCII font */

    hfont = SelectAsciiFont(hdc, DefAsciiFont.facename, DefAsciiFont.size, &tm);
    SelectObject(hdc, hfont);
    GetCharWidth(hdc, 0, 255, Widths);

    WhiteSpaces[0] = ((global.paper[0] * global.resolution.x) - global.printable.x) / 2;
    WhiteSpaces[1] = ((global.paper[1] * global.resolution.y) - global.printable.y) / 2;

    MarginDots[0] = (f->pagesetup.margins[0] * global.resolution.x) - WhiteSpaces[0];
    MarginDots[1] = (f->pagesetup.margins[1] * global.resolution.x) - WhiteSpaces[0];
    MarginDots[2] = (f->pagesetup.margins[2] * global.resolution.y) - WhiteSpaces[1];
    MarginDots[3] = (f->pagesetup.margins[3] * global.resolution.y) - WhiteSpaces[1];


    /* Set the abort procedures */

    AbortPrinting = MakeProcInstance(PrintAbortProc, hInstance);

    EnableWindow(global.hwnd, FALSE);
    EnableWindow(global.clienthwnd, FALSE);

    UserAbort = PrintError = FALSE;
    PrintDialogBoxHwnd = CreateDialog(hInstance, "Printing", f->parent, PrintDlgProc);
    AnimateIcon(GetDlgItem(PrintDialogBoxHwnd, 4211), "X_PRINTING", 5, 1.0);

    /* Reformat from draft view? */

    if (global.draftview) {
        SetDlgItemText(PrintDialogBoxHwnd, 4201, "Reformatting File...");
        AbortPrinting(hdc, 0);
        global.draftview = FALSE;
        ReformatFile(f);
        ChangeView = TRUE;
    }


    /* Pagenate */

    SetDlgItemText(PrintDialogBoxHwnd, 4201, "Repagenating File...");
    RepagenateFile(f);

    AbortPrinting(hdc, 0);

    if (PrintFrom < 0 && PrintTo < 0) {     /* Current Page */
        StartPos = f->current;
        PageNum = FindPage(f, 0, &StartPos);        /* Find current page */
        if (PageNum == 0) {
            ErrorMessage(global.hwnd, "There %s less than %d page%s in this document!",
                            (PrintFrom > 1) ? "are" : "is", PrintFrom, (PrintFrom > 1) ? "s" : "");
            goto CleanUp;
        }
        NrPages = 1;
    } else {
        if (PrintFrom <= 0) {
            PARAOF(StartPos) = f->paragraph;
            LINEOF(StartPos) = f->paragraph->lines;
            PageNum = 1;
        } else {
            i = FindPage(f, PrintFrom, &StartPos);
            if (i == 0) {
                ErrorMessage(global.hwnd, "There %s less than %d page%s in this document!",
                                (PrintFrom > 1) ? "are" : "is", PrintFrom, (PrintFrom > 1) ? "s" : "");
                goto CleanUp;
            }
            PageNum = PrintFrom;
        }

        if (PrintTo <= 0) {
            NrPages = -1;
        } else if (PrintFrom <= 0) {
            NrPages = PrintTo;
        } else {
            NrPages = PrintTo - PrintFrom + 1;
            if (NrPages <= 0) NrPages = 1;
        }
    }

    i = FindPage(f, PageNum + 1, &NextPos);
    if (i == 0) PARAOF(NextPos) = NULL;


    if (!(f->filename[0])) {
        strcpy(buffer, "Printing ");
        GetWindowText(f->parent, buffer + strlen(buffer), MAXLINELEN);
        sprintf(buffer + strlen(buffer), "\n\n%s", global.printer);
    } else {
        sprintf(buffer, "Printing %s\n\n%s", f->filename, global.printer);
    }

    SetDlgItemText(PrintDialogBoxHwnd, 4201, buffer);

    SetDlgItemText(PrintDialogBoxHwnd, 4202, "Initializing printer...");
    AbortPrinting(hdc, 0);

    sprintf(buffer, "%s %s - ", PROGNAME, JWPVERSION);
    GetWindowText(f->parent, buffer + strlen(buffer), MAXLINELEN);

#ifdef USEWIN32
    {
        DOCINFO dinfo;

        dinfo.cbSize = sizeof(DOCINFO);
        dinfo.lpszDocName = buffer;
        dinfo.lpszOutput = NULL;

        errorcode = StartDoc(hdc, &dinfo);
    }
#else
    errorcode = Escape(hdc, STARTDOC, strlen(buffer), buffer, NULL);

#endif

    if (errorcode <= 0) {
        PrintErrorMessage(errorcode, "start printing");
        goto CleanUp;
    }

#ifdef USEWIN32
    SetAbortProc(hdc, (ABORTPROC) AbortPrinting);
#else
    Escape(hdc, SETABORTPROC, 0, (LPSTR) AbortPrinting, NULL);
#endif USEWIN32


    /* Do the true printing */

Loop:

    sprintf(buffer, "Currently Printing Page %d...", PageNum);
    SetDlgItemText(PrintDialogBoxHwnd, 4202, buffer);
    AbortPrinting(hdc, 0);

    /* Process each band */

    errorcode = Escape(hdc, NEXTBAND, 0, NULL, (LPSTR) &rect);
    if (errorcode < 0) {
        PrintError = TRUE;
        PrintErrorMessage(errorcode, "continue printing");
    }

    SelectObject(hdc, hfont);
    SetBkMode(hdc, TRANSPARENT);

    while (!PrintError && !UserAbort && !IsRectEmpty(&rect)) {
        /* Header / Footer ? */

        if (!AbortPrinting(hdc, 0)) break;

        for (h = 0; h < 2; h++) {
            KANJI kbuffer[MAXLINELEN];

            if (f->nofirstpage && PageNum <= 1) break;

            switch (h) {
                default: continue;

                case 0: y = TOPMG - PrintHeight - PrintSpacing;
                        if (rect.top >= y) continue;
                        y -= PrintHeight;
                        if (f->lrheader) {
                            num = (PageNum % 2 == 0) ? 1 : 0;
                        } else {
                            num = 0;
                        }
                        break;

                case 1: y = global.printable.y - BOTTOMMG + PrintHeight + PrintSpacing;
                        if (rect.bottom <= y) continue;
                        if (f->lrheader) {
                            num = (PageNum % 2 == 0) ? 2 : 3;
                        } else {
                            num = 2;
                        }
                        break;
            }

            for (i = 0; i < NRHEADERS; i++) {
                if (f->header[num][i] == NULL) continue;
                if (!AbortPrinting(hdc, 0)) break;

                BuildHeaderString (f, f->header[num][i], PageNum, kbuffer);

                n = kanjilen(kbuffer);
                if (n <= 0) continue;

                for (j = k = 0; kbuffer[j]; j++) {
                    if (!AbortPrinting(hdc, 0)) break;
                    if (ISKANJI(kbuffer[j])) {
                        k += PrintCell;
                    } else {
                        k += Widths[kbuffer[j]];
                        if (buffer[j+1] && ISKANJI(kbuffer[j+1])) k += PrintLeading;
                    }
                }

                switch (i) {
                    default:
                    case 0: x = LEFTMG; break;
                    case 1: x = (f->linelen * PrintCell - PrintLeading - k) / 2 + LEFTMG; break;
                    case 2: x = f->linelen * PrintCell - PrintLeading - k + LEFTMG; break;
                }

                for (j = 0; kbuffer[j]; j++) {
                    if (!AbortPrinting(hdc, 0)) break;
                    if (ISKANJI(kbuffer[j])) {
                        r = Jis2Index(kbuffer[j], PRINTFONT->holes);
                        if (r < 0) r = Jis2Index(BADKANJI, PRINTFONT->holes);
                        bmsize = GetKanjiBitmap(PRINTFONT, r, &cbufp, f->vertical && !WithinSet(kbuffer[j], NoRotate));
                        SetBitmapBits(hbitmap, (DWORD) bmsize, cbufp);
                        SelectObject(hdcmem, hbitmap);

                        if (f->vertical && WithinSet(kbuffer[j], NeedOffset)) {
                            CalcRotateOffset(cbufp, PRINTFONT->height, bmsize, &x1, &y1);
                            x1 = (int) ((double) x1 * PrintRatioX);
                            y1 = (int) ((double) y1 * PrintRatioY);
                        } else {
                            x1 = y1 = 0;
                        }

                        StretchBlt(hdc, x - x1, y - y1, PrintWidth, PrintHeight, hdcmem, 0, 0,
                                   PRINTFONT->width, PRINTFONT->height, SRCAND);
                        x += PrintCell;
                    } else {
                        TextOut(hdc, x, y + PrintHeight - tm.tmHeight, kbuffer + j, 1);
                        x += Widths[kbuffer[j]];
                        if (kbuffer[j+1] && ISKANJI(kbuffer[j+1])) x += PrintLeading;
                    }
                }
            }
        }

        if (UserAbort) break;

        pp = PARAOF(StartPos);
        lp = LINEOF(StartPos);
        y = TOPMG;

        for (;;) {
            if (pp == PARAOF(NextPos) && lp == LINEOF(NextPos)) break;

            if (y > rect.bottom) break;

            if (pp->pagebreak) {
                if (!MoveToNextLine(&pp, &lp, +1)) break;
                continue;
            }

            if (!AbortPrinting(hdc, 0)) break;

            ThisParaHeight = floor((double) PRINTFONT->height * printscale * ((double) pp->spacemulti) / 100.0 + 0.5);
            ThisParaSpacing = floor((double) PRINTFONT->spacing * printscale * ((double) pp->spacemulti) / 100.0 + 0.5);

            if (y + ThisParaHeight + ThisParaSpacing < rect.top) {
                if (!MoveToNextLine(&pp, &lp, +1)) break;
                y += ThisParaHeight + ThisParaSpacing;
                continue;
            }

            x = ((lp == pp->lines) ? pp->firstindent : pp->leftindent) * PrintCell + LEFTMG;
            ScreenLen = ((lp == pp->lines) ? pp->firstindent : pp->leftindent) * BASEWIDTH(f);

            up = pp->text + lp->position;

            for (i = 0; i < lp->length; i++) {
                if (x > rect.right) break;

                if (!AbortPrinting(hdc, 0)) break;

                if (ISKANJI(up->kanji) && x + PrintCell < rect.left) {
                    x += PrintCell;
                    ScreenLen += BASEWIDTH(f);
                    up++;
                    continue;
                }

                if (up->kanji == 0x2121 /* Blank */) {
                    x += PrintCell;
                    ScreenLen += BASEWIDTH(f);
                    up++;
                } else if (ISKANJI(up->kanji)) {
                    r = Jis2Index(up->kanji, PRINTFONT->holes);
                    if (r < 0) r = Jis2Index(BADKANJI, PRINTFONT->holes);
                    bmsize = GetKanjiBitmap(PRINTFONT, r, &cbufp, f->vertical && !WithinSet(up->kanji, NoRotate));
                    SetBitmapBits(hbitmap, (DWORD) bmsize, cbufp);
                    SelectObject(hdcmem, hbitmap);

                    if (f->vertical && WithinSet(up->kanji, NeedOffset)) {
                        CalcRotateOffset(cbufp, PRINTFONT->height, bmsize, &x1, &y1);
                        x1 = (int) ((double) x1 * PrintRatioX);
                        y1 = (int) ((double) y1 * PrintRatioY);
                    } else {
                        x1 = y1 = 0;
                    }

                    StretchBlt(hdc, x - x1, y - y1, PrintWidth, PrintHeight, hdcmem, 0, 0,
                               PRINTFONT->width, PRINTFONT->height, SRCAND);
                    x += PrintCell;
                    ScreenLen += BASEWIDTH(f);
                    up++;
                } else {
                    /* Get the text string */

                    SetTextAlign(hdc, TA_LEFT | TA_BOTTOM | TA_NOUPDATECP);
                    SetTextJustification(hdc, 0, 0);

                    ratio = (double) global.resolution.y / (double) global.resolution.x;
                    ratio *= global.dispscale / printscale;

                    for (j = k = 0; i < lp->length; i++, j++, up++) {
                        if (ISKANJI(up->kanji)) break;
                        if (up->kanji == '\t') break;

                        buffer[j] = LOBYTE(up->kanji);

                        //temp = FontCharWidth(up->kanji, 1) - FontDisplayError(up->kanji, pp, lp);
                        temp = FontCharWidth(up->kanji, 1) - FontDisplayError(up->kanji, i);

                        ScreenLen += temp;
                        k += Widths[up->kanji];
                    }
                    buffer[j] = 0;

                    Jmode = global.shortjust;

                    if (i >= lp->length || i <= j) {    /* Open ended */
                        Jmode = global.longjust;
                        if (Jmode == J_RIGHT || Jmode == J_CENTER) Jmode = J_JUSTIFY;
                        if (lp->next == NULL && Jmode == J_JUSTIFY) Jmode = J_LEFT;

                        if (buffer[0] && i >= lp->length) {
                            for (temp = strlen(buffer) - 1; temp >= 0 && buffer[temp] == ' '; temp--);
                            buffer[++temp] = '\0';
                            k -= (j - temp) * Widths[' '];
                        }
                    } else if (up->kanji != '\t') {     /* In between */
                        Jmode = J_LEFT;
                    }

#ifdef USEWIN32
                    {
                        SIZE tsize;

                        GetTextExtentPoint32(hdc, buffer, strlen(buffer), &tsize);
                        k = tsize.cx;
                    }
#else
                    k = LOWORD(GetTextExtent(hdc, buffer, strlen(buffer)));
#endif USEWIN32

                    /* Now, get the amount of tab */

                    if (Jmode != J_LEFT) {
                        if (i >= lp->length) {
                            ScreenLen = (f->linelen - pp->rightindent) * BASEWIDTH(f) - global.leading;
                            r = (f->linelen - pp->rightindent) * PrintCell - PrintLeading - x - k + LEFTMG;
                        } else {
                            ScreenLen += BASEWIDTH(f) - (ScreenLen % BASEWIDTH(f));
                            r = (ScreenLen / BASEWIDTH(f)) * PrintCell - PrintLeading - x - k + LEFTMG;
                        }
                    } else if (up->kanji == '\t') {
                        ScreenLen += BASEWIDTH(f) - (ScreenLen % BASEWIDTH(f));
                        r = (ScreenLen / BASEWIDTH(f)) * PrintCell - PrintLeading - x - k + LEFTMG;
                    } else {
                        r = 0;
                    }

                    if (r < 0) Jmode = J_JUSTIFY;

                    switch (Jmode) {
                        LeftText:
                        case J_LEFT:
                            if (!AbortPrinting(hdc, 0)) break;
                            TextOut(hdc, x, y + PrintHeight, buffer, strlen(buffer));
                            x += k + r;
                            break;

                        RightText:
                        case J_RIGHT:
                            if (!AbortPrinting(hdc, 0)) break;
                            TextOut(hdc, x + r, y + PrintHeight, buffer, strlen(buffer));
                            x += k + r;
                            break;

                        CenterText:
                        case J_CENTER:
                            TextOut(hdc, x + r/2, y + PrintHeight, buffer, strlen(buffer));
                            x += k + r;
                            break;

                        case J_JUSTIFY:
                            /* Count the nuber of spaces */
                            for (temp = n = 0; buffer[temp]; temp++)
                                if (buffer[temp] == ' ') n++;

                            if (n > 0) {
                                if (r > Widths[' ']) {
                                    if (buffer[0] != ' ' && i > j) {
                                        x += Widths[' ']/2;
                                        r -= Widths[' ']/2;
                                    }
                                    if (i < lp->length) {
                                        SetTextJustification(hdc, r - Widths[' ']/2, n);
                                    } else {
                                        SetTextJustification(hdc, r, n);
                                    }
                                } else {
                                    SetTextJustification(hdc, r, n);
                                }
                                if (!AbortPrinting(hdc, 0)) break;
                                TextOut(hdc, x, y + PrintHeight, buffer, strlen(buffer));
                                x += k + r;
                            } else if (i >= lp->length) {
                                if (i <= j) goto LeftText; else goto RightText;
                            } else {
                                if (i <= j) goto LeftText; else goto CenterText;
                            }
                            break;
                    }

                    if (i >= lp->length) break;
                    if (ISKANJI(up->kanji)) {
                        i--;
                        x += PrintLeading;
                        ScreenLen += global.leading;
                    }
                    if (up->kanji == '\t') {
                        x += PrintLeading;
                        up++;
                    }
                }
            }

            y += ThisParaHeight + ThisParaSpacing;

            if (!AbortPrinting(hdc, 0)) break;

            if (!MoveToNextLine(&pp, &lp, +1)) break;
        }

        errorcode = Escape(hdc, NEXTBAND, 0, NULL, (LPSTR) &rect);
        if (errorcode < 0) {
            PrintError = TRUE;
            PrintErrorMessage(errorcode, "continue printing");
            break;
        }
    }


    /* Another page? */

    if (!UserAbort && !PrintError && (NrPages < 0 || NrPages > 1) && PARAOF(NextPos) != NULL) {
        if (NrPages > 0) NrPages--;
        StartPos = NextPos;
        PageNum++;
        i = FindPage(f, PageNum + 1, &NextPos);
        if (i == 0) PARAOF(NextPos) = NULL;
        goto Loop;
    }

    /* Finishes the accounting */

    if (!PrintError) {
        if (UserAbort) {
            SetDlgItemText(PrintDialogBoxHwnd, 4202, "Printing Aborted!");
            AbortPrinting(hdc, 0);
#ifdef USEWIN32
            AbortDoc(hdc);
#else
            Escape(hdc, ABORTDOC, 0, NULL, NULL);
#endif
        } else {
            SetDlgItemText(PrintDialogBoxHwnd, 4202, "Printing Completed.");
            AbortPrinting(hdc, 0);

#ifdef USEWIN32
            EndDoc(hdc);
#else
            Escape(hdc, ENDDOC, 0, NULL, NULL);
#endif
        }
    }


CleanUp:

    /* Clean up... */

    if (ChangeView) {
        SetDlgItemText(PrintDialogBoxHwnd, 4201, "Reformatting File...");
        AbortPrinting(hdc, 0);
        global.draftview = TRUE;
        ReformatFile(f);
    }

    AnimateIcon(NULL, NULL, 0, 0.0);

    EnableWindow(global.hwnd, TRUE);
    EnableWindow(global.clienthwnd, TRUE);

    DestroyWindow(PrintDialogBoxHwnd);
    PrintDialogBoxHwnd = NULL;

    FreeProcInstance(AbortPrinting);

    ShowCursor(FALSE);
    SetCursor(hcursor);

    if (hdcmem != NULL) DeleteDC(hdcmem);
    if (hdc != NULL) DeleteDC(hdc);
    if (hbitmap != NULL) DeleteObject(hbitmap);
    if (hfont != NULL) DeleteObject(hfont);
}



BOOL CALLBACK PageSetupProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    static MEASUREMENT OldMeasure;

    switch (message) {
        case WM_INITDIALOG: {
            int i;
            MEASUREINFO *mp;
            char buffer[MAXLINELEN];
            char format[30];

            SetDlgItemText(hwnd, 4201, global.printer);

            /* Page orientation */

            if (global.hdevnames != NULL && global.hdevmode != NULL) {
                SendDlgItemMessage(hwnd, 4211, CB_ADDSTRING, 0, (LONG) "Portrait");
                SendDlgItemMessage(hwnd, 4211, CB_ADDSTRING, 0, (LONG) "Landscape");
                if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
                    SendDlgItemMessage(hwnd, 4211, CB_SETCURSEL, global.pagesetup.landscape ? 1 : 0, 0L);
                } else {
                    SendDlgItemMessage(hwnd, 4211, CB_SETCURSEL, global.active->pagesetup.landscape ? 1 : 0, 0L);
                }
            } else {
                EnableWindow(GetDlgItem(hwnd, 4211), FALSE);
            }

            /* Measurements */

            for (i = 0; Measurements[i].name != NULL; i++) {
                SendDlgItemMessage(hwnd, 4212, CB_ADDSTRING, 0, (LONG) Measurements[i].name);
            }
            mp = Measurements + global.measure;
            SendDlgItemMessage(hwnd, 4212, CB_SETCURSEL, global.measure, 0L);

            /* Line length */

            if (global.active != NULL) {
                SetDlgItemInt(hwnd, 4213, global.active->linelen, TRUE);
            }

            /* Margins */

            sprintf(format, "%%%d.%dlf%%s", mp->decimal + 2, mp->decimal);

            if (global.active != NULL) {
                for (i = 0; i < 4; i++) {
                    sprintf(buffer, format, global.active->pagesetup.margins[i] * mp->conversion, "");
                    SetDlgItemText(hwnd, 4221 + i, buffer);
                }
            } else {
                for (i = 0; i < 4; i++) {
                    sprintf(buffer, format, global.pagesetup.margins[i] * mp->conversion, "");
                    SetDlgItemText(hwnd, 4221 + i, buffer);
                }
            }

            for (i = 0; i < 4; i++) {
                sprintf(buffer, format, global.pagesetup.margins[i] * mp->conversion, mp->symbol);
                SetDlgItemText(hwnd, 4231 + i, buffer);
            }

            OldMeasure = global.measure;

            TrapHelpInDialogBoxes(hwnd, "PageSetupDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND: {
            int i, j;
            double num, temp, minmargins[4];
            BOOL NeedReformat;
            MEASUREINFO *mp, *mp1;
            FILEOPTIONS *f;
            char buffer[MAXLINELEN];
            char format[30];

            minmargins[0] = minmargins[1] = ((global.paper[0] * global.resolution.x) - global.printable.x) / 2;
            minmargins[2] = minmargins[3] = ((global.paper[1] * global.resolution.y) - global.printable.y) / 2;

            switch (WINCMDID) {
                case 4211:      /* Orientation */
                    if (global.hdevnames == NULL || global.hdevmode == NULL)
                        break;
                    /* Drops into next case... */

                case 4221:      /* Margins */
                case 4222: {
                    double marginsbackup[2];
                    double *mp;

                    if (global.active == NULL) mp = global.pagesetup.margins;
                    else mp = global.active->pagesetup.margins;

                    for (i = 0; i < 2; i++) {
                        GetDlgItemText(hwnd, 4221 + i, buffer, MAXLINELEN);
                        num = atof(buffer);

                        temp = (double) minmargins[i] / (double) global.resolution.x;
                        if (num < temp) num = temp;
                        num /= Measurements[global.measure].conversion;

                        marginsbackup[i] = mp[i];
                        mp[i] = num;
                    }

                    SetDlgItemInt(hwnd, 4213, CalcLineLength(global.active), TRUE);

                    for (i = 0; i < 2; i++) {
                        mp[i] = marginsbackup[i];
                    }
                    break;
                }

                case 4241:      /* Make default */
                    if (YesNo(hwnd, "Make these margin settings the default?") == IDYES) {
                        mp = Measurements + OldMeasure;
                        sprintf(format, "%%%d.%dlf%%s", mp->decimal + 2, mp->decimal);

                        for (i = 0; i < 4; i++) {
                            GetDlgItemText(hwnd, 4221 + i, buffer, MAXLINELEN);
                            num = atof(buffer);

                            switch (i) {
                                case 0:
                                case 1: temp = (double) minmargins[i] / (double) global.resolution.x;
                                        break;
                                case 2:
                                case 3: temp = (double) minmargins[i] / (double) global.resolution.x;
                                        break;
                            }

                            if (num < temp) num = temp;

                            sprintf(buffer, format, num, mp->symbol);
                            SetDlgItemText(hwnd, 4231 + i, buffer);
                        }
                    }
                    return (TRUE);

                case 4212: {    /* Measurement */
                    int x;
                    double printscale;

                    j = SendDlgItemMessage(hwnd, 4212, CB_GETCURSEL, 0, 0L);
                    if (j == CB_ERR) return (TRUE);
                    if (j == OldMeasure) return (TRUE);

                    mp = Measurements + OldMeasure;
                    mp1 = Measurements + j;
                    sprintf(format, "%%%d.%dlf%%s", mp1->decimal + 2, mp1->decimal);

                    for (i = 0; i < 4; i++) {
                        GetDlgItemText(hwnd, 4221 + i, buffer, MAXLINELEN);
                        num = atof(buffer);

                        switch (i) {
                            case 0:
                            case 1: temp = (double) minmargins[i] / (double) global.resolution.x;
                                    break;
                            case 2:
                            case 3: temp = (double) minmargins[i] / (double) global.resolution.x;
                                    break;
                        }

                        if (num < temp) num = temp;

                        sprintf(buffer, format, num * mp1->conversion / mp->conversion, "");
                        SetDlgItemText(hwnd, 4221 + i, buffer);
                    }
                    for (i = 0; i < 4; i++) {
                        GetDlgItemText(hwnd, 4231 + i, buffer, MAXLINELEN);
                        num = atof(buffer);

                        switch (i) {
                            case 0:
                            case 1: temp = (double) minmargins[i] / (double) global.resolution.x;
                                    break;
                            case 2:
                            case 3: temp = (double) minmargins[i] / (double) global.resolution.x;
                                    break;
                        }

                        if (num < temp) num = temp;

                        sprintf(buffer, format, num * mp1->conversion / mp->conversion, mp1->symbol);
                        SetDlgItemText(hwnd, 4231 + i, buffer);
                    }

                    OldMeasure = j;

                    x = (int) floor(global.printscale * (double) global.resolution.y / 72.0 + 0.5);      /* dots */
                    printscale = (double) x / (double) PRINTFONT->height;

                    num = (double) (PRINTFONT->width + PRINTFONT->leading) * printscale;
                    num *= (double) global.resolution.x / (double) global.resolution.y;
                    SetDlgItemInt(hwnd, 4213, (int) floor((double) global.printable.x / num), TRUE);
                    return (TRUE);
                }

                case IDOK:
                    NeedReformat = FALSE;

                    /* Orientation */

                    if (global.hdevnames != NULL && global.hdevmode != NULL) {
                        i = SendDlgItemMessage(hwnd, 4211, CB_GETCURSEL, 0, 0L);
                        if (i != CB_ERR) {
                            switch (i) {
                                case 0: i = FALSE; break;
                                case 1: i = TRUE; break;
                            }
                            if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
                                if (i != global.pagesetup.landscape) {
                                    OptionsChanged = TRUE;
                                    global.pagesetup.landscape = i;
                                }
                            } else {
                                if (i != global.active->pagesetup.landscape) {
                                    global.active->pagesetup.landscape = i;
                                    global.active->changed = global.active->repage = TRUE;
                                    (void) PageOrientation(i, TRUE);
                                    ChangePrinterOrFonts(global.active, FALSE, TRUE);
                                }
                            }
                        }
                    }

                    /* Measurement */

                    i = SendDlgItemMessage(hwnd, 4212, CB_GETCURSEL, 0, 0L);
                    if (i != CB_ERR && i != global.measure) {
                        OptionsChanged = TRUE;
                        global.measure = i;
                    }

                    /* Margins */

                    if (global.active != NULL) {
                        for (i = 0; i < 4; i++) {
                            GetDlgItemText(hwnd, 4221 + i, buffer, MAXLINELEN);
                            num = atof(buffer);

                            switch (i) {
                                case 0:
                                case 1: temp = (double) minmargins[i] / (double) global.resolution.x;
                                        break;
                                case 2:
                                case 3: temp = (double) minmargins[i] / (double) global.resolution.x;
                                        break;
                            }

                            if (num < temp) num = temp;
                            num /= Measurements[global.measure].conversion;

                            if (fabs(num - global.active->pagesetup.margins[i]) > 0.01) {
                                global.active->pagesetup.margins[i] = num;
                                NeedReformat = TRUE;
                                OptionsChanged = TRUE;
                            }
                        }
                        if (NeedReformat) {
                            global.active->linelen = CalcLineLength(global.active);
                            ReformatFile(global.active);
                            InvalidateRect(global.active->hwnd, NULL, TRUE);
                        }
                    }

                    for (i = 0; i < 4; i++) {
                        GetDlgItemText(hwnd, 4231 + i, buffer, MAXLINELEN);
                        num = atof(buffer);

                        switch (i) {
                            case 0:
                            case 1: temp = (double) minmargins[i] / (double) global.resolution.x;
                                    break;
                            case 2:
                            case 3: temp = (double) minmargins[i] / (double) global.resolution.x;
                                    break;
                        }

                        if (num < temp) num = temp;
                        num /= Measurements[global.measure].conversion;

                        if (fabs(num - global.pagesetup.margins[i]) > 0.01) {
                            global.pagesetup.margins[i] = num;
                            OptionsChanged = TRUE;
                        }
                    }

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, NeedReformat);
                    return (TRUE);

                case IDCANCEL:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, 0);
                    return (TRUE);
            }
            break;
        }
        break;
    }
    return (FALSE);
}
