#include "jwp.h"
#include "radlook.h"

/*

   kanji lookup via natural join on radicals and stroke count

   basic idea:

        a "radical set" is a set of characters which contain a
        particular radical.
        a "stroke set" is a set of characters which are composed of
        a particular number of strokes.
        a "kanji set" is either a radical set or stroke set.
        a "merge compare" is analagous to a merge sort on two arrays,
        except that identical elements are stored into another array
        and different elements are discarded.

        Step 1: sort the radical set and stroke sets.

       Step 2: allocate two arrays, one for the size of the
                smallest group and one for the largest.  Since we're
                  doing a natural join, we know that the most number of
                  common elements is at most the size of the smallest
                  array.

      Step 3: starting with the smallest size kanji set, do a
                 "merge compare" against the next largest size kanji
                  set, storing identical elements back into the smaller
                  array.

        Step 4: repeat step 3 until all arrays have been compared or
                 the original array is empty.

      Step 5: extract characters found in all kanji sets into list
                 box of dialog.

    since a character can have at most 23 strokes, we declare an array of
    30 (MAXKANJISETS) to handle all possible radical groups (23 + 1 for
    stroke count group + a few more for safety)

    assumptions about radical lookup data file structure:

        'Escape-code-less' JIS codes in rad2knji.dat and stroknji.dat
      are sorted in order of increasing JIS code value.


        Given a radical or stroke no. 'n', the record for radical or
      stroke 'n' in rad2knji.idx is at the offset 'n'*4.  The first two
      bytes of each record represent the offset into the data file for the
      record, and the last two bytes indicate the set size.

*/
#define RADINDEXNAME    global.radkanjiidx
#define RADDATANAME     global.radkanjidat
#define STROKEINDEXNAME global.stroknjiidx
#define STROKEDATANAME  global.stroknjidat

#define MAXKANJISETS 30

#define MAXSTROKES 30

// must be consistent with record structure of index file
struct kanjiSet {
    unsigned INT16 offset;
    INT16 count;
};

static struct kanjiSet kanjiSets[MAXKANJISETS];
static int lastKanjiSet;

static HFILE radDataFile;
static HFILE radIndexFile;
static HFILE strokeDataFile;
static HFILE strokeIndexFile;

JISCode far *joinSet = NULL;
static BOOL globalErr = TRUE;

/*
   initializes the KanjiSets array and opens global file handles to
    data files.
*/
BOOL openKanjiQuery(void)
{
    OFSTRUCT of;

    radDataFile = OpenFile(RADDATANAME, &of, OF_READ);
    radIndexFile = OpenFile(RADINDEXNAME, &of, OF_READ);
    strokeDataFile = OpenFile(STROKEDATANAME, &of, OF_READ);
    strokeIndexFile = OpenFile(STROKEINDEXNAME, &of, OF_READ);

    if (radDataFile == HFILE_ERROR) {
        ErrorMessage(global.hwnd, "Unable to load %s !", RADDATANAME);
        closeKanjiQuery();
        return FALSE;
    }
    if (radIndexFile == HFILE_ERROR) {
        ErrorMessage(global.hwnd, "Unable to load %s!", RADINDEXNAME);
        closeKanjiQuery();
        return FALSE;
    }
    if (strokeDataFile == HFILE_ERROR) {
        ErrorMessage(global.hwnd, "Unable to load %s!", STROKEDATANAME);
        closeKanjiQuery();
        return FALSE;
    }
    if (strokeIndexFile == HFILE_ERROR) {
        ErrorMessage(global.hwnd, "Unable to load %s!", STROKEINDEXNAME);
        closeKanjiQuery();
        return FALSE;
    }

    return TRUE;
}

void closeKanjiQuery(void)
{
    _lclose(radDataFile);
    _lclose(radIndexFile);
    _lclose(strokeDataFile);
    _lclose(strokeIndexFile);
}

/*

   Uses Rad2Knji.idx to determine the count of kanji characters
   containing the specified radical and the beginning of that set's
   offset in the Rad2Knji.dat file.

    Inserts the positive occurence count value into the kanjiSets array.

*/
void addRadicalSet(int radicalNo)
{
    struct kanjiSet ks;

    if (++lastKanjiSet < MAXKANJISETS) {
        _llseek(radIndexFile, (long) (radicalNo - 1) * 4, 0);
        _lread(radIndexFile, &ks, sizeof(struct kanjiSet));
        kanjiSets[lastKanjiSet] = ks;
    }
}

/*

   Uses StroKnji.idx to determine the count of kanji characters
    containing the specified number of strokes and the beginning of
    that set's offset in the StroKnji.dat file.

   Inserts the negative occurence count value into the kanjiSets array.
    This is so that when we merge the arrays, we'll know to look in
    the stroke data file (instead of the radical data file) for the
    file offset.

*/
void addStrokeSet(int nStrokes)
{
    struct kanjiSet ks;

    if ((nStrokes < 1) || (nStrokes > MAXSTROKES)) {
        globalErr = TRUE;
        return;
    }

    if (++lastKanjiSet < MAXKANJISETS) {
        _llseek(strokeIndexFile, (long) (nStrokes - 1) * 4, 0);
        _lread(strokeIndexFile, &ks, sizeof(struct kanjiSet));
        ks.count *= -1;
        kanjiSets[lastKanjiSet] = ks;
    }
}

// must be called before every doKanjiJoin()!
void initKanjiJoin(void)
{
    lastKanjiSet = -1;
    globalErr = FALSE;

}



/*

  sortJoinSet() will sort the kanji set by stroke count, taking
  info from KINFO.DAT.

*/

static int kinfofd;

static int kanjiStrokesCompare (const void *s1, const void *s2)
{
    switch (global.rlookupsort) {
        case 2: {
            int cat1, cat2;
            KANJI c1, c2;

            c1 = *((JISCode *) s1);
            c2 = *((JISCode *) s2);

            if (HIBYTE(c1) < 0x30) cat1 = 0x00ff;
            else if (0x30 <= HIBYTE(c1) && HIBYTE(c1) <= 0x4f) cat1 = 0x0030;
            else if (HIBYTE(c1) > 0x4f) cat1 = 0x0050;
            else cat1 = 0x00ff;

            if (HIBYTE(c1) < 0x30) cat2 = 0x00ff;
            else if (0x30 <= HIBYTE(c2) && HIBYTE(c2) <= 0x4f) cat2 = 0x0030;
            else if (HIBYTE(c2) > 0x4f) cat2 = 0x0050;
            else cat2 = 0x00ff;

            if (cat1 != cat2) return (cat1 - cat2);
            /* Drop into next case... */
        }
        case 0:
        case 1: return (GetKinfo(kinfofd, *((JISCode *) s1), 'S') -
                        GetKinfo(kinfofd, *((JISCode *) s2), 'S'));
    }
}


static BOOL sortJoinSet(INT32 nr_elements)
{
    JISCode *ljoinset;
    HCURSOR hCursor;

    if (nr_elements <= 0) return (FALSE);

    kinfofd = OpenKinfoFile(NULL);
    if (kinfofd < 0) return (FALSE);

    hCursor = SetCursor(LoadCursor (NULL, IDC_WAIT));
    ShowCursor(TRUE);

#ifdef USEWIN32
    qsort(joinSet, nr_elements, sizeof(JISCode), kanjiStrokesCompare);
#else
    ljoinset = (JISCode *) MemAlloc(nr_elements * sizeof(JISCode) + 5);
    if (ljoinset != NULL) {
        _fmemcpy(ljoinset, joinSet, nr_elements * sizeof(JISCode));
        qsort(ljoinset, nr_elements, sizeof(JISCode), kanjiStrokesCompare);
        _fmemcpy(joinSet, ljoinset, nr_elements * sizeof(JISCode));
        FreeMem(ljoinset);
    }
#endif USEWIN32

    _lclose(kinfofd);

    ShowCursor(FALSE);
    SetCursor(hCursor);
    return (TRUE);
}



/*

   Executes what is basically a join query on the sets of kanji
    characters in the kanjiSets array.  Kanji characters found
    in all sets are left in the global joinSet array when the function
    returns.  The function returns the number of characters in
    the array, or -1 if an unexpected error occurs.

*/
int doKanjiJoin()
{

    int minCount;
    struct kanjiSet temp;
    int currSet;
    int tSetSize;
    int tSetElm;
    int tSetIdx;
    int mSetSize;
    int mSetElm;
    int swapSet;
    JISCode far *mergeSet;

    // since the most number of radicals a character can contain is less
    // than MAXKANJISETS - 1 kanji sets, don't bother doing the merge if
    // there can't be any matching characters.
    // also, note that we use MAXKANJISETS - 1 because we're reserving a
    // kanjiSet element for a stroke count record.
    // also return 0 if no search criteria were added.
    if ((lastKanjiSet > MAXKANJISETS - 1) || (lastKanjiSet < 0) ||
         (globalErr)) {
        return 0;
    }

    // first sort the kanjiSets array in order of increasing set size.
    // for simplicity and because the expected number of sets is relatively
   // small, we'll use selection sort.

   for (currSet = 0; currSet < lastKanjiSet; currSet++) {
        minCount = currSet;
        for (swapSet = currSet+1; swapSet <= lastKanjiSet; swapSet++) {
            if (abs(kanjiSets[swapSet].count) < abs(kanjiSets[minCount].count)) {
                minCount = swapSet;
            }
        }
        temp = kanjiSets[currSet];
        kanjiSets[currSet] = kanjiSets[minCount];
        kanjiSets[minCount] = temp;
    }

   // allocate temporary space to merge arrays.
    if (joinSet != NULL) {
        FreeBlock(joinSet);
    }
    joinSet = (JISCode far *) BlockAlloc(abs(kanjiSets[0].count) * sizeof(JISCode));

    if (joinSet == NULL) {
        return -1;
    }

    mergeSet = (JISCode far *) BlockAlloc(abs(kanjiSets[lastKanjiSet].count) * sizeof(JISCode));

    if (mergeSet == NULL) {
        FreeBlock(joinSet);
        joinSet = NULL;
        return -1;
    }

    // merge compare all the kanji sets, leaving the common elements
    // in the first kanjiSet array element.

    tSetSize = abs(kanjiSets[0].count);

   // read the elements of the kanji set to compare into mergeSet
    _llseek(kanjiSets[0].count > 0 ? radDataFile : strokeDataFile,
        kanjiSets[0].offset, 0);
    _lread(kanjiSets[0].count > 0 ? radDataFile : strokeDataFile,
             joinSet, abs(kanjiSets[0].count) * sizeof(INT16));

    for (currSet = 1; currSet <= lastKanjiSet; currSet++) {

        // read the elements of the kanji set to compare into mergeSet
        _llseek(kanjiSets[currSet].count > 0 ? radDataFile : strokeDataFile,
            kanjiSets[currSet].offset, 0);
        _lread(kanjiSets[currSet].count > 0 ? radDataFile : strokeDataFile,
                mergeSet, abs(kanjiSets[currSet].count) * sizeof(INT16));


        // now merge compare the joinSet and mergeSet
        tSetElm = 0;
       tSetIdx = 0;
        mSetSize = abs(kanjiSets[currSet].count);
        mSetElm = 0;

        while ((mSetElm < mSetSize) && (tSetElm < tSetSize)) {

            if (joinSet[tSetElm] == mergeSet[mSetElm]) {
                joinSet[tSetIdx++] = joinSet[tSetElm];
                tSetElm++;
                mSetElm++;
                continue;
            }

            while ((mSetElm < mSetSize) &&
                (joinSet[tSetElm] > mergeSet[mSetElm])) {
                mSetElm++;
            }

            while ((tSetElm < tSetSize) &&
                    (joinSet[tSetElm] < mergeSet[mSetElm])) {
                tSetElm++;
            }

        }

        tSetSize = tSetIdx;

        if (tSetSize == 0) {
            break;
        }
    }

    FreeBlock(mergeSet);

    if (global.rlookupsort != 0) sortJoinSet(tSetSize);

    return tSetSize;

}
