#include "jwp.h"
#include "radlook.h"
#include <assert.h>

/*
  +-----------------------------------------------------------+
  |                                                           |
  | Matches:                   kanji list                     |
  |            <               scroll bars                  > |
  |                                                           |
  |  Search   Reset   Get Info   Insert to File   Forget it   |<-cmnd buttons
  |                                                           |
  |   Strokes...  Any   10+ 20+ 0  1  2  3  4  5  6  7  8  9  |<-stroke btns
  |                                                           |
  |  Radicals... 1 <radical buttons...>                       |
  |                <radical buttons...>                       |
  ...              <radical buttons...>                     ...
  +-----------------------------------------------------------+

  The space between logically grouped sections (e.g. cmnd buttons and stroke
  buttons) is called the section gap.

  The number of radical button columns determines the width of the dialog.
*/

#define RADBTN_NCOLS        24   // # of columns of radical buttons
#define RADFIRSTROWOFFSET    6   // offset to col. of first button of 1st row

#define RADBTN_WIDTH        24   // width of radical buttons
#define RADBTN_HEIGHT       24   // height of radical buttons
#define RADBITMAPSIZE       16   // dimension of radical bitmap
#define RADBTN_VSPACE        0   // space between rows of radical buttons
#define RADBTN_HSPACE        0   // space between cols of radical buttons

#define CMDBTN_WIDTH       100   // width of command buttons
#define CMDBTN_HEIGHT       28   // height of command buttons

#define STRBTN_WIDTH        22   // width of stroke count buttons
#define STRBTN_HEIGHT       16   // height of stroke count buttons

#define MTCBOX_WIDTH        80   // width of match count box
#define MTCBOX_HEIGHT       20   // height of match count box
#define MTCBOX_HORZGAP       5   // gap between match area and kanji list

#define SECTIONGAP           5   // vertical space between sections of dlg

#define DLGBOX_HMARG         5   // left (and right) margin of dialog box
#define DLGBOX_VMARG         5   // top (and bottom) margin of dialog box

static int dlgWidth;             // width of dialog box
static int dlgHeight;            // height of dialog box

static HBITMAP buttonDown;
static HBITMAP buttonUp;
static HBITMAP buttonSel;

static HBITMAP strokeUp;
static HBITMAP strokeDown;
static HBITMAP strokeUp2;
static HBITMAP strokeDown2;

static HBRUSH grayBrush = NULL;

#define NRADBUTTONS 241
#define RADID_OFFSET 4000

static HBITMAP kanjiBitmap = NULL;
static HWND radButtons[NRADBUTTONS];
static int radState[NRADBUTTONS];

// check all uses of these defines before changing their values.
#define BTN_UP   0
#define BTN_SEL  1
#define BTN_DOWN 2
#define BTN_TEMP 3     // redundant radical flag (see ID_DOSEARCH below)

#define NRADPAIRS 18             // # of paired radicals

typedef struct {
    int firstrad;
    int secondrad;
} RADPAIRTYPE;

// ** IMPORTANT **
// Radicals are paired such that the *first* radical of each pair is
// the true lookup radical in the index file.  In other words, when
// we add radicals to the search, if we see a second radical we ignore
// it.  We can ignore it because its paired radical is the real lookup
// radical.  This is necessary because the offset values into the data
// file for alternate radicals aren't 'fixed up'/relocated at data
// creation time (because of the way the original data is structured).
// So we do the relocation here.  WAKATTA, WAKATTA, HEN DA KEDO...

static RADPAIRTYPE radPair[NRADPAIRS] = {
   {186,  32}, {193,  33}, { 51,  52}, { 54,  55}, { 64,  65}, { 64,  66},
   {164,  73}, { 74,  75}, { 80,  81}, { 86,  87}, { 97,  96}, {118, 113},
   {149, 115}, {154, 116}, {146, 140}, {236, 199}, {238, 219}, {239, 226} };


#define NSTROKEBUTTONS 13        // 0-9, 10+, 20+, Any
#define STROKEID_OFFSET 3000
static HWND strokeButtons[NSTROKEBUTTONS];
static int strokeState[NSTROKEBUTTONS];

#define ID_INSFILE  2000
#define ID_DONE     2001
#define ID_GETINFO  2002
#define ID_DOSEARCH 2003
#define ID_RESET    2004
#define ID_MATCHES1 2005
#define ID_RADLBL   2006
#define ID_STROKELBL 2007
#define ID_KANJILIST 2008
#define ID_MATCHES2 2009
static HWND insToFileButton;
static HWND doneButton;
static HWND getInfoButton;
static HWND doSearchButton;
static HWND resetButton;
static HWND matchBox1;
static HWND matchBox2;
static HWND strokeLabel;
static HWND radLabel;
static HWND kanjiList;

static int nMatches;

#define LABELID_OFFSET  1000
#define NLABELS 17
static HWND strokeLbls[NLABELS];
static HBITMAP labelBitmap[NLABELS];
// indicates radical numbers before which a stroke count label appears
static int labelPos[NLABELS] = {
    0, 6, 34, 74, 117, 142, 171, 190, 200,
    211, 220, 227, 231, 235, 237, 238, 240 };


long FAR PASCAL _export RadLookupProc (HWND, UINT, UINT, LONG) ;
long FAR PASCAL _export StrokesProc(HWND, UINT, WPARAM, LPARAM);
long FAR PASCAL _export CursorMovementProc(HWND, UINT, WPARAM, LPARAM);
LONG FAR PASCAL RadLookupListProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
extern BOOL FAR PASCAL KanjiInfoProc (HWND, UINT, WPARAM, LPARAM);

static char szMenuName[]  = "MainMenu";
static char szClassName[] = "SampleApp";
static char szTitle[]     = "Radical Lookup";

//static int emulateDlg = FALSE;
//extern int emulateDlg;
//static HWND hKanjiLookupDlg;
extern HWND hKanjiLookupDlg;

static FARPROC lpfnDefButtonProc, lpfnDefStrokesProc;
static int currentStrokes;
static int currentRadical;  // used to track current tabstop in radical group

#define WM_CHANGEHIGHLIGHT WM_USER

#define KANJILISTBORDER   BORDERSPACE

static int start;         // index to first/leftmost kanji in kanjiList window
static int NrInList;      // total # kanji viewable in kanjiList window
static int NrSelected;    // current element of joinSet selected
static int EdgeGap;


long FAR PASCAL _export RadLookupProc (HWND hwnd, UINT message, UINT wParam,
LPARAM lParam)
{
    HDC              hdcmem;
    LPDRAWITEMSTRUCT lpdis;
    int vPos, hPos;
    int colNo, rowNo;
    int kListHeight;
    int buttonNo, labelNo;
    int radNo;
    int nStrokes;
    char buf[20];
    HCURSOR hCursor;
    HMENU sysMenu;

    switch (message)
    {

        case WM_CREATE:

            // load hourglass cursor
            hCursor = SetCursor(LoadCursor (NULL, IDC_WAIT));
            ShowCursor(TRUE);

            sysMenu = GetSystemMenu(hwnd, FALSE);
            DeleteMenu(sysMenu, SC_SIZE, MF_BYCOMMAND);
            DeleteMenu(sysMenu, SC_MAXIMIZE, MF_BYCOMMAND);
            DeleteMenu(sysMenu, SC_TASKLIST, MF_BYCOMMAND);
            DeleteMenu(sysMenu, 5, MF_BYPOSITION);

            grayBrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));

            dlgWidth = 2*DLGBOX_HMARG +
                        RADBTN_NCOLS * (RADBTN_WIDTH + RADBTN_HSPACE)
                          - RADBTN_HSPACE;

            // create the match count output static area

            vPos = DLGBOX_VMARG;

           matchBox1 = CreateWindow("static", "&Matches: ",
                WS_CHILD | WS_VISIBLE | SS_CENTER,
                DLGBOX_HMARG, vPos,
                MTCBOX_WIDTH, MTCBOX_HEIGHT,
                hwnd, (HMENU) ID_MATCHES1, hInstance, NULL);

           matchBox2 = CreateWindow("button", "",
                WS_CHILD | WS_VISIBLE | BS_OWNERDRAW | WS_DISABLED,
                DLGBOX_HMARG, vPos+MTCBOX_HEIGHT,
                MTCBOX_WIDTH, MTCBOX_HEIGHT,
                hwnd, (HMENU) ID_MATCHES2, hInstance, NULL);

            hPos = DLGBOX_HMARG + MTCBOX_WIDTH + MTCBOX_HORZGAP;

            kListHeight = GetSystemMetrics(SM_CYHSCROLL) +
                SYSFONT->height + 1 + 2*KANJILISTBORDER;

            // kanji list scroll bar
            kanjiList = CreateWindow(PROGNAME " Radical kanji lookup list", "",
                WS_CHILD | WS_VISIBLE | WS_HSCROLL | WS_GROUP | WS_TABSTOP,
                hPos, vPos,
                dlgWidth - hPos - DLGBOX_HMARG, kListHeight,
                hwnd, (HMENU) ID_KANJILIST, hInstance, NULL);

            hPos = (dlgWidth/2) - (5*(CMDBTN_WIDTH + 15)/2);
            vPos += kListHeight + SECTIONGAP;

            // create the dialog command buttons section

         doSearchButton = CreateWindow("button", (global.autorlookup ? "Auto &Search" : "&Search!"),
                BS_PUSHBUTTON | WS_CHILD | WS_VISIBLE | WS_GROUP | WS_TABSTOP,
                hPos, vPos,
                CMDBTN_WIDTH, CMDBTN_HEIGHT,
                hwnd, (HMENU) ID_DOSEARCH, hInstance, NULL);

            hPos += CMDBTN_WIDTH + 15;

         if (global.autorlookup) {
            EnableWindow(doSearchButton, FALSE);
         }

         resetButton = CreateWindow("button", "&Reset",
                BS_PUSHBUTTON | WS_CHILD | WS_VISIBLE | WS_GROUP | WS_TABSTOP,
                hPos, vPos,
                CMDBTN_WIDTH, CMDBTN_HEIGHT,
                hwnd, (HMENU) ID_RESET, hInstance, NULL);

            hPos += CMDBTN_WIDTH + 15;

            getInfoButton = CreateWindow("button", "&Get Info...",
                BS_PUSHBUTTON | WS_CHILD | WS_VISIBLE | WS_GROUP | WS_DISABLED | WS_TABSTOP,
                hPos, vPos,
                CMDBTN_WIDTH, CMDBTN_HEIGHT,
                hwnd, (HMENU) ID_GETINFO, hInstance, NULL);

            hPos += CMDBTN_WIDTH + 15;

            insToFileButton = CreateWindow("button", "&Insert to file",
                BS_PUSHBUTTON | WS_CHILD | WS_VISIBLE | WS_GROUP | WS_DISABLED | WS_TABSTOP,
                hPos, vPos,
                CMDBTN_WIDTH, CMDBTN_HEIGHT,
                hwnd, (HMENU) ID_INSFILE, hInstance, NULL);

            hPos += CMDBTN_WIDTH + 15;

            doneButton = CreateWindow("button", "&Done",
                BS_DEFPUSHBUTTON | WS_CHILD | WS_VISIBLE | WS_GROUP | WS_TABSTOP,
                hPos, vPos,
                CMDBTN_WIDTH, CMDBTN_HEIGHT,
                hwnd, (HMENU) ID_DONE, hInstance, NULL);

         vPos += CMDBTN_HEIGHT + SECTIONGAP;

            // stroke count buttons section

            // the "Any", "10+", and "20+" buttons are double-width
            // leave a 1 button-width gap between button groups

            hPos = (dlgWidth/2) - (24*STRBTN_WIDTH/2);

            // create the stroke count group label

           strokeLabel = CreateWindow("static", "&Total # of strokes:",
                WS_CHILD | WS_VISIBLE | SS_LEFT | WS_GROUP,
                hPos, vPos,
                6*STRBTN_WIDTH, STRBTN_HEIGHT,
                hwnd, (HMENU) ID_STROKELBL, hInstance, NULL);

            hPos += 7*STRBTN_WIDTH;

            strokeButtons[12] = CreateWindow("button", "",
                                     WS_CHILD | WS_VISIBLE | BS_OWNERDRAW | WS_TABSTOP,
                                          hPos, vPos,
                                          2*STRBTN_WIDTH, STRBTN_HEIGHT,
                                          hwnd, (HMENU) (STROKEID_OFFSET+12), hInstance, NULL);
            strokeState[12] = BTN_DOWN;

            hPos += 3*STRBTN_WIDTH;

            lpfnDefStrokesProc = (FARPROC) GetWindowLong(strokeButtons[12], GWL_WNDPROC);
            SetWindowLong(strokeButtons[12], GWL_WNDPROC, (LONG) StrokesProc);

            strokeButtons[10] = CreateWindow("button", "",
                                     WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
                                          hPos, vPos,
                                          2*STRBTN_WIDTH, STRBTN_HEIGHT,
                                          hwnd, (HMENU) (STROKEID_OFFSET+10), hInstance, NULL);
            strokeState[10] = BTN_UP;

            SetWindowLong(strokeButtons[10], GWL_WNDPROC, (LONG) StrokesProc);

            hPos += 2*STRBTN_WIDTH;

            strokeButtons[11] = CreateWindow("button", "",
                                     WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
                                          hPos, vPos,
                                          2*STRBTN_WIDTH, STRBTN_HEIGHT,
                                          hwnd, (HMENU) (STROKEID_OFFSET+11), hInstance, NULL);
            strokeState[11] = BTN_UP;

            SetWindowLong(strokeButtons[11], GWL_WNDPROC, (LONG) StrokesProc);

            hPos += 2*STRBTN_WIDTH;

            for (buttonNo = 0; buttonNo < 10; buttonNo++) {
//              sprintf(buf, "%d", buttonNo);
                strokeButtons[buttonNo] = CreateWindow("button", "", //buf,
                                        WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
                                              hPos, vPos,
                                              STRBTN_WIDTH, STRBTN_HEIGHT,
                                          hwnd, (HMENU) (STROKEID_OFFSET+buttonNo), hInstance, NULL);
               SetWindowLong(strokeButtons[buttonNo], GWL_WNDPROC, (LONG) StrokesProc);

                strokeState[buttonNo] = BTN_UP;

            hPos += STRBTN_WIDTH;

            }

            currentStrokes = 12;

            vPos += STRBTN_HEIGHT + SECTIONGAP;

            // radical buttons section

            hPos = DLGBOX_HMARG + (RADBTN_WIDTH+RADBTN_HSPACE) / 2;

           // the reset button is 4 times the width of a radical button and
            // 4 pixels shorter in total height

            radLabel = CreateWindow("static", "Radical &Elements:",
                           WS_CHILD | WS_VISIBLE | SS_LEFT | WS_GROUP,
                              hPos, vPos+2,
                              (RADBTN_WIDTH+RADBTN_HSPACE)*5, RADBTN_HEIGHT-4,
                              hwnd, (HMENU) ID_RADLBL, hInstance, NULL);

            buttonNo = 0;
            labelNo = 0;
            rowNo = 0;
            colNo = RADFIRSTROWOFFSET;  // makes it flush with lower-right corner

            // create the radical stroke index labels and pushbuttons

            kanjiBitmap = LoadBitmap(hInstance, "RADICAL_BITMAPS");

            while (buttonNo < NRADBUTTONS) {
                if ((labelNo < NLABELS) && (buttonNo == labelPos[labelNo])) {
                strokeLbls[labelNo] = CreateWindow("button", "",
                                              WS_CHILD | WS_VISIBLE | BS_OWNERDRAW | WS_DISABLED,
                                                     DLGBOX_HMARG-1+colNo*(RADBTN_WIDTH+RADBTN_HSPACE),
                                                     vPos+rowNo*(RADBTN_HEIGHT+RADBTN_VSPACE),
                                                     RADBTN_WIDTH, RADBTN_HEIGHT,
                                                     hwnd, (HMENU) (labelNo+LABELID_OFFSET), hInstance, NULL);

                    sprintf(buf, "STROKELBL_%02d", (labelNo+1));
                    labelBitmap[labelNo] = LoadBitmap(hInstance, buf);

                    labelNo++;
                    colNo++;
                    if (colNo > (RADBTN_NCOLS - 1)) {
                        colNo = 0;
                        rowNo++;
                    }
                }

                radButtons[buttonNo] = CreateWindow ("button", "",
                                        WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
                                              DLGBOX_HMARG-1+colNo*(RADBTN_WIDTH+RADBTN_HSPACE),
                                              vPos+rowNo*(RADBTN_HEIGHT+RADBTN_VSPACE),
                                              RADBTN_WIDTH, RADBTN_HEIGHT,
                                          hwnd, (HMENU) (buttonNo+RADID_OFFSET), hInstance, NULL);
                radState[buttonNo] = BTN_UP;


                if (buttonNo == 0) {
                    lpfnDefButtonProc = (FARPROC) GetWindowLong(radButtons[0], GWL_WNDPROC);
                    SetWindowLong(radButtons[0], GWL_STYLE, WS_CHILD | WS_VISIBLE | BS_OWNERDRAW | WS_TABSTOP);
                    currentRadical = 0;
                }
                SetWindowLong(radButtons[buttonNo], GWL_WNDPROC, (LONG) CursorMovementProc);
//              SetClassWord(radButtons[buttonNo], GCW_HBRBACKGROUND, NULL);

                buttonNo++;
                colNo++;
                if (colNo > (RADBTN_NCOLS - 1)) {
                    colNo = 0;
                    rowNo++;
                }
            }

            dlgHeight = vPos + rowNo*(RADBTN_HEIGHT+RADBTN_VSPACE) + DLGBOX_VMARG;
            MoveWindow(hwnd, 0, 0,
//              dlgWidth+2*GetSystemMetrics(SM_CXDLGFRAME),
                dlgWidth,
//              dlgHeight+2*GetSystemMetrics(SM_CYDLGFRAME)+
                dlgHeight+
                    GetSystemMetrics(SM_CYCAPTION),
                FALSE);

            TrapHelpInDialogBoxes(hwnd, "RlookupDlg", 1, TRUE);
            CenterDialogBox(hwnd);

            buttonUp = LoadBitmap(hInstance, "btn_up");
            buttonSel = LoadBitmap(hInstance, "btn_sel");
            buttonDown = LoadBitmap(hInstance, "btn_down");

            strokeUp = LoadBitmap(hInstance, "strokecountup");
            strokeDown = LoadBitmap(hInstance, "strokecountdown");
            strokeUp2 = LoadBitmap(hInstance, "strokecountup2");
            strokeDown2 = LoadBitmap(hInstance, "strokecountdown2");

            SetFocus(doneButton);

            ShowCursor(FALSE);
            SetCursor(hCursor);

            return 0 ;

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#else
        case WM_CTLCOLOR:
            if ((LOWORD(lParam) == strokeLabel) ||
                 (LOWORD(lParam) == radLabel) ||
                 (LOWORD(lParam) == matchBox1) ||
                 (LOWORD(lParam) == matchBox2)) {
                SetBkColor((HDC) wParam, GetSysColor(COLOR_BTNFACE));
               return (HBRUSH) grayBrush;
            }
            break;
#endif USEWIN32

        case WM_COMMAND: {
            WPARAM WinCmdIDCode = WINCMDID;

         if (WinCmdIDCode == ID_DONE) {
                UnhookHelpTraps(hwnd, TRUE);
                DestroyWindow(hwnd);
                return 0;
            }

            if (WinCmdIDCode == ID_RESET) {
                for (buttonNo = 0; buttonNo < NRADBUTTONS; buttonNo++) {
                    if (radState[buttonNo] == BTN_SEL) {
                        radState[buttonNo] = BTN_UP;
                        InvalidateRect(radButtons[buttonNo], NULL, TRUE);
                    }
                }
                for (buttonNo = 0; buttonNo < NSTROKEBUTTONS-1; buttonNo++) {
                    if (strokeState[buttonNo] == BTN_DOWN) {
                        strokeState[buttonNo] = BTN_UP;
                        InvalidateRect(strokeButtons[buttonNo], NULL, TRUE);
                    }
                }
                if (strokeState[12] == BTN_UP) {
                    strokeState[12] = BTN_DOWN;
                    InvalidateRect(strokeButtons[12], NULL, TRUE);
                }
                nMatches = 0;
                NrSelected = 0;
                InvalidateRect(matchBox2, NULL, TRUE);
                EnableWindow(getInfoButton, FALSE);
                EnableWindow(insToFileButton, FALSE);
                InvalidateRect(kanjiList, NULL, TRUE);
//              EnableWindow(kanjiList, FALSE);

                return 0;
            }

            if (WinCmdIDCode == ID_DOSEARCH) {
                int pairNo;

                if (!openKanjiQuery()) return 0;

                initKanjiJoin();

                // set flag to avoid setting redundant radicals as search criteria
                for (pairNo = 0; pairNo < NRADPAIRS; pairNo++) {
                    if (radState[radPair[pairNo].secondrad] == BTN_SEL) {
                        radState[radPair[pairNo].secondrad] = BTN_TEMP;
                    }
                }

                for (buttonNo = 0; buttonNo < NRADBUTTONS; buttonNo++) {
                    if (radState[buttonNo] == BTN_SEL) {
                        addRadicalSet(buttonNo+1);
                    }
                }
                // reset redundant radical state
                for (pairNo = 0; pairNo < NRADPAIRS; pairNo++) {
                    if (radState[radPair[pairNo].secondrad] == BTN_TEMP) {
                        radState[radPair[pairNo].secondrad] = BTN_SEL;
                    }
                }

                if (strokeState[12] != BTN_DOWN) {
                    nStrokes = 0;
                    for (buttonNo = 0; buttonNo < 10; buttonNo++) {
                        if (strokeState[buttonNo] == BTN_DOWN) {
                            nStrokes = buttonNo;
                            break;
                        }
                    }
                    if (strokeState[10] == BTN_DOWN) {   // 10+ down?
                        nStrokes += 10;
                    }
                    if (strokeState[11] == BTN_DOWN) {   // 20+ down?
                        nStrokes += 20;
                    }
                    if (nStrokes > 0) {  // user could have pressed only 0 button
                       addStrokeSet(nStrokes);
                    }
                }

                nMatches = doKanjiJoin();
                closeKanjiQuery();

                if (nMatches == -1) {
                    ErrorMessage(hwnd, "Not enough memory to perform search!");
                    return 0;
                }

                InvalidateRect(matchBox2, NULL, TRUE);

                if (nMatches > 0) {
                    EnableWindow(getInfoButton, TRUE);
//                  EnableWindow(kanjiList, TRUE);
                    if (nMatches < NrInList) {
                        SetScrollRange(kanjiList, SB_HORZ, 0,
                            (nMatches == 1) ? 1 : nMatches-1, TRUE);
                    } else {
                        SetScrollRange(kanjiList, SB_HORZ, NrInList, nMatches, TRUE);
                    }

                    if (!IsIconic(global.hwnd) && (curfile != NULL)) {
                        EnableWindow(insToFileButton, TRUE);
                    }
                } else {
                    MessageBeep(MB_ICONHAND);
                }

                start = 0;
                NrSelected = 0;
                InvalidateRect(kanjiList, NULL, TRUE);

                return 0;
            }

            if (WinCmdIDCode == ID_GETINFO) {
                HWND hack = GetFocus();
                SetKanjiInfoChar(joinSet[NrSelected]);
                DialogBox(hInstance, "KanjiInfo", hwnd, KanjiInfoProc);
                SetFocus(hack);
                return 0;
            }

            if (WinCmdIDCode == ID_INSFILE) {
                KANJI buf[2];
                FILEOPTIONS *f;

                f = global.active;

                if (CheckInputTextType(f, PARAOF(f->current)->text, buf)) {
                    buf[0] = joinSet[NrSelected];
                    buf[1] = 0;

                    TurnOffSelection(f);
                    UndoAddTyping(f, f->current, 1, FALSE);
                    InsertString(f, f->current, buf,
                        OP_REFORMAT | OP_UPDATE | OP_MOVESEL | OP_MOVETOEND);
                } else {
                    MessageBeep(MB_ICONHAND);
                }
                return 0;
            }

            // WARNING: major hack to handle the 3-set of redundant radicals!
           if ((WinCmdIDCode == 64+RADID_OFFSET) ||
                 (WinCmdIDCode == 65+RADID_OFFSET) ||
                 (WinCmdIDCode == 66+RADID_OFFSET)) {
                radNo = WinCmdIDCode - RADID_OFFSET;
                radState[radNo] = radState[radNo] == BTN_UP ? BTN_SEL : BTN_UP;
                InvalidateRect(radButtons[radNo], NULL, TRUE);
                if (radNo != 64) {
                    radState[64] = radState[64] == BTN_UP ? BTN_SEL : BTN_UP;
                    InvalidateRect(radButtons[64], NULL, TRUE);
                }
                if (radNo != 65) {
                    radState[65] = radState[65] == BTN_UP ? BTN_SEL : BTN_UP;
                    InvalidateRect(radButtons[65], NULL, TRUE);
                }
                if (radNo != 66) {
                    radState[66] = radState[66] == BTN_UP ? BTN_SEL : BTN_UP;
                    InvalidateRect(radButtons[66], NULL, TRUE);
                }

                if (global.autorlookup) SendMessage(hwnd, WM_COMMAND, ID_DOSEARCH, 0L);
                return 0;
            }

            if ((WinCmdIDCode >= RADID_OFFSET) && (WinCmdIDCode < RADID_OFFSET+NRADBUTTONS)) {
                radNo = WinCmdIDCode - RADID_OFFSET;
                radState[radNo] = radState[radNo] == BTN_UP ? BTN_SEL : BTN_UP;
                InvalidateRect(radButtons[radNo], NULL, TRUE);

                for (buttonNo = 0; buttonNo < NRADPAIRS; buttonNo++) {
                    if (radPair[buttonNo].firstrad == radNo) {
                        radState[radPair[buttonNo].secondrad] =
                            radState[radPair[buttonNo].secondrad] == BTN_UP ? BTN_SEL : BTN_UP;
                        InvalidateRect(radButtons[radPair[buttonNo].secondrad], NULL, TRUE);
                        break;
                    }
                    if (radPair[buttonNo].secondrad == radNo) {
                        radState[radPair[buttonNo].firstrad] =
                            radState[radPair[buttonNo].firstrad] == BTN_UP ? BTN_SEL : BTN_UP;
                        InvalidateRect(radButtons[radPair[buttonNo].firstrad], NULL, TRUE);
                        break;
                    }
                }

                if (global.autorlookup) SendMessage(hwnd, WM_COMMAND, ID_DOSEARCH, 0L);
                return 0;
            }

            if ((WinCmdIDCode >= STROKEID_OFFSET) &&
                 (WinCmdIDCode < STROKEID_OFFSET+NSTROKEBUTTONS)) {
                nStrokes = WinCmdIDCode - STROKEID_OFFSET;
                if (nStrokes == 12) {  // "Any" strokes button pushed
                    for (buttonNo = NSTROKEBUTTONS-1; buttonNo >= 0; buttonNo--) {
                        if (strokeState[buttonNo] == BTN_DOWN) {
                            strokeState[buttonNo] = BTN_UP;
                            InvalidateRect(strokeButtons[buttonNo], NULL, TRUE);
                        }
                    }
                    strokeState[12] = BTN_DOWN;
                    InvalidateRect(strokeButtons[12], NULL, TRUE);
                } else {
                    if (strokeState[12] == BTN_DOWN) {
                       // last state was Any (so all other buttons should be up!)
                        strokeState[12] = BTN_UP;
                       InvalidateRect(strokeButtons[12], NULL, TRUE);
                    }
                    if ((nStrokes == 10) || (nStrokes == 11)) {  // 10+
                        if (strokeState[nStrokes] == BTN_DOWN) {
                            strokeState[nStrokes] = BTN_UP;
                        } else {
                            strokeState[nStrokes] = BTN_DOWN;
                        }
                        InvalidateRect(strokeButtons[nStrokes], NULL, TRUE);
                    } else {  // one of [0-9] pressed
                        if (strokeState[nStrokes] == BTN_DOWN) {
                            strokeState[nStrokes] = BTN_UP;
                            InvalidateRect(strokeButtons[nStrokes], NULL, TRUE);
                        } else {
                            for (buttonNo = 0; buttonNo < 10; buttonNo++) {
                                if (strokeState[buttonNo] == BTN_DOWN) {
                                    strokeState[buttonNo] = BTN_UP;
                                    InvalidateRect(strokeButtons[buttonNo], NULL, TRUE);
                                }
                            }
                            strokeState[nStrokes] = BTN_DOWN;
                            InvalidateRect(strokeButtons[nStrokes], NULL, TRUE);
                        }
                    }

                    for (buttonNo = NSTROKEBUTTONS-1; buttonNo >= 0; buttonNo--) {
                        if (strokeState[buttonNo] == BTN_DOWN) {
                            break;
                        }
                    }
                    // if no buttons are down, then set the "Any" button
                    if (buttonNo == -1) {
                        strokeState[12] = BTN_DOWN;
                        InvalidateRect(strokeButtons[12], NULL, TRUE);
                    }
                }
                if (global.autorlookup) SendMessage(hwnd, WM_COMMAND, ID_DOSEARCH, 0L);
                return 0;
            }

            return 0;
        }

        case WM_SETFOCUS:
            if ((nMatches > 0) &&
                 (!IsIconic(global.hwnd)) && (curfile != NULL)) {
                EnableWindow(insToFileButton, TRUE);
            } else {
                EnableWindow(insToFileButton, FALSE);
            }
            SetFocus(doneButton);
            return 0;

/*
      case WM_GETMINMAXINFO:

            mmInfo = (MINMAXINFO FAR *) lParam;

            mmInfo->ptMaxSize.x = dlgWidth;
            mmInfo->ptMaxSize.y = dlgHeight;
            mmInfo->ptMaxPosition.x = 30;
            mmInfo->ptMaxPosition.y = 30;
            mmInfo->ptMinTrackSize.x = dlgWidth;
            mmInfo->ptMinTrackSize.y = dlgHeight;
            mmInfo->ptMaxTrackSize.x = dlgWidth;
            mmInfo->ptMaxTrackSize.y = dlgHeight;

            return 0;
*/
        case WM_DRAWITEM:

            lpdis = (LPDRAWITEMSTRUCT) lParam ;
//          lpdis->CtlID  is the button id

           // is the item the Matches box?
            if (lpdis->CtlID == ID_MATCHES2) {
                HBRUSH graybrush;

                graybrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
                FillRect(lpdis->hDC, &(lpdis->rcItem), graybrush);
                DeleteObject(graybrush);

                if (nMatches <= 0) {
                    strcpy(buf, "None");
                } else {
                    sprintf(buf, "%d", nMatches);
                }
                SetBkMode(lpdis->hDC, TRANSPARENT);
                SetTextAlign(lpdis->hDC, TA_CENTER);
                TextOut(lpdis->hDC, lpdis->rcItem.right/2, 0, buf, strlen(buf));
                return 0;
            }

            hdcmem = CreateCompatibleDC(lpdis->hDC);


            // is the item a stroke button?
            if ((lpdis->CtlID >= STROKEID_OFFSET) &&
                (lpdis->CtlID < STROKEID_OFFSET+NSTROKEBUTTONS)) {
                nStrokes = lpdis->CtlID - STROKEID_OFFSET;
                if (nStrokes == 12) {
                    strcpy(buf, "Any");
                } else if (nStrokes == 11) {
                    strcpy(buf, "20+");
                } else if (nStrokes == 10) {
                    strcpy(buf, "10+");
               } else {
                    sprintf(buf, "%d", nStrokes);
               }
                if (lpdis->itemState & ODS_SELECTED) {
//                  if (strokeState[nStrokes] == BTN_DOWN) {
//                      SelectObject(hdcmem, nStrokes < 10 ? strokeUp : strokeUp2);
//                  } else {
                        SelectObject(hdcmem, nStrokes < 10 ? strokeDown : strokeDown2);
//                     }
                } else {
                    if (strokeState[nStrokes] == BTN_DOWN) {
                        SelectObject(hdcmem, nStrokes < 10 ? strokeDown : strokeDown2);
                    } else {
                        SelectObject(hdcmem, nStrokes < 10 ? strokeUp : strokeUp2);
                   }
                }
                BitBlt(lpdis->hDC,
                    0, 0, lpdis->rcItem.right, lpdis->rcItem.bottom,
                    hdcmem, 0, 0, SRCCOPY);

                SetBkMode(lpdis->hDC, TRANSPARENT);

                if (lpdis->itemState & ODS_SELECTED) {
//                  if (strokeState[nStrokes] == BTN_DOWN) {
//                      TextOut(lpdis->hDC, nStrokes < 10 ? 6 : 10, 0, buf, strlen(buf));
//                  } else {
                        TextOut(lpdis->hDC, nStrokes < 10 ? 7 : 12, 2, buf, strlen(buf));
//                  }
                } else {
                    if (strokeState[nStrokes] == BTN_DOWN) {
                        TextOut(lpdis->hDC, nStrokes < 10 ? 7 : 11, 1, buf, strlen(buf));
                    } else {
                        TextOut(lpdis->hDC, nStrokes < 10 ? 6 : 10, 0, buf, strlen(buf));
                    }
                }

                DeleteDC(hdcmem);

                if (lpdis->itemState & ODS_FOCUS) {
                    if (strokeState[nStrokes] == BTN_DOWN) {
                        lpdis->rcItem.left   += 3;
                        lpdis->rcItem.top    += 3;
                        lpdis->rcItem.right  -= 1;
                        lpdis->rcItem.bottom -= 1;
                    } else {
                        lpdis->rcItem.left   += 2;
                        lpdis->rcItem.top    += 2;
                        lpdis->rcItem.right  -= 2;
                        lpdis->rcItem.bottom -= 2;
                    }
                    DrawFocusRect (lpdis->hDC, &lpdis->rcItem) ;
                }

                return 0;
            }

            // if the item is a label, just paint it in its static state
            if ((lpdis->CtlID >= LABELID_OFFSET) &&
                 (lpdis->CtlID < LABELID_OFFSET + 17)) {
                SelectObject(hdcmem, labelBitmap[lpdis->CtlID - LABELID_OFFSET]);
                BitBlt(lpdis->hDC,
                    0, 0, lpdis->rcItem.right, lpdis->rcItem.bottom,
                    hdcmem, 0, 0, SRCCOPY);

                DeleteDC(hdcmem);
                return 0;
            }

            // If a button is being selected, draw it in its down state
            if (lpdis->itemState & ODS_SELECTED) {
                SelectObject(hdcmem, buttonDown);
            } else {
                if (radState[lpdis->CtlID-RADID_OFFSET] == BTN_SEL) {
                    SelectObject(hdcmem, buttonSel);
                } else {
                    SelectObject(hdcmem, buttonUp);
                }
            }

            BitBlt(lpdis->hDC, 0, 0, lpdis->rcItem.right, lpdis->rcItem.bottom, hdcmem, 0, 0, SRCCOPY);

            SelectObject(hdcmem, kanjiBitmap);

            SetTextColor(lpdis->hDC, RGB(0, 0, 0));
            SetBkColor(lpdis->hDC, RGB(255, 255, 255));
            if (lpdis->itemState & ODS_SELECTED) {
                BitBlt(lpdis->hDC, 6, 6, RADBITMAPSIZE, RADBITMAPSIZE,
                        hdcmem, 0, (lpdis->CtlID - RADID_OFFSET) * RADBITMAPSIZE, SRCAND);
            } else if (radState[lpdis->CtlID-RADID_OFFSET] == BTN_SEL) {
                BitBlt(lpdis->hDC, 5, 5, RADBITMAPSIZE, RADBITMAPSIZE,
                        hdcmem, 0, (lpdis->CtlID - RADID_OFFSET) * RADBITMAPSIZE, SRCAND);
            } else {
                BitBlt(lpdis->hDC, 4, 4, RADBITMAPSIZE, RADBITMAPSIZE,
                        hdcmem, 0, (lpdis->CtlID - RADID_OFFSET) * RADBITMAPSIZE, SRCAND);
            }

            DeleteDC(hdcmem);

            if (lpdis->itemState & ODS_FOCUS) {
                if (lpdis->itemState & ODS_SELECTED) {
                    lpdis->rcItem.left   += 3;
                    lpdis->rcItem.top    += 3;
                    lpdis->rcItem.right  -= 1;
                    lpdis->rcItem.bottom -= 1;
                } else {
                    lpdis->rcItem.left   += 3;
                    lpdis->rcItem.top    += 3;
                    lpdis->rcItem.right  -= 3;
                    lpdis->rcItem.bottom -= 3;
                }

                DrawFocusRect (lpdis->hDC, &lpdis->rcItem) ;
            }

            return 0 ;

        case WM_DESTROY :
            DeleteObject(buttonUp);
            DeleteObject(buttonDown);
            DeleteObject(buttonSel);
            DeleteObject(strokeUp);
            DeleteObject(strokeDown);
            DeleteObject(strokeUp2);
            DeleteObject(strokeDown2);

            for (labelNo = 0; labelNo < NLABELS; labelNo++) {
                sprintf(buf, "RAD_%03d", labelNo);
                DeleteObject(labelBitmap[labelNo]);
            }

            if (joinSet != NULL) {
                FreeBlock(joinSet);
                joinSet = NULL;
            }

            if (grayBrush != NULL) {
                DeleteObject(grayBrush);
                grayBrush = NULL;
            }

            if (kanjiBitmap != NULL) {
                DeleteObject(kanjiBitmap);
                kanjiBitmap = NULL;
            }

            nMatches = 0;
            NrSelected = 0;

//          emulateDlg = FALSE;
            hKanjiLookupDlg = 0;
            SetActiveWindow(global.hwnd);

            return 0 ;
    }

    return DefWindowProc (hwnd, message, wParam, lParam) ;

}

long FAR PASCAL _export CursorMovementProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int buttonId;
    int currCol;
    int destCol;
    int labelNo;
    DWORD style;

    switch (message)
    {
        case WM_GETDLGCODE:
            return DLGC_WANTARROWS;

        case WM_KEYDOWN:
            switch (wParam)
            {
                case VK_UP:
#ifdef USEWIN32
                    buttonId = GetWindowLong(hwnd, GWL_ID) - RADID_OFFSET;
#else
                    buttonId = GetWindowWord(hwnd, GWW_ID) - RADID_OFFSET;
#endif USEWIN32
                    for (labelNo = 0;
                         (labelNo < NLABELS) && (labelPos[labelNo] <= buttonId);
                         labelNo++) {
                         ;
                    }
                    destCol = (buttonId+labelNo+RADFIRSTROWOFFSET) % RADBTN_NCOLS;

                    do {

                        if (labelPos[labelNo-1] == buttonId) {
                            labelNo--;
                        }

                        buttonId--;
                        if (buttonId < 0) {
                            buttonId = NRADBUTTONS-1;
                        }


                        if (labelNo == 0) {
                            if (buttonId == NRADBUTTONS-1) {
                                labelNo = NLABELS;
                            }
                        }

                  currCol = (buttonId+labelNo+RADFIRSTROWOFFSET) % RADBTN_NCOLS;
                    } while (currCol != destCol);
                    SetFocus(radButtons[buttonId]);

                    break;
               case VK_DOWN:
#ifdef USEWIN32
                    buttonId = GetWindowLong(hwnd, GWL_ID) - RADID_OFFSET;
#else
                    buttonId = GetWindowWord(hwnd, GWW_ID) - RADID_OFFSET;
#endif USEWIN32
                    for (labelNo = 0;
                         (labelNo < NLABELS) && (labelPos[labelNo] <= buttonId);
                         labelNo++) {
                         ;
                    }

                    destCol = (buttonId+labelNo+RADFIRSTROWOFFSET) % RADBTN_NCOLS;
                    do {
                  buttonId = (buttonId+1) % NRADBUTTONS;

                        if (labelNo == NLABELS) {
                            if (buttonId == 0) {
                                labelNo = 0;
//                              if (labelPos[labelNo] == buttonId) {
                                if (labelPos[0] == 0) {
                                    labelNo++;
                                }
                            }
                        } else if (labelPos[labelNo] == buttonId) {
                            labelNo++;
                        }
                  currCol = (buttonId+labelNo+RADFIRSTROWOFFSET) % RADBTN_NCOLS;
                    } while (currCol != destCol);
                    SetFocus(radButtons[buttonId]);

                    break;
               case VK_LEFT:
#ifdef USEWIN32
                    buttonId = GetWindowLong(hwnd, GWL_ID) - RADID_OFFSET - 1;
#else
                    buttonId = GetWindowWord(hwnd, GWW_ID) - RADID_OFFSET - 1;
#endif USEWIN32
                    if (buttonId < 0) {
                        buttonId = NRADBUTTONS-1;
                    }
                    SetFocus(radButtons[buttonId]);
                    break;
               case VK_RIGHT:
#ifdef USEWIN32
                   SetFocus(radButtons[(GetWindowLong(hwnd, GWL_ID)-RADID_OFFSET+1) % NRADBUTTONS]);
#else
                   SetFocus(radButtons[(GetWindowWord(hwnd, GWW_ID)-RADID_OFFSET+1) % NRADBUTTONS]);
#endif
                    break;
            }
            break;
       case WM_SETFOCUS:
            style = WS_CHILD | WS_VISIBLE | BS_OWNERDRAW;
            SetWindowLong(radButtons[currentRadical], GWL_STYLE, style);
            style = style | WS_TABSTOP;
            SetWindowLong(hwnd, GWL_STYLE, style);
#ifdef USEWIN32
            currentRadical = GetWindowLong(hwnd, GWL_ID) - RADID_OFFSET;
#else
            currentRadical = GetWindowWord(hwnd, GWW_ID) - RADID_OFFSET;
#endif USEWIN32
            break;
    }
    return CallWindowProc(lpfnDefButtonProc, hwnd, message, wParam, lParam);
}


long FAR PASCAL _export StrokesProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    DWORD style;

    switch (message)
    {
       case WM_SETFOCUS:
            style = WS_CHILD | WS_VISIBLE | BS_OWNERDRAW;
            SetWindowLong(strokeButtons[currentStrokes], GWL_STYLE, style);
            style = style | WS_TABSTOP;
            SetWindowLong(hwnd, GWL_STYLE, style);
#ifdef USEWIN32
            currentStrokes = GetWindowLong(hwnd, GWL_ID) - STROKEID_OFFSET;
#else
            currentStrokes = GetWindowWord(hwnd, GWW_ID) - STROKEID_OFFSET;
#endif USEWIN32
            break;
    }
    return CallWindowProc(lpfnDefStrokesProc, hwnd, message, wParam, lParam);
}


LONG FAR PASCAL RadLookupListProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int i, j, r;
    RECT rect;
    HDC hdc;
    HBRUSH hbrush;
    HPEN hpen;
    PAINTSTRUCT ps;
    BYTE far *cbufp;

    switch (message) {

        case WM_CREATE:
              GetClientRect(hwnd, &rect);
              i = GetSystemMetrics(SM_CYHSCROLL) + SYSFONT->height + 1
                  + 2 * KANJILISTBORDER;

              SetWindowPos(hwnd, NULL, 0, 0, rect.right, i, SWP_NOMOVE | SWP_NOZORDER);

              NrInList = (rect.right - 2 * KANJILISTBORDER) / SYSFONT->width;
              NrInList /= 2;

           EdgeGap = (rect.right - (NrInList * 2 - 1) * SYSFONT->width) / 2;
              if (EdgeGap < 0) EdgeGap = 1;

              return (0);

        case WM_GETDLGCODE:
              return (DLGC_WANTARROWS);

        case WM_LBUTTONDOWN:
          {
              int x;

              x = LOWORD(lParam);
              GetClientRect(hwnd, &rect);

              SetFocus(hwnd);

           /* Find out which character it clicks on */

           i = (x + (SYSFONT->width / 2) - EdgeGap) / (2 * SYSFONT->width);

              if (start+i >= nMatches) {
                  MessageBeep(MB_ICONHAND);
                  return (0);
              }

           /* Change the high-light */

              hdc = GetDC(hwnd);
              SetTextColor(hdc, GetSysColor(COLOR_WINDOWTEXT));
              SetBkColor(hdc, GetSysColor(COLOR_WINDOW));

              if ((NrSelected >= start) && (NrSelected < start+NrInList)) {
                  PatBlt(hdc, ((NrSelected-start) * 2 * SYSFONT->width) + EdgeGap - (SYSFONT->width / 2),
                          2, 2 * SYSFONT->width, rect.bottom - 4, DSTINVERT);
              }
              PatBlt(hdc, (i * 2 * SYSFONT->width) + EdgeGap - (SYSFONT->width / 2),
                      2, 2 * SYSFONT->width, rect.bottom - 4, DSTINVERT);
              ReleaseDC(hwnd, hdc);

              NrSelected = start+i;

              if (nMatches < NrInList) {
                  SetScrollPos(hwnd, SB_HORZ, NrSelected, TRUE);
              }

              return (0);
        }

        case WM_LBUTTONDBLCLK:
          {
              int x;

              x = LOWORD(lParam);
              GetClientRect(hwnd, &rect);

              /* Find out which character it clicks on */

              i = (x + (SYSFONT->width / 2) - EdgeGap) / (2 * SYSFONT->width);

              if (start+i >= nMatches) {
                  MessageBeep(MB_ICONHAND);
              } else {
                  SendMessage(GetParent(hwnd), WM_COMMAND, ID_INSFILE, 0L);
              }
              SetFocus(hwnd);

              return (0);
          }

        case WM_PAINT:
              GetClientRect(hwnd, &rect);
              hdc = BeginPaint(hwnd, &ps);

              SetTextColor(hdc, GetSysColor(COLOR_WINDOWTEXT));
              SetBkColor(hdc, GetSysColor(COLOR_WINDOW));
              hbrush = SelectObject(hdc, GetStockObject(NULL_BRUSH));
              hpen = SelectObject(hdc, CreatePen(PS_SOLID, 1, GetSysColor(COLOR_WINDOWFRAME)));

              Rectangle(hdc, 0, 0, rect.right, rect.bottom);

              DeleteObject(SelectObject(hdc, CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNSHADOW))));

              Rectangle(hdc, 1, 1, rect.right - 1, rect.bottom - 1);

              SelectObject(hdc, CreateSolidBrush(GetSysColor(COLOR_WINDOW)));

              //EnableFontCache(FALSE);

              for (i = 0, j = EdgeGap; i < NrInList; i++, j += 2 * SYSFONT->width) {
                  if (start+i >= nMatches) {
                      PatBlt(hdc, j, KANJILISTBORDER, SYSFONT->width,
                              SYSFONT->height, PATCOPY);
                      continue;
                  }

                  r = Jis2Index(joinSet[i+start], SYSFONT->holes);
                  if (r < 0) r = Jis2Index(BADKANJI, SYSFONT->holes);
                  r = GetKanjiBitmap(SYSFONT, r, &cbufp, FALSE);

                  if (start+i == NrSelected) {
                      HBRUSH hbrush1;

                      hbrush1 = SelectObject(hdc, CreateSolidBrush(~GetSysColor(COLOR_WINDOW)));
                      PatBlt(hdc, j - (SYSFONT->width / 2), 2,
                               2 * SYSFONT->width, rect.bottom - 4, PATCOPY);
                      DeleteObject(SelectObject(hdc, hbrush1));
                  }
                  DisplayKanjiBitmap(hdc, j, KANJILISTBORDER + SYSFONT->height,
                        SYSFONT->width, SYSFONT->height, r,
                        (start+i == NrSelected) ? NOTSRCCOPY : SRCCOPY,
                        cbufp, FALSE);
              }

              if (nMatches < NrInList) {
                  SetScrollPos(hwnd, SB_HORZ, NrSelected, TRUE);
              } else {
                  SetScrollPos(hwnd, SB_HORZ, start+NrInList, TRUE);
              }

              DeleteObject(SelectObject(hdc, hpen));
              DeleteObject(SelectObject(hdc, hbrush));

              //EnableFontCache(TRUE);

              EndPaint(hwnd, &ps);
              return (0);

          case WM_CHANGEHIGHLIGHT:
              GetClientRect(hwnd, &rect);
              hdc = GetDC(hwnd);

//            SetTextColor(hdc, GetSysColor(COLOR_WINDOWTEXT));
//            SetBkColor(hdc, GetSysColor(COLOR_WINDOW));
              hbrush = SelectObject(hdc, CreateSolidBrush(GetSysColor(COLOR_WINDOW)));
              for (i = 0, j = EdgeGap; i < NrInList; i++, j += 2 * SYSFONT->width) {
                  if ((start+i == wParam) || (start+i == NrSelected)) {

                      r = Jis2Index(joinSet[i+start], SYSFONT->holes);
                      if (r < 0) r = Jis2Index(BADKANJI, SYSFONT->holes);
                      r = GetKanjiBitmap(SYSFONT, r, &cbufp, FALSE);

                      if (start+i == NrSelected) {
                          HBRUSH hbrush1;

                          hbrush1 = SelectObject(hdc, CreateSolidBrush(~GetSysColor(COLOR_WINDOW)));
                          PatBlt(hdc, j - (SYSFONT->width / 2), 2,
                              2 * SYSFONT->width, rect.bottom - 4, PATCOPY);
                          DeleteObject(SelectObject(hdc, hbrush1));

                      } else {
                          PatBlt(hdc, j - (SYSFONT->width / 2), 2,
                              2 * SYSFONT->width, rect.bottom - 4, PATCOPY);
                      }
                      DisplayKanjiBitmap(hdc, j, KANJILISTBORDER + SYSFONT->height,
                          SYSFONT->width, SYSFONT->height, r,
                          (start+i == NrSelected) ? NOTSRCCOPY : SRCCOPY,
                          cbufp, FALSE);
                  }
              }

              DeleteObject(SelectObject(hdc, hbrush));

              ReleaseDC(hwnd, hdc);
              return (0);

        case WM_KEYDOWN:
              switch (wParam)
              {
                  case VK_SPACE:
                      if (nMatches > 0) {
                          SendMessage(GetParent(hwnd), WM_COMMAND, ID_GETINFO, 0L);
                      }
                      return 0;
                  case VK_LEFT:
                  case VK_UP:
                  case VK_PRIOR:
                  {
                      int oldNrSelected = NrSelected;
                      if (NrSelected <= 0) {
                          return 0;
                      }
                      NrSelected -= (wParam == VK_PRIOR) ? NrInList : 1;
                      if (NrSelected < 0) {
                          NrSelected = 0;
                      }
                      if ((NrSelected < start) || (NrSelected >= start+NrInList)) {
                          start = max(NrSelected - NrInList + 1, 0);
                      } else {
                          SendMessage(hwnd, WM_CHANGEHIGHLIGHT, oldNrSelected, 0L);
                          if (nMatches < NrInList) {
                              SetScrollPos(hwnd, SB_HORZ, NrSelected, TRUE);
                          }
                          return (0);
                      }
                      break;
                  }
              case VK_RIGHT:
              case VK_DOWN:
              case VK_NEXT:
                  {
                      int oldNrSelected = NrSelected;
                      if (NrSelected >= nMatches - 1) {
                          return 0;
                      }
                      NrSelected += (wParam == VK_NEXT) ? NrInList : 1;
                      if (NrSelected >= nMatches) {
                          NrSelected = nMatches - 1;
                      }
                      if ((NrSelected < start) || (NrSelected >= start+NrInList)) {
                          start = NrSelected;
                      } else {
                          SendMessage(hwnd, WM_CHANGEHIGHLIGHT, oldNrSelected, 0L);
                          if (nMatches < NrInList) {
                              SetScrollPos(hwnd, SB_HORZ, NrSelected, TRUE);
                          }
                          return (0);
                      }
                      break;
                  }
                  case VK_HOME:
                      start = 0;
                      NrSelected = 0;
                      break;
                  case VK_END:
                      if (NrSelected >= nMatches - 1) {
                          return 0;
                      }
                      start = max(nMatches - NrInList, 0);
                      NrSelected = nMatches - 1;
                      break;
              }
              InvalidateRect(hwnd, NULL, TRUE);
              return (0);

        case WM_HSCROLL:
              switch (SBCODE)
              {
                  case SB_LINEUP:
                  case SB_PAGEUP:
                      if (start <= 0) {
                          return 0;
                      }
                      start -= (SBCODE == SB_LINEUP) ? 1 : NrInList;
                      if (start < 0) {
                          start = 0;
                      }
                      break;

                  case SB_LINEDOWN:
                  case SB_PAGEDOWN:
                      if (nMatches <= NrInList) {
                          return 0;
                      }
                      start += (SBCODE == SB_LINEDOWN) ? 1 : NrInList;
                      if (start + NrInList > nMatches) {
                          start = nMatches - NrInList;
                      }
                      break;
                  case SB_THUMBPOSITION:
                      if (nMatches >= NrInList) {
                          start = SBPOS - NrInList;
                      }
                      break;
                  default:
                      return (TRUE);
              }
              InvalidateRect(hwnd, NULL, TRUE);

              return (TRUE);
    }

    return (DefWindowProc(hwnd, message, wParam, lParam));
}

